/*
        CommandScroll.m
	by Joe Freeman
        Subprocess Example, Release 2.0
        NeXT Computer, Inc.

	You may freely copy, distribute and reuse the code in this example.
	NeXT disclaims any warranty of any kind, expressed or implied, as to
	its fitness for any particular use.
*/

#import "CommandScroll.h"
#import <appkit/nextstd.h>
#import <appkit/Font.h>
#import <appkit/Text.h>

static id cs_myself;

@implementation CommandScroll

- initFrame:(const NXRect *)frameRect
{	
    [super initFrame:frameRect];
    [self setVertScrollerRequired: YES];
    [self setBackgroundGray: NX_WHITE];
    [self awake];
    return self;
}

- setMASTER:sender
{
	MASTER=sender;
	delegate=sender;
	return self;
}

- awake
    // these initializations implemented here so that this object can be
    // made an IB palatte more easily 
{    
    NXRect textRect;
    id theText;
    cs_myself = self;
    textRect.origin.x = textRect.origin.y = 0.0;
    [self getContentSize: &(textRect.size)];
    theText =
	[[Text alloc]
	    initFrame: &textRect  text:NULL alignment: NX_LEFTALIGNED];
    [theText notifyAncestorWhenFrameChanged:YES];
    [theText setHorizResizable:NO];
    [theText setVertResizable:YES];
    
    textRect.size.width = 0.0;
    [theText setMinSize:&(textRect.size)];
    [self getContentSize: &(textRect.size)];
    textRect.size.height = 1000000;
    [theText setMaxSize:&(textRect.size)];
    
    [[theText superview] setAutoresizeSubviews:YES];
    [[theText superview] setAutosizing: NX_HEIGHTSIZABLE | NX_WIDTHSIZABLE];
    [theText setCharFilter: MyNXFilter];

    [theText setMonoFont:FALSE];
    [self setDocView: theText];
    machineFont = [Font newFont:"Ohlfs" size:10];
    userFont = [Font newFont:"Courier-Bold" size:10];
    xLength=[docView textLength];
    history[0]=(char *)malloc(5);
    strcpy(history[0],"help");
    return self;
}

- setDocView:anObject
{
    [super setDocView:anObject];
    docView = anObject;
    [docView setDelegate:self];
    return self;
}

/*==========================================================
 *
 * New method for subclass
 *
 *==========================================================*/
- clearContents
{
   	[docView setSel:0 :[docView textLength]];
    	[docView setSelFont:machineFont];
    	[docView replaceSel:""];
	[self appendString:"> "];
	DPSFlush();
	return self;

}
 
- appendString:(char *)buffer
    // append the buffer to the end of the text object
{
    [docView setSel:[docView textLength] :0];
    [docView setSelFont:machineFont];
    [docView replaceSel:buffer];
    [docView scrollSelToVisible];
    lastTextCount = [docView textLength];
    xLength=lastTextCount;
    lastPrintLength=xLength;
    return self;
}

- reinsertSpace
{
    [docView setSel:[docView textLength]-1 :[docView textLength]-1];
    [docView setSelFont:userFont];
    [docView replaceSel:" "];
    [docView setSel:[docView textLength] :0];
    nAppended++;
    [docView scrollSelToVisible];
    lastTextCount = [docView textLength];
    return self;
}

- userFont
{
	return userFont;
}

- machineFont
{
	return machineFont;
}

- deleteLastChar
{
	[docView setSel:[docView textLength]-1 :[docView textLength]];
        [docView replaceSel:""];
        [docView setSel:[docView textLength] :0];
        [docView scrollSelToVisible];
        lastTextCount = [docView textLength];
        nAppended--;
        return self;
}

/*==========================================================
 *
 * Text Object Delegation
 *
 *==========================================================*/
 
- (BOOL)textWillChange:textObject 
    // moving the selection to the end before the change means the keys
    // the user hits will always appear at the end of the scroll view
{
    [docView setSel:[docView textLength] :0];
    [docView setSelFont:userFont];
    [docView scrollSelToVisible];
    if (nRemember)     [docView setSel:Begin.cp:End.cp];
    nRemember=0;
    return NO;
}

- setEndFont
{
    int tLength=[docView textLength];
    [docView getSel:&Begin:&End];
    [docView setSel:tLength-1 :tLength];
    [docView setSelFont:userFont];
    [docView setSel:Begin.cp:End.cp];
    return self;
}

- textDidEnd:textObject endChar:(unsigned short)whyEnd  
    // Get this on every Return if using NXFieldFilter
    // really simplistic.  Assumes user does not type ahead
{

    if (whyEnd==16) {
    [docView setSel:[docView textLength] :0];
    [docView replaceSel: "\n"];
    lastTextCount = [docView textLength];
    if (delegate && [delegate respondsTo:@selector(userEntered:)])
	[delegate perform:@selector(userEntered:) with:(void *)"\n"];
    } else {
    	[docView getSel:&Begin:&End];
	nRemember=1;
    }
    return self;
}


- resetLTC
{
	lastTextCount = [docView textLength];
	return self;
}

- textDidGetKeys:textObject isEmpty:(BOOL)flag 
    // Send each character as it is entered.
    // The delta calculation is in case the user has backspaced.
{	
    int	delta;		// the difference in this length from previous
    int	theLength;	// current length of text object
    char *theText, miniBuffer[2];
    BOOL mallocedTheText = NO;

    if (doNotAppend==YES) {
	doNotAppend=NO;
	return self;
    }

    theLength = [docView textLength];

    delta = theLength - lastTextCount;
    if (delta > 0)
    {
        theText = (char *)malloc(delta + 1);
        mallocedTheText = YES;
	[docView
	    getSubstring:theText
	    start:([docView textLength]-delta)
	    length:delta];
	theText[delta] = '\0';

     }
    else
    {
        theText = miniBuffer;
	strcpy(theText,"\177");
    }
  	 if (delegate && [delegate respondsTo:@selector(userEntered:)])
		[delegate perform:@selector(userEntered:) with:(void *)theText];
    	lastTextCount = theLength+nAppended;
        nAppended=0;

    if (mallocedTheText) {
        free(theText);
	mallocedTheText=NO;
    }
   


    return self;
}

- sendSel
{
    NXSelPt Begin_local,End_local;	
   [docView getSel:&Begin_local:&End_local];
   if (Begin_local.cp>lastPrintLength) return self;
   if ((Begin_local.cp!=[docView textLength])&&(Begin_local.cp!=End_local.cp)) {
	int delta=End_local.cp-Begin_local.cp;
        char *theText = (char *)malloc(delta + 1);

	[docView
	    getSubstring:theText
	    start:Begin_local.cp
	    length:delta];
	theText[delta] = '\0';
    if (delegate && [delegate respondsTo:@selector(userEntered:)])
	[delegate perform:@selector(userEntered:) with:(void *)theText];
        free(theText);
    }
   return self;
}

- setLastCommand:(char *)pcString
{
	if (strlen(pcString)==0) {
		return self;
	}
	presentHistory++;
	if (presentHistory>99) presentHistory=0;

	if (history[presentHistory]) free(history[presentHistory]);
	history[presentHistory]=(char *)malloc(strlen(pcString)+1);
	strcpy(history[presentHistory], pcString);
	lastHistory=presentHistory;
	useHistory=presentHistory;
	return self;
}

- getLastCommand
{

	useHistory--; if (useHistory<0) {
		useHistory=99;
		while ((!history[useHistory])&&useHistory) useHistory--;
	}
	if (!history[useHistory]) useHistory=0;
	[docView setSel:xLength:[docView textLength]];
	[docView replaceSel:history[useHistory]];
	[docView setSel:xLength:xLength+strlen(history[useHistory])];

	return self;
}

- getPreviousCommand
{

	useHistory++; if (useHistory>99) {
		useHistory=0;
	}
	if (!history[useHistory]) useHistory=0;
	[docView setSel:xLength:[docView textLength]];
	[docView replaceSel:history[useHistory]];
	[docView setSel:xLength:xLength+strlen(history[useHistory])];

	return self;

}

- setCurrentPosition
{
	xLength=[docView textLength];
        if ((xLength-lastPrintLength)<1) [self setEndFont]; 
	return self;
}

- (int)possibleDeletion
{
	int delta;
    	delta = [docView textLength] - lastPrintLength;
	if (delta>0) return 1;
	return 0;
}

- acceptHistory
{
    	NXSelPt Begin_local,End_local;	
   	[docView getSel:&Begin_local:&End_local];
   	if ((Begin_local.cp!=xLength)||(End_local.cp!= (xLength+ strlen(history[useHistory])))) return self;
	[docView setSel:[docView textLength]:0];
	[self setCurrentPosition];
	xLength=[docView textLength]+1;	// for insert char afterwards
	return self;
}

- rejectHistory
{
    	NXSelPt Begin_local,End_local;	
   	[docView getSel:&Begin_local:&End_local];
   	if ((Begin_local.cp!=xLength)||(End_local.cp!= (xLength+ strlen(history[useHistory])))) return NULL;
	[docView replaceSel:""];
	return self;
}

unsigned short MyNXFilter(unsigned short theChar, int flags, unsigned short charSet)
{
	switch (theChar) {
		case 173: [cs_myself perform:@selector(getLastCommand) 
				with:cs_myself afterDelay:1 cancelPrevious:YES];
			  cs_myself->doNotAppend=YES;
			  return ' ';
		case 175: [cs_myself perform:@selector(getPreviousCommand) 
				with:cs_myself afterDelay:1 cancelPrevious:YES];
			  cs_myself->doNotAppend=YES;
			  return ' ';
		case 172: if([cs_myself rejectHistory]) {
			  	cs_myself->doNotAppend=YES;
			  	return NX_ILLEGAL;
			  }
			  theChar=127;
			  if (![cs_myself possibleDeletion]) return NX_ILLEGAL;
			  [cs_myself setEndFont];
			  break;
		case 174: [cs_myself acceptHistory];
			  return ' ';
		case 3:
		case 13:
			  [cs_myself sendSel];
			  break;
		case 127: if (![cs_myself possibleDeletion]) return NX_ILLEGAL;
			  [cs_myself setEndFont];
	}
	cs_myself->useHistory=cs_myself->presentHistory;
	[cs_myself setCurrentPosition];

	return (NXFieldFilter(theChar, flags, charSet));
}



@end
