
#import "HP48Object.h"
#import "HP48View.h"
#import "CommandScroll.h"
#import <mach/cthreads.h>
#import <mach/mach.h>
#import <mach/message.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "hp48.h"

#define DEBUG 1

extern mutex_t timerMutex;
extern condition_t shutdownCondition;
extern mutex_t shutdownWaitMutex;
extern condition_t shutdownWaitCondition;
extern struct stats stats;
extern CPU cpu;
extern int SHUTDOWNED;
extern int keydirty;
#define MAXKEYROW 10
extern int keyeventlist[MAXKEYROW];
#define SMALL_HP 1
#define BIG_HP 2

int 	last_exec_flags=0;
int      x11_loaded = TRUE;
int	 mustdisplay = TRUE, mustDisplayAnnunc = 0;
int 	 ModeSize=SMALL_HP;
char 	*pcDumpString;
char 	*CALCbuffer;
int	nCurrentDumpLen;

int	nMaxX;
int 	nMinX;
int	nMaxY;
int 	nMinY;
int	nDisplayDirty;
id	 myself;

#define UP 0
#define DOWN 1
#define MOUSE 2
struct tsKeyevent {
	int nType;
	int nTime;
	int nKey;
};
#define MAXDEL 32
#define MAXREC 32
struct tsKeyevent Kqueue[32];
int receivedSomething;
int currentReceived;
struct tsKeyevent Dqueue[MAXDEL];
int deletedSomething;
int currentDeleted;
int waitingForCmd=FALSE;
int globalTimer=0;
int last_annunc_state;
static mutex_t runMutex;
static mutex_t kbdMutex;		
any_t emulatorThread(any_t arg);
void heartbeathandler(DPSTimedEntry tag, double now, void *userData);
// TYPEDEFS

typedef struct { unsigned char bit; int column; } Annunciator_t ;


// LOCAL DATA

#define ANN_LEFT         0x81
#define ANN_RIGHT        0x82
#define ANN_ALPHA        0x84
#define ANN_BATTERY      0x88
#define ANN_BUSY         0x90
#define ANN_IO           0xa0

Annunciator_t AnnunciatorInfo[] = {
	{ ANN_BUSY, 4 },
	{ ANN_LEFT, 0 },
	{ ANN_RIGHT, 1},
	{ ANN_ALPHA, 2 },
	{ ANN_BATTERY, 3 },
	{ ANN_IO, 5 },
	{ 0 }
} ;

// LOCAL PRIVATE FUNCTIONS

void x11_reset_maxmin();
void x11_checkkeyinput();
void setStatus();
int kbd_set();
int kbd_clear();

// IMPLEMENTATION

@implementation HP48Object

- displayAnnunciators
{
	unsigned char newAnnunciators = last_annunc_state;

	if (newAnnunciators != annunciatorState)
	{
		BOOL touched = NO;
		int i;

		lastAnnunciatorState = annunciatorState;
		annunciatorState = newAnnunciators;

		for (i = 0; AnnunciatorInfo[i].bit; i++)
		{
			id cell = [HPCalcAnnView cellAt:0 :AnnunciatorInfo[i].column];
			int state = [cell state];
			int newState = ((AnnunciatorInfo[i].bit & annunciatorState) ==
							AnnunciatorInfo[i].bit);
			if (state != newState)
			{
				[cell setState:newState];
				touched = YES;
			}			
		}
		if (touched)
		{
			[HPCalcAnnView display];
		}
	}

	return self;
}

- init
{
	if (!HPCalcWindow) {
		if (![NXApp loadNibSection:"calculator.nib" owner:self withNames:NO]) {
	   	 	NXLogError ("Could not load calculator.nib");
		}
	}
	NXRunAlertPanel("Warning","This software is under test.\nA lot bugs are to be expected","OK",NULL,NULL);
	x11_reset_maxmin();
	nCurrentDumpLen=1024;
	pcDumpString=(char *)malloc(1025);
	CALCbuffer=(char *)malloc(8192);
	*pcDumpString=0;
	*CALCbuffer='c';
	*(CALCbuffer+1)=0;
	[HPCalcWindow makeKeyAndOrderFront:self];
	[HPCalcLCDView declareReceiver:self];
	DPSFlush();
	fprintf(stderr,"%s\n",[[NXBundle mainBundle] directory]);
	chdir([[NXBundle mainBundle] directory]);
	myself=self;
	
	runMutex=mutex_alloc();
	kbdMutex=mutex_alloc();

	x11_clear();

	cthread_detach(cthread_fork((cthread_fn_t)emulatorThread, (any_t)0));
	[HPCalcWindow makeFirstResponder:HPCalcLCDView];
	[DebugCScroll setMASTER:self];
	DPSAddTimedEntry(0.15, heartbeathandler,self, NX_MODALRESPTHRESHOLD);
	return self;
}

- testinterface
{
	mutex_lock(runMutex);
	if (mustdisplay) {
		mustdisplay=0;
		[HPCalcLCDView display];
		mutex_unlock(runMutex);
	} else mutex_unlock(runMutex);
	x11_checkkeyinput();

	return self;
}

void heartbeathandler(DPSTimedEntry tag, double now, void *userData)
{
	static int mustBroadcastAgain;
	mutex_lock(runMutex);
	if (mustdisplay || mustDisplayAnnunc) {
		if (mustdisplay) {
			mustdisplay=0;
			[((id)userData)->HPCalcLCDView display];
		}
		if (mustDisplayAnnunc) {
			[(id)userData displayAnnunciators];
			mustDisplayAnnunc = 0;
			NXPing();
		}
		mutex_unlock(runMutex);
	} else mutex_unlock(runMutex);
	if (exec_flags^last_exec_flags) {
		if (!exec_flags) {
			setStatus("");
		} else {
			setStatus("stopped");
		}
		last_exec_flags=exec_flags;
	}
	if (mustBroadcastAgain>0) {
		mustBroadcastAgain--;
		condition_broadcast(shutdownCondition);
	} else
		if (keydirty||(receivedSomething!=currentReceived)) 		
			condition_broadcast(shutdownCondition);

	if ((receivedSomething!=currentReceived)&&(!keydirty)) {
		int nKey;
#ifdef ALWAYSLOCK
		mutex_lock(timerMutex);
#endif
		condition_broadcast(shutdownCondition);
		switch (Kqueue[currentReceived].nType) {
			case DOWN:
				if (kbd_set(nKey=Kqueue[currentReceived++].nKey)) {
				#ifdef ALWAYSLOCK
					currentReceived--;
					mutex_unlock(timerMutex);
				#endif
					return;
				}
				break;
			case UP:
				if (kbd_clear(nKey=Kqueue[currentReceived++].nKey)) {
				#ifdef ALWAYSLOCK
					currentReceived--;
					mutex_unlock(timerMutex);
				#endif
					return;
				}
				break;
			case MOUSE:
				if (kbd_set(nKey=Kqueue[currentReceived++].nKey)) {
				#ifdef ALWAYSLOCK
					currentReceived--;
					mutex_unlock(timerMutex);
				#endif
					return;
				}

				[(id)userData perform:@selector(delayedClear:)
					 with:(id)nKey
					 afterDelay:KEY_DELAY
					 cancelPrevious:NO];
				break;
		}
#ifdef ALWAYSLOCK
		mutex_unlock(timerMutex);
#endif
		if (currentReceived>=MAXDEL) currentReceived=0;
		mustBroadcastAgain++;
		
	}
	if (*pcDumpString) {
		mutex_lock(runMutex);
		[((id)userData)->DebugCScroll appendString:pcDumpString];
		*pcDumpString=0;
		mutex_unlock(runMutex);
	}
	globalTimer++;

	return;
}
- receiveKeystroke:sender
{
	int nKey=[[sender selectedCell] tag];
#ifdef ALWAYSLOCK
	if (!mutex_try_lock(timerMutex)) {
		[self perform:@selector(receiveKeystroke:) with:sender afterDelay:10 cancelPrevious:NO];
		return self;
	}
#endif
	Kqueue[receivedSomething].nType=MOUSE;
	Kqueue[receivedSomething].nKey=nKey;
	Kqueue[receivedSomething].nTime=globalTimer;
	receivedSomething++;
	if (receivedSomething>=MAXDEL) receivedSomething=0;
#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	condition_broadcast(shutdownCondition);
	heartbeathandler(0,0.0,self);
	return self;
}

- receiveKeystrokeDirect:(int)nKey
{
#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	Kqueue[receivedSomething].nType=DOWN;
	Kqueue[receivedSomething].nKey=nKey;
	Kqueue[receivedSomething].nTime=globalTimer;
	receivedSomething++;
	if (receivedSomething>=MAXDEL) receivedSomething=0;
#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	condition_broadcast(shutdownCondition);
	heartbeathandler(0,0.0,self);

	switch (nKey)
	{							// workround to handle "autorepeat" keys
	case 0x71: case 0x62: case 0x61: case 0x60: case 0x41: case 0x40:	
		[self perform:@selector(delayedClear:)
			  with:(id)nKey
			  afterDelay:2
			  cancelPrevious:NO];
		break;
	}

	return self;
}


- receiveUpKeystrokeDirect:(int)nKey
{
	switch (nKey)
	{							// workround to handle "autorepeat" keys
	case 0x71: case 0x62: case 0x61: case 0x60: case 0x41: case 0x40:	
		return self;
	}
#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	Kqueue[receivedSomething].nType=UP;
	Kqueue[receivedSomething].nKey=nKey;
	Kqueue[receivedSomething].nTime=globalTimer;
	receivedSomething++;
	if (receivedSomething>=MAXDEL) receivedSomething=0;

#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	condition_broadcast(shutdownCondition);
	heartbeathandler(0,0.0,self);
	return self;

}
- delayedClear:(int)nKey
{
#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	if (kbd_clear(nKey)) [self perform:@selector(delayedClear:) with:nKey afterDelay:1 cancelPrevious:NO];

#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	condition_broadcast(shutdownCondition);
	return self;
}

- (void)callbackTimer:(int)nTimer
{
	switch (nTimer) {
		case 1:
			cpu.timer1=0;
			break;
		case 2:
			cpu.timer2=0;
			break;
	}
	dotimerintr();
	condition_broadcast(shutdownCondition);
	//fprintf(stderr,"TIMER OUT %d\n",nTimer);
}

- makeBig:sender
{
	int x,y;
	[SmallWindow orderOut:self];
	HPCalcWindow=BigWindow;
	HPCalcLCDView=BigLCDView;
	HPCalcAnnView=BigAnnView;
	ModeSize=BIG_HP;
	nMaxY=70;
	nMinY=0;

	for (x=0;x<139;x++) {
		for (y=0;y<69;y++) {
		int nWhereX=(x*2)+1,nWhereY=(y*2)+1;
		biglcd_buffer[nWhereY][nWhereX]= 
		biglcd_buffer[nWhereY+1][nWhereX]= 
		biglcd_buffer[nWhereY][nWhereX+1]= 
		biglcd_buffer[nWhereY+1][nWhereX+1]= 
		lcd_buffer[y][x];
		}
	}
#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	clear_lcd();
	draw_lcd();
#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	x11_flush();
	[HPCalcLCDView declareReceiver:self];
	[HPCalcWindow makeFirstResponder:HPCalcLCDView];
	[BigWindow makeKeyAndOrderFront:self];	
	return self;
}
- makeSmall:sender
{
	int x,y;
	[BigWindow orderOut:self];
	HPCalcWindow=SmallWindow;
	HPCalcLCDView=SmallLCDView;
	HPCalcAnnView=SmallAnnView;
	ModeSize=SMALL_HP;
	nMaxY=70;
	nMinY=0;
	for (x=0;x<139;x++) {
		for (y=0;y<69;y++) {
		int nWhereX=(x*2)+1,nWhereY=(y*2)+1;
		lcd_buffer[y][x]= biglcd_buffer[nWhereY][nWhereX];
		}
	}

#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	clear_lcd();
	draw_lcd();
#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif
	x11_flush();
	[HPCalcLCDView declareReceiver:self];
	[HPCalcWindow makeFirstResponder:HPCalcLCDView];
	[SmallWindow makeKeyAndOrderFront:self];	
	return self;
}


- displayLCD
{
	[HPCalcLCDView display];

	return self;
}


static char hpStatePath[MAXPATHLEN] = { '\0' };

char *HPLoadPath()
{

	if (hpStatePath[0] == '\0')
	{
		struct stat buf;

		strcpy(hpStatePath, NXHomeDirectory());
		strcat(hpStatePath, "/Library/");
		strcat(hpStatePath, [NXApp appName]);

		if (stat(hpStatePath, &buf) == -1)
		{
			if (errno == ENOENT)
			{
				mkdir(hpStatePath, 0644);
				if (stat(hpStatePath, &buf) == 0)
				{
					goto directory_ok;
				}
			}
			strcpy(hpStatePath, "HP48.SAVE");

			return hpStatePath;
		}
	directory_ok:
		strcat(hpStatePath, "/HP48.SAVE");
	}

	return hpStatePath;
}

char *HPSavePath()
{
	return hpStatePath;
}


any_t emulatorThread(any_t arg)
{
		oldmain(0,NULL);
		return (any_t)0;
}
/**********************************************************************/
/*   Remember  state  of  LCD  so  when  we refresh we only draw the  */
/*   bits which are changing.					      */
/**********************************************************************/


void
x11_init(argcp, argv)
int	*argcp;
char	**argv;
{
}


void x11_reset_maxmin()
{
	nDisplayDirty=FALSE;
	nMaxY=0;
	nMinY=70;
	nMaxX=0;
	nMinX=140;

}

// FUNCTIONS

/**********************************************************************/
/*   Routine to draw a nibble into the display.			      */
/**********************************************************************/
void
x11_draw_nibble(x, y, val)
int	x;
int	y;
int	val;
{
	int	r, c;
	PIXEL b;
	nDisplayDirty=TRUE;

	if (y>nMaxY) nMaxY=y;
	if (y<nMinY) nMinY=y;
	if (x>nMaxX) nMaxX=x;
	if (x<nMinX) nMinX=x;

	switch (ModeSize) {

	case SMALL_HP:
		c = (x) * 4 ;
		r = (y); 
		if (val & 1) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		lcd_buffer[r][c++]=b;
		if (val & 2) {
			b=COLOR_1;
		} else {
			b=COLOR_0;	
		}
		lcd_buffer[r][c++]=b;
		if (val & 4) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		lcd_buffer[r][c++]=b;
		if (val & 8) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		lcd_buffer[r][c]=b;
		break;
	case BIG_HP:
		c = (x) * 8;
		r = (y)*2; 
		if (val & 1) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		biglcd_buffer[++r][++c]=b;
		biglcd_buffer[r][c+1]=b;
		biglcd_buffer[r+1][c]=b;
		biglcd_buffer[r+1][c+1]=b; 
		c+=2; 
		if (val & 2) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		biglcd_buffer[r][c]=b;
		biglcd_buffer[r][c+1]=b;
		biglcd_buffer[r+1][c]=b;
		biglcd_buffer[r+1][c+1]=b;
		c+=2;
		if (val & 4) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		biglcd_buffer[r][c]=b;
		biglcd_buffer[r][c+1]=b;
		biglcd_buffer[r+1][c]=b;
		biglcd_buffer[r+1][c+1]=b;
		c+=2;
		if (val & 8) {
			b=COLOR_1;
		} else {
			b=COLOR_0;
		}
		biglcd_buffer[r][c]=b;
		biglcd_buffer[r][c+1]=b;
		biglcd_buffer[r+1][c]=b;
		biglcd_buffer[r+1][c+1]=b;
		break;
	}

}

void x11_checkkeyinput()
{
}

void
x11_flush()
{
	mutex_lock(runMutex);
	mustdisplay++;
	mutex_unlock(runMutex);
	return;
}

void
x11_clear()
{
	int i, lim;
	PIXEL *buffer;

	switch (ModeSize) {
	case BIG_HP:
		lim = BROWS * BCOLUMNS;
		buffer = &(biglcd_buffer[0][0]);
		break;
	case SMALL_HP:
		lim = ROWS * COLUMNS;
		buffer = &(lcd_buffer[0][0]);
		break;
	}
	for (i = 0; i < lim; i++)
	{
		buffer[i] = COLOR_0;
	}
	nMaxX = 130;
	nMaxY = 63;
	nMinX = nMinY = 0;
	nDisplayDirty = TRUE;

	[myself displayLCD];
}

void
x11_refresh_window()
{
	int	n1, n2;
	int	f;
	mutex_lock(runMutex);
	mustdisplay++;

	n1 = read_nibble(0x10B);
	n2 = read_nibble(0x10C);
	mutex_unlock(runMutex);

	f = (n2 << 4) | n1;
	if (f == last_annunc_state)
		return;
	last_annunc_state = f;
	mustDisplayAnnunc++;	
	//printf("ANNUNC: 0x%x\n",f);
	return;
}
void
x11_set_async(flag)
int	flag;
{

}
void
x11_draw_annunc()
{
}
void 
nx_print(char *pcString)
{
#ifdef ALWAYSLOCK
	mutex_lock(timerMutex);
#endif
	mutex_lock(runMutex);
	if ((strlen(pcDumpString)+strlen(pcString))>=nCurrentDumpLen) {
		nCurrentDumpLen+=1024;
		pcDumpString=realloc(pcDumpString, nCurrentDumpLen+1);
	}
	strcat(pcDumpString, pcString);
	mutex_unlock(runMutex);	
#ifdef ALWAYSLOCK
	mutex_unlock(timerMutex);
#endif	
}

void checkTimer()
{
	int n1,n2;
	cpu.debug_mode = TRUE;
	n1 = read_nibble(TIMER1_CTRL);
	n2 = read_nibble(TIMER2_CTRL);
	cpu.debug_mode = FALSE;
			
	if (n2 & 0x01) {
		timer_callback(cpu.timer2,2);
		//fprintf(stderr,"ESTABLISHED TIMER2\n");
	}
	if (n1 & 0x01) {
		timer_callback(cpu.timer1,1);
		//fprintf(stderr,"ESTABLISHED TIMER1\n");
	}
}	


void 
timer_callback(int actin, int timerNo)
{
	int nTimer;
	switch (timerNo) {
		case 1: nTimer=(actin*1000)/16;
			break;
		case 2: nTimer=(actin*1000)/8192;
			break;
	}
	mutex_lock(runMutex);
	[myself perform:@selector(callbackTimer:) with:timerNo afterDelay:nTimer cancelPrevious:NO];
	mutex_unlock(runMutex);
}

#if 0
void
kbd_callback(code)
int	code;
{	int	r, c;
	int	i;
	extern FILE	*cmd_fp;

	if (code<0) return;

	/***********************************************/
	/*   Handle the ATTN key special.	       */
	/***********************************************/
	if (code == 0x8000) {
		do_intr(0, (long *) NULL);
		for (i = 0; i < 9; i++)
			cpu.keybuf.rows[i] ^= 0x8000;
		return;
		}
	r = code >> 4;
	c = 1 << (code & 0x0f);
	/***********************************************/
	/*   Generate an interrupt if key going down.  */
	/***********************************************/
	if ((cpu.keybuf.rows[r] & c) == 0) {
		cpu.keybuf.rows[r] |= c;
		if (cpu.ien)
			do_intr(0, (long *) NULL);
		
		/***********************************************/
		/*   Log command at exact time it occured.     */
		/***********************************************/
		fprintf(cmd_fp, "instr %lx\nc\nkey %x\n", 
			cpu.instr_count, code); 
		}
	else {
		cpu.keybuf.rows[r] &= ~c;
		/***********************************************/
		/*   Turn off ATTN key as well.		       */
		/***********************************************/
		for (i = 0; i < 9; i++)
			cpu.keybuf.rows[i] &= ~0x8000;
		}
	keydirty++;
}
#endif

int
kbd_set(code)
int	code;
{	int	r, c;
	int	i;
	extern FILE	*cmd_fp;

	if (code<0) return 0;

	/***********************************************/
	/*   Handle the ATTN key specially.	       */
	/***********************************************/
	if (code == 0x8000) {
		do_intr(0, (long *) NULL);
		for (i = 0; i < 9; i++)
			cpu.keybuf.rows[i] ^= 0x8000;
		return 0;
	}
	r = code >> 4;
	c = 1 << (code & 0x0f);
	/***********************************************/
	/*   Generate an interrupt if key going down.  */
	/***********************************************/
	if (keyeventlist[r]) return 1;
	cpu.keybuf.rows[r] |= c;

	keyeventlist[r]=1;
		
	if (cpu.ien)
	{
		do_intr(0, (long *) NULL);
	}
	/***********************************************/
	/*   Log command at exact time it occured.     */
	/***********************************************/
	fprintf(cmd_fp, "instr %lx\nc\nkey %x\n", 
		cpu.instr_count, code); 

	switch (code)
	{							// workround to handle "autorepeat" keys
	case 0x71:	case 0x62:
	case 0x61:	case 0x60:
	case 0x41:	case 0x40:	
		break;
	default:
	 	keydirty++;	
	}

	return 0;
}

int
kbd_clear(code)
int	code;
{	int	r, c;
	int	i;

//	extern FILE	*cmd_fp;

	if (code<0) return 0;

	/***********************************************/
	/*   Handle the ATTN key specially.	       */
	/***********************************************/
	if (code == 0x8000) {
		return;
		}
	r = code >> 4;
	c = 1 << (code & 0x0f);

	switch (code)
	{							// workround to handle "autorepeat" keys
	case 0x71:	case 0x62:
	case 0x61:	case 0x60:
	case 0x41:	case 0x40:	
		keyeventlist[r] = 0;
		break;
	}
								// do nothing if key not yet read???
 	if (keyeventlist[r]) return 1;

	cpu.keybuf.rows[r] &= ~c;
	/***********************************************/
	/*   Turn off ATTN key as well.		       */
	/***********************************************/
	for (i = 0; i < 9; i++)
		cpu.keybuf.rows[i] &= ~0x8000;

	return 0;
}


- showConsole:sender
{
	[[DebugCScroll window] makeKeyAndOrderFront:self];
	return self;
}

#define IS_NUMBER(ch) ((ch) >= '0' && (ch) <= '9')

- receiveKeystrokesFromCS:(char)cStroke
{

	char acDummy[2]={0,0};
	acDummy[0]=cStroke;
	mutex_lock(shutdownWaitMutex);
	if ((cStroke==127)||(cStroke==8)) {
		if (strlen(CALCbuffer)>1) {
			*(CALCbuffer+strlen(CALCbuffer)-1)=0;
		}
	} else {		
		strcat(CALCbuffer, acDummy);
	}
	//printf("<%s>\n", CALCbuffer);
	if (strlen(CALCbuffer)>8000) {
		cStroke='\n';		
	}

	if (cStroke!='\n') {
		mutex_unlock(shutdownWaitMutex);
		return self;
	}
	mutex_unlock(shutdownWaitMutex);
	condition_broadcast(shutdownCondition);
	intr_handler();
	condition_broadcast(shutdownWaitCondition);

	return self;
}

- userEntered:(char *)buffer
{
	char *pcPtr=buffer;

	while (*pcPtr) {
		[self receiveKeystrokesFromCS:*pcPtr];
		pcPtr++;
	}
	return self;	
}

void setStatus(char *pcString)
{
	[(id)(myself)->SmallStatus setStringValue:pcString];
	[(id)(myself->BigStatus) setStringValue:pcString];	
}


void getACommand(char *buffer)
{
	mutex_lock(shutdownWaitMutex);
	if (!strlen(CALCbuffer)) {
		waitingForCmd=TRUE;
		condition_wait(shutdownWaitCondition, shutdownWaitMutex);
	}
	waitingForCmd=FALSE;
	strcpy(buffer, CALCbuffer);
	*CALCbuffer=0;
	mutex_unlock(shutdownWaitMutex);
	return ;
}

@end
