// IMPORTS

#import "HP48View.h"
#import "HP48Object.h"

// DEFINES

#define DOWN 1
#define UP 0

#define calcSize  bounds.size

// "GLOBAL" DATA

PIXEL	lcd_buffer[ROWS][COLUMNS];
PIXEL	biglcd_buffer[BROWS][BCOLUMNS];

// LOCAL DATA

int keyeqiv[256];

NXPoint zeroPoint={0.0,0.0};

struct tsVKeyevent {
	int nUpDown;
	int nKey;
};
extern int	nMaxX;
extern int 	nMinX;
extern int	nMaxY;
extern int 	nMinY;
extern int	nDisplayDirty;


#define MAXVEVENT 10

struct tsVKeyevent vEvent[MAXVEVENT+1];

int nCurrentVEvent;
int nTransmittedVEvent;

#define BUTTON_A	0
#define BUTTON_B	1
#define BUTTON_C	2
#define BUTTON_D	3
#define BUTTON_E	4
#define BUTTON_F	5

#define BUTTON_MTH	6
#define BUTTON_PRG	7
#define BUTTON_CST	8
#define BUTTON_VAR	9
#define BUTTON_UP	10
#define BUTTON_NXT	11

#define BUTTON_COLON	12
#define BUTTON_STO	13
#define BUTTON_EVAL	14
#define BUTTON_LEFT	15
#define BUTTON_DOWN	16
#define BUTTON_RIGHT	17

#define BUTTON_SIN	18
#define BUTTON_COS	19
#define BUTTON_TAN	20
#define BUTTON_SQRT	21
#define BUTTON_POWER	22
#define BUTTON_INV	23

#define BUTTON_ENTER	24
#define BUTTON_NEG	25
#define BUTTON_EEX	26
#define BUTTON_DEL	27
#define BUTTON_BS	28

#define BUTTON_ALPHA	29
#define BUTTON_7	30
#define BUTTON_8	31
#define BUTTON_9	32
#define BUTTON_DIV	33

#define BUTTON_SHL	34
#define BUTTON_4	35
#define BUTTON_5	36
#define BUTTON_6	37
#define BUTTON_MUL	38

#define BUTTON_SHR	39
#define BUTTON_1	40
#define BUTTON_2	41
#define BUTTON_3	42
#define BUTTON_MINUS	43

#define BUTTON_ON	44
#define BUTTON_0	45
#define BUTTON_PERIOD	46
#define BUTTON_SPC	47
#define BUTTON_PLUS	48

#define LAST_BUTTON	48

/* SYNONYMS */

#define BUTTON_G	6
#define BUTTON_H	7
#define BUTTON_I	8
#define BUTTON_J	9
#define BUTTON_K	10
#define BUTTON_L	11

#define BUTTON_M	12
#define BUTTON_N	13
#define BUTTON_O	14
#define BUTTON_P	15
#define BUTTON_Q	16
#define BUTTON_R	17

#define BUTTON_S	18
#define BUTTON_T	19
#define BUTTON_U	20
#define BUTTON_V	21
#define BUTTON_W	22
#define BUTTON_X	23

#define BUTTON_Y	24
#define BUTTON_Z	25

// TYPEDEFS

typedef struct button_t {

  char		*name;
  int		code;
  char		*letter;
  char		*left;
  char		*right;

} button_t;


// KEYMAPS

// MAP IS BASED ON A PC KEYBOARD - 
// ESC = ON
// TAB = ALPHA
// (LEFT SHIFT = ORANGE, RIGHT SHIGT = BLUE, BUT THESE ARE MODIFIER KEYS)
// F2-F7 MAP TO MENU KEYS (A-F)
// A-Z MAP TO LETTER KEYS (A-Z)
// ARROW KEYS MAP TO K,P,Q,R
// PERIOD AND COMMA BOTH MAP TO PERIOD FOR DECIMAL POINT

int buttonMapping[256] =
{
-1, -1, -1, BUTTON_ENTER, -1, -1, -1, -1,
BUTTON_DEL, BUTTON_ALPHA, -1, -1, -1, BUTTON_ENTER, -1, -1, // 15
-1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, BUTTON_ON, -1, -1, -1, -1, // 31
 BUTTON_SPC, -1, -1, -1, -1, -1, -1, BUTTON_COLON, -1, -1,
 BUTTON_MUL, BUTTON_PLUS, BUTTON_PERIOD, BUTTON_MINUS, BUTTON_PERIOD, BUTTON_DIV, // 47
 BUTTON_0, BUTTON_1, BUTTON_2, BUTTON_3, BUTTON_4, BUTTON_5,
 BUTTON_6, BUTTON_7, BUTTON_8, BUTTON_9, -1, -1, -1, BUTTON_DEL, -1, -1, // 63
 BUTTON_ALPHA,
 BUTTON_A, BUTTON_B, BUTTON_C, BUTTON_D, BUTTON_E, BUTTON_F,
 BUTTON_G, BUTTON_H, BUTTON_I, BUTTON_J, BUTTON_K, BUTTON_L,
 BUTTON_M, BUTTON_N, BUTTON_O, BUTTON_P, BUTTON_Q, BUTTON_R,
 BUTTON_S, BUTTON_T, BUTTON_U, BUTTON_V, BUTTON_W, BUTTON_X,
 BUTTON_Y, BUTTON_Z, -1, -1, -1, -1, -1,
 -1,
 BUTTON_A, BUTTON_B, BUTTON_C, BUTTON_D, BUTTON_E, BUTTON_F,
 BUTTON_G, BUTTON_H, BUTTON_I, BUTTON_J, BUTTON_K, BUTTON_L,
 BUTTON_M, BUTTON_N, BUTTON_O, BUTTON_P, BUTTON_Q, BUTTON_R,
 BUTTON_S, BUTTON_T, BUTTON_U, BUTTON_V, BUTTON_W, BUTTON_X,
 BUTTON_Y, BUTTON_Z, -1, -1, -1, -1, BUTTON_BS,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0x8*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0x9*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, BUTTON_LEFT, BUTTON_UP, BUTTON_RIGHT, BUTTON_DOWN, //0xa*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0xb*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0xc*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0xd*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, //0xe*
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1	//0xf*
};


button_t buttons_sx[] = {
// 0
  { "A", 0x14, "A", 0, 0, },
  { "B", 0x84, "B", 0, 0, },
  { "C", 0x83, "C", 0, 0, },
  { "D", 0x82, "D", 0, 0, },
  { "E", 0x81, "E", 0, 0, },
  { "F", 0x80, "F", 0, 0, },
// 6
  { "MTH", 0x24, "G", "PRINT", 0, },
  { "PRG", 0x74, "H", "I/O", 0, },
  { "CST", 0x73, "I", "MODES", 0, },
  { "VAR", 0x72, "J", "MEMORY", 0, },
  { "UP", 0x71, "K", "LIBRARY", 0, },
  { "NXT", 0x70, "L", "PREV", 0, },
// 12
  { "COLON", 0x04, "M", "UP", "HOME" },
  { "STO", 0x64, "N", "DEF", "RCL" },
  { "EVAL", 0x63, "O", "aQ", "aNUM" },
  { "LEFT", 0x62, "P", "GRAPH", 0 },
  { "DOWN", 0x61, "Q", "REVIEW", 0 },
  { "RIGHT", 0x60, "R", "SWAP", 0 },
// 18
  { "SIN", 0x34, "S", "ASIN", "b" },
  { "COS", 0x54, "T", "ACOS", "c" },
  { "TAN", 0x53, "U", "ATAN", "d" },
  { "SQRT", 0x52, "V", "e", "f" },
  { "POWER", 0x51, "W", "g", "LOG" },
  { "INV", 0x50, "X", "h", "LN" },
// 24
  { "ENTER", 0x44, 0, "EQUATION", "MATRIX" },
  { "NEG", 0x43, "Y", "EDIT", "VISIT" },
  { "EEX", 0x42, "Z", "2D", "3D" },
  { "DEL", 0x41, 0, "PURGE", 0 },
  { "BS", 0x40, 0, "DROP", "CLR" },
// 29
  { "ALPHA", 0x35, 0, "USR", "ENTRY" },
  { "7", 0x33, "7", "SOLVE", 0 },
  { "8", 0x32, "8", "PLOT", 0 },
  { "9", 0x31, "9", "ALGEBRA", 0 },
  { "DIV", 0x30, 0, "( )", "#" },
// 34
  { "SHL", 0x25, 0, 0, 0 },
  { "4", 0x23, "4", "TIME", 0 },
  { "5", 0x22, "5", "STAT", 0 },
  { "6", 0x21, "6", "UNITS", 0 },
  { "MUL", 0x20, 0, "[ ]", "_" },
// 39
  { "SHR", 0x15, 0, 0, 0 },
  { "1", 0x13, "1", "RAD", "POLAR" },
  { "2", 0x12, "2", "STACK", "ARG" },
  { "3", 0x11, "3", "CMD", "MENU" },
  { "MINUS", 0x10, 0, "i", "j" },
// 44
  { "ON", 0x8000, 0, "CONT", "OFF" },
  { "0", 0x03, "0", "= ", " a" },
  { "PERIOD", 0x02, ".", ",", " k" },
  { "SPC", 0x01, " ", "l ", " m" },
  { "PLUS", 0x00, 0, "{ }", ": :" },

  { 0 }
};

button_t *buttons = buttons_sx;


// IMPLEMENTATION

@implementation HP48View

- initFrame:(const NXRect *)frm
{   

    int i;
    [super initFrame:frm];
    for (i=0;i<255;i++) keyeqiv[i]=(-1);
    keyeqiv['0']=0x3;
    keyeqiv['1']=0x13; 
    keyeqiv['2']=0x12;
    keyeqiv['3']=0x11;
    keyeqiv['4']=0x23;
    keyeqiv['5']=0x22;
    keyeqiv['6']=0x21;
    keyeqiv['7']=0x33;
    keyeqiv['8']=0x32;
    keyeqiv['9']=0x31;
    keyeqiv['+']=0x0;
    keyeqiv['-']=0x10;
    keyeqiv['*']=0x20;
    keyeqiv['/']=0x30;
    keyeqiv[13]=0x44;
    keyeqiv[3]=0x44;
    keyeqiv[44]=0x2;    // ','
    keyeqiv[46]=0x2;    // '.'
    keyeqiv[32]=0x1;    // ' '
    imRect.origin.x=0.0;
    imRect.origin.y=0.0;
    nVEventState=TRUE;
    if (frm->size.width>150) {
	  nDisplayType=2;
#if defined(DEBUG)
	  fprintf(stderr,"HP BIG INIT\n");
#endif
	  xSize=COLUMNS*2;
	  ySize=ROWS*2;

	  imRect.size.width=280.0;
	  imRect.size.height=140.0;

  	  LCD=[[NXBitmapImageRep alloc] initData:(unsigned char *)biglcd_buffer
									pixelsWide:(COLUMNS*2)
									pixelsHigh:(ROWS*2)
									bitsPerSample:BITS_PER_SAMPLE
									samplesPerPixel:SAMPLES_PER_PIXEL
									hasAlpha:HAS_ALPHA
									isPlanar:IS_PLANAR
									colorSpace:COLOR_SPACE
									bytesPerRow:0 bitsPerPixel:0];
   	  data[0] = (unsigned char *)biglcd_buffer;
    } else {
	  nDisplayType=1;
#if defined(DEBUG)
	  fprintf(stderr,"HP SMALL INIT\n");
#endif
	  xSize=COLUMNS;
	  ySize=ROWS;
	  imRect.size.width=140.0;
	  imRect.size.height=70.0;
  	  LCD=[[NXBitmapImageRep alloc] initData:(unsigned char *)lcd_buffer
									pixelsWide:COLUMNS
									pixelsHigh:ROWS
									bitsPerSample:BITS_PER_SAMPLE
									samplesPerPixel:SAMPLES_PER_PIXEL
									hasAlpha:HAS_ALPHA
									isPlanar:IS_PLANAR
									colorSpace:COLOR_SPACE
									bytesPerRow:0 bitsPerPixel:0];
	  data[0] = (unsigned char *)lcd_buffer;
    }
    LCDImage=[[NXImage alloc] initSize:&imSize];
    [LCDImage useRepresentation:LCD];
    [self setFlipped:YES];
    return self;
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- declareReceiver:sender
{
	receiver=sender;
	return self;
}

- keyDown:(NXEvent *)theEvent
{
	int nFaster;
#if 1
	if ((theEvent->data.key.repeat == 0) &&
		(nFaster = buttonMapping[theEvent->data.key.charCode]) >= 0)
	{
		[receiver receiveKeystrokeDirect:buttons[nFaster].code];
		return self;
	}
#else
	if ((nFaster= keyeqiv[theEvent->data.key.charCode])>=0) {
		[receiver receiveKeystrokeDirect:nFaster];
		return YES;
	}
#endif
	//printf("Event:%d is %d\n",theEvent->data.key.charCode,nFaster);
	return self;
}
- keyUp:(NXEvent *)theEvent
{
	int nFaster;
#if 1
	if ((theEvent->data.key.repeat == 0) &&
		(nFaster = buttonMapping[theEvent->data.key.charCode]) >= 0)
	{
		[receiver perform:@selector(receiveUpKeystrokeDirect:)
				  with:(id)buttons[nFaster].code
				  afterDelay:KEY_DELAY
				  cancelPrevious:NO];
		return self;
	}
	else printf("CharCode = 0x%x\n", theEvent->data.key.charCode);
#else
	if ((theEvent->data.key.repeat == 0) &&
		((nFaster= keyeqiv[theEvent->data.key.charCode])>=0)) {
		[receiver perform:@selector(receiveUpKeystrokeDirect:) with:nFaster afterDelay:KEY_DELAY cancelPrevious:NO];
		return YES;
	}
#endif
	//printf("Event:%d is %d\n",theEvent->data.key.charCode,nFaster);
	return self;
}

- drawSelf:(NXRect *)rects :(int)rectCount 
{
   register int virYSize;
   // way one: draw the LCD using NXBitmapImageRep draw op 
   // [LCD draw];
   // way two: (nofuc for now): draw the LCD using NXImage composite op 
   // [LCDImage composite:NX_COPY toPoint:&zeroPoint];
   // way three: using NXDrawBitmap
   if (!nDisplayDirty) return self;

   switch (nDisplayType) {
	case 1: // small display
   		*data=(unsigned char *)&lcd_buffer[nMinY][0];
   		imRect.size.width=xSize;
   		imRect.size.height=virYSize=(nMaxY-nMinY)+1;
   		imRect.origin.y=nMinY+8;
   		NXDrawBitmap(&imRect, xSize, virYSize,
					 BITS_PER_SAMPLE,SAMPLES_PER_PIXEL,BITS_PER_PIXEL,
					 (xSize * BYTES_PER_SAMPLE),
					 IS_PLANAR,HAS_ALPHA, COLOR_SPACE,(&data));
		break;
	case 2: // big display
   		*data=(unsigned char *)&biglcd_buffer[nMinY*2][0];
   		imRect.size.width=xSize;
   		imRect.size.height=virYSize=1+(((nMaxY-nMinY)+1)*2);
   		imRect.origin.y=(nMinY*2)+8;
   		NXDrawBitmap(&imRect,xSize, virYSize,
					 BITS_PER_SAMPLE,SAMPLES_PER_PIXEL,BITS_PER_PIXEL,
					 (xSize * BYTES_PER_SAMPLE),
					 IS_PLANAR,HAS_ALPHA, COLOR_SPACE, (&data));
		break;
   }
   x11_reset_maxmin();
   nRedrawCount++;
   return self;
}

// KEYBOARD ENHANCEMENTS

- (int)alphaCode { return buttons[BUTTON_ALPHA].code; }
- (int)leftShiftCode { return buttons[BUTTON_SHL].code; }
- (int)rightShiftCode { return buttons[BUTTON_SHR].code; }

- flagsChanged:(NXEvent *)theEvent
{
	int newFlag = theEvent->flags & NX_ALPHASHIFTMASK;

	if (newFlag != alpha)
	{
		alpha = newFlag;
#if defined(DEBUG) && 0
		printf("Alpha %s\n", alpha ? "up" : "down");
#endif
		if (alpha)
			[receiver receiveKeystrokeDirect:[self alphaCode]];
		else
			[receiver perform:@selector(receiveUpKeystrokeDirect:)
					  with:(id)[self alphaCode]
					  afterDelay:KEY_DELAY
					  cancelPrevious:NO];
	}
	newFlag = theEvent->flags & NX_NEXTLSHIFTKEYMASK;
	if (newFlag != leftShift)
	{
		leftShift = newFlag;
#if defined(DEBUG) && 0
		printf("Left Shift %s\n", leftShift ? "up" : "down");
#endif
		if (leftShift)
			[receiver receiveKeystrokeDirect:[self leftShiftCode]];
		else
			[receiver perform:@selector(receiveUpKeystrokeDirect:)
					  with:(id)[self leftShiftCode]
					  afterDelay:KEY_DELAY
					  cancelPrevious:NO];
	}
	newFlag = theEvent->flags & NX_NEXTRSHIFTKEYMASK;
	if (newFlag != rightShift)
	{
		rightShift = newFlag;
#if defined(DEBUG) && 0
		printf("Right Shift %s\n", rightShift ? "up" : "down");
#endif
		if (rightShift)
			[receiver receiveKeystrokeDirect:[self rightShiftCode]];
		else
			[receiver perform:@selector(receiveUpKeystrokeDirect:)
					  with:(id)[self rightShiftCode]
					  afterDelay:KEY_DELAY
					  cancelPrevious:NO];
	}

	return self;
}

@end
