/**********************************************************************/
/*   Routines to perform 64 bit arithmetic.			      */
/**********************************************************************/
# include	"hp48.h"

# define	MASK_16	0xffff

# define	CHECK_DIGIT(x)	if (x > 9 || x < 0) printf("%05lx: BAD DIGIT****\n", cpu.pc);
extern CPU	cpu;
int	show_dec;

void
add_64(wp1, wp2)
WORD64	*wp1;
WORD64	*wp2;
{	unsigned short	s1[4], s2[4];
	unsigned long	res[4];

	/***********************************************/
	/*   We  could  do  this  a number of ways --  */
	/*   many  involving  testing  the endianness  */
	/*   of  the  CPU.  We  avoid this at present  */
	/*   at  do  things with shorts and longs. We  */
	/*   assume  a  short  is  16 bits and a long  */
	/*   is  32  bits.  This  ensures  it will be  */
	/*   more portable at the expense of speed.    */
	/***********************************************/
	s1[0] = (unsigned short) (wp1->lo & MASK_16);
	s1[1] = (unsigned short) ((wp1->lo >> 16) & MASK_16);
	s1[2] = (unsigned short) (wp1->hi & MASK_16);
	s1[3] = (unsigned short) ((wp1->hi >> 16) & MASK_16);

	s2[0] = (unsigned short) (wp2->lo & MASK_16);
	s2[1] = (unsigned short) ((wp2->lo >> 16) & MASK_16);
	s2[2] = (unsigned short) (wp2->hi & MASK_16);
	s2[3] = (unsigned short) ((wp2->hi >> 16) & MASK_16);

	res[0] = (unsigned long) s1[0] + (unsigned long) s2[0];
	res[1] = (unsigned long) s1[1] + (unsigned long) s2[1];
	res[2] = (unsigned long) s1[2] + (unsigned long) s2[2];
	res[3] = (unsigned long) s1[3] + (unsigned long) s2[3];
	if (res[0] >> 16)
		res[1]++;
	if (res[1] >> 16)
		res[2]++;
	if (res[2] >> 16)
		res[3]++;
	if (res[3] >> 16)
		cpu.cy = 1;
	else
		cpu.cy = 0;

	wp1->lo = (res[1] << 16) | (res[0] & 0xffffL);
	wp1->hi = (res[3] << 16) | (res[2] & 0xffffL);
}
void
sub_64(wp1, wp2)
WORD64	*wp1;
WORD64	*wp2;
{	unsigned short	s1[4], s2[4];
	unsigned long	res[4];

	/***********************************************/
	/*   We  could  do  this  a number of ways --  */
	/*   many  involving  testing  the endianness  */
	/*   of  the  CPU.  We  avoid this at present  */
	/*   at  do  things with shorts and longs. We  */
	/*   assume  a  short  is  16 bits and a long  */
	/*   is  32  bits.  This  ensures  it will be  */
	/*   more portable at the expense of speed.    */
	/***********************************************/
	s1[0] = (unsigned short) (wp1->lo & MASK_16);
	s1[1] = (unsigned short) ((wp1->lo >> 16) & MASK_16);
	s1[2] = (unsigned short) (wp1->hi & MASK_16);
	s1[3] = (unsigned short) ((wp1->hi >> 16) & MASK_16);

	s2[0] = (unsigned short) (wp2->lo & MASK_16);
	s2[1] = (unsigned short) ((wp2->lo >> 16) & MASK_16);
	s2[2] = (unsigned short) (wp2->hi & MASK_16);
	s2[3] = (unsigned short) ((wp2->hi >> 16) & MASK_16);

	res[0] = (unsigned long) s1[0] - (unsigned long) s2[0];
	res[1] = (unsigned long) s1[1] - (unsigned long) s2[1];
	res[2] = (unsigned long) s1[2] - (unsigned long) s2[2];
	res[3] = (unsigned long) s1[3] - (unsigned long) s2[3];
	if (res[0] >> 16)
		res[1]--;
	if (res[1] >> 16)
		res[2]--;
	if (res[2] >> 16)
		res[3]--;
	if (res[3] >> 16)
		cpu.cy = 1;
	else
		cpu.cy = 0;

	wp1->lo = (res[1] << 16) | (res[0] & 0xffffL);
	wp1->hi = (res[3] << 16) | (res[2] & 0xffffL);
}
/**********************************************************************/
/*   Function   to   be   called  from  debugger  only  for  testing  */
/*   arithmetic.						      */
/**********************************************************************/
void
s64(a, b)
long	a, b;
{
	WORD64	w1;
	WORD64	w2;

	w1.lo = a;
	w1.hi = 0;
	w2.lo = b;
	w2.hi = 0;
	sub_64(&w1, &w2);
	printf("Result=%08lx:%08lx Cy=%d\n",
		w1.hi, w1.lo, cpu.cy);
}
/**********************************************************************/
/*   Perform BCD arithmetic.					      */
/**********************************************************************/
void
add_dec64(w1, w2)
WORD64	*w1;
WORD64	*w2;
{	unsigned char	a1[16];
	unsigned char	a2[16];
	int	i, cy;

	if (show_dec)
		printf("$%05lx: %08lx:%08lx + %08lx:%08lx => ", cpu.pc,
			w1->hi, w1->lo,
			w2->hi, w2->lo);
	/***********************************************/
	/*   Break the src & dst into the digits.      */
	/***********************************************/
	for (i = 0; i < 8; i++) {
		a1[i] = (w1->lo >> (i*4)) & 0x0f;
		a2[i] = (w2->lo >> (i*4)) & 0x0f;
		CHECK_DIGIT(a1[i]);
		CHECK_DIGIT(a2[i]);
		}
	for (; i < 16; i++) {
		a1[i] = (w1->hi >> ((i-8)*4)) & 0x0f;
		a2[i] = (w2->hi >> ((i-8)*4)) & 0x0f;
		}

	for (i = cy = 0; i < 16; i++) {
		a1[i] += a2[i] + cy;
		if (a1[i] >= 10) {
			a1[i] -= 10;
			cy = 1;
			}
		else
			cy = 0;
		}
	cpu.cy = cy;
	/***********************************************/
	/*   Now put the result back together again.   */
	/***********************************************/
	w1->lo = 0;
	w1->hi = 0;
	for (i = 0; i < 8; i++) {
		w1->lo |= a1[i] << (i*4);
		}
	for (; i < 16; i++) {
		w1->hi |= a1[i] << ((i-8)*4);
		}
	if (show_dec)
		printf("%08lx:%08lx cy=%d\n", w1->hi, w1->lo, cpu.cy);
}
/**********************************************************************/
/*   Perform 64 bit BCD subtraction.				      */
/**********************************************************************/
void
sub_dec64(w1, w2)
WORD64	*w1;
WORD64	*w2;
{	unsigned char	a1[16];
	unsigned char	a2[16];
	int	i, cy;

	if (show_dec)
		printf("$%05lx: %08lx:%08lx - %08lx:%08lx => ", cpu.pc,
			w1->hi, w1->lo,
			w2->hi, w2->lo);
	/***********************************************/
	/*   Break the src & dst into the digits.      */
	/***********************************************/
	for (i = 0; i < 8; i++) {
		a1[i] = (w1->lo >> (i*4)) & 0x0f;
		a2[i] = (w2->lo >> (i*4)) & 0x0f;
		CHECK_DIGIT(a1[i]);
		CHECK_DIGIT(a2[i]);
		}
	for (; i < 16; i++) {
		a1[i] = (w1->hi >> ((i-8)*4)) & 0x0f;
		a2[i] = (w2->hi >> ((i-8)*4)) & 0x0f;
		}
	for (i = cy = 0; i < 16; i++) {
		a1[i] -= a2[i] + cy;
		if (a1[i] >= 10) {
			a1[i] += 10;
			cy = 1;
			}
		else
			cy = 0;
		}
	cpu.cy = cy;
	/***********************************************/
	/*   Now put the result back together again.   */
	/***********************************************/
	w1->lo = 0;
	w1->hi = 0;
	for (i = 0; i < 8; i++) {
		w1->lo |= a1[i] << (i*4);
		}
	for (; i < 16; i++) {
		w1->hi |= a1[i] << ((i-8)*4);
		}
	if (show_dec)
		printf("%08lx:%08lx cy=%d\n", w1->hi, w1->lo, cpu.cy);
}
/**********************************************************************/
/*   Routine  to  add  two  BCD numbers where each number <= 32 bits  */
/*   long							      */
/**********************************************************************/
unsigned long
add_dec(w1, w2, n)
unsigned long	w1, w2;
int	n;
{	unsigned char	a1[16];
	unsigned char	a2[16];
	int	i, cy;

	if (show_dec)
		printf("$%05lx: (#%d) %08lx + %08lx => ", cpu.pc, n,
			w1, w2);
	/***********************************************/
	/*   Break the src & dst into the digits.      */
	/***********************************************/
	for (i = 0; i < n; i++) {
		a1[i] = (w1 >> (i*4)) & 0x0f;
		a2[i] = (w2 >> (i*4)) & 0x0f;
		CHECK_DIGIT(a1[i]);
		CHECK_DIGIT(a2[i]);
		}

	for (i = cy = 0; i < n; i++) {
		a1[i] += a2[i] + cy;
		if (a1[i] >= 10) {
			a1[i] -= 10;
			cy = 1;
			}
		else
			cy = 0;
		}
	cpu.cy = cy;
	/***********************************************/
	/*   Now put the result back together again.   */
	/***********************************************/
	w1 = 0;
	for (i = 0; i < n; i++)
		w1 |= a1[i] << (i*4);
	if (show_dec)
		printf("%08lx cy=%d\n", w1, cy);
	return w1;
}
/**********************************************************************/
/*   Routine  to  subtract  two  BCD numbers where each number <= 32  */
/*   bits long							      */
/**********************************************************************/
unsigned long
sub_dec(w1, w2, n)
unsigned long	w1, w2;
int	n;
{	unsigned char	a1[16];
	unsigned char	a2[16];
	int	i, cy;

	if (show_dec)
		printf("$%05lx: (#%d) %08lx - %08lx => ", cpu.pc, n,
			w1, w2);
	/***********************************************/
	/*   Break the src & dst into the digits.      */
	/***********************************************/
	for (i = 0; i < n; i++) {
		a1[i] = (w1 >> (i*4)) & 0x0f;
		a2[i] = (w2 >> (i*4)) & 0x0f;
		CHECK_DIGIT(a1[i]);
		CHECK_DIGIT(a2[i]);
		}

	for (i = cy = 0; i < n; i++) {
		a1[i] -= a2[i] + cy;
		if (a1[i] >= 10) {
			a1[i] += 10;
			cy = 1;
			}
		else
			cy = 0;
		}
	cpu.cy = cy;
	/***********************************************/
	/*   Now put the result back together again.   */
	/***********************************************/
	w1 = 0;
	for (i = 0; i < n; i++)
		w1 |= a1[i] << (i*4);
	if (show_dec)
		printf("%08lx cy=%d\n", w1, cy);
	return w1;
}
/**********************************************************************/
/*   Compute the tens complement of a number.			      */
/**********************************************************************/
unsigned long
tens_complement(w, n)
unsigned long	w;
int	n;
{	unsigned char	a1[16];
	int	i, cy;

	if (show_dec)
		printf("$%05lx: (#%d) -%08lx => ", cpu.pc, n, w);
	/***********************************************/
	/*   Break number apart.		       */
	/***********************************************/
	for (i = 0; i < n; i++) {
		a1[i] = (w >> (i*4)) & 0x0f;
		CHECK_DIGIT(a1[i]);
		}

	cy = 1;
	for (i = 0; i < n; i++) {
		a1[i] = 9 - a1[i] + cy;
		if (a1[i] >= 10) {
			a1[i] -= 10;
			cy = 1;
			}
		else
			cy = 0;
		}
	cpu.cy = cy;
	/***********************************************/
	/*   Now put the result back together again.   */
	/***********************************************/
	w = 0;
	for (i = 0; i < n; i++)
		w |= a1[i] << (i*4);
	if (show_dec)
		printf("%08lx cy=%d\n", w, cy);
	return w;
}
