/**********************************************************************/
/*   Code to implement command line debugger.			      */
/**********************************************************************/

# include	"hp48.h"
# include	"dis48.h"
# include	<signal.h>
# include "proto.h"
# include <string.h>
# include <fcntl.h>

char acOut[512];

extern int x11_loaded;
extern int show_dec;
extern CPU cpu;

/**********************************************************************/
/*   Some    constants   in   the   HP48   ROM   for   disassembling  */
/*   code/program/unit objects.					      */
/**********************************************************************/
# define	DOBINT	0x02911
# define	DOEREAL	0x02955
# define	DOREAL	0x02933
# define	DOCHAR	0x029BF
# define	DOHSTR	0x02A4E
# define	DOCSTR	0x02A2C
# define	DOEXT	0x02ADA
# define	DOCOL	0x02D9D
# define	DOCODE	0x02DCC
# define	DOIDNT	0x02E48
# define	DOLAM	0x02E6D
# define	SEMI	0x0312B
/**********************************************************************/
/*   Number  of  breakpoints.  Last  entry  in table used for finish  */
/*   command.							      */
/**********************************************************************/
# define	MAX_BREAKPOINTS	16
int	num_bkpts;
int	num_ranges;
struct breakpoint {
	long	addr;
	long	end_addr;
	int	flags;
	} bkpt_tbl[MAX_BREAKPOINTS+1];
/**********************************************************************/
/*   If none-zero, instruction number to stop at.		      */
/**********************************************************************/
long	stop_at_instr = 0;
/**********************************************************************/
/*   Table used for looking up register names.			      */
/**********************************************************************/
struct map {
	char	*name;
	void	*ptr;
	int	size;
	} reg_tbl[] = {
	{"d0",	(void *) &cpu.D0, sizeof cpu.D0},
	{"d1",	(void *) &cpu.D1, sizeof cpu.D1},
	{"a",	(void *) &cpu.A, sizeof cpu.A},
	{"b",	(void *) &cpu.B, sizeof cpu.B},
	{"c",	(void *) &cpu.C, sizeof cpu.C},
	{"d",	(void *) &cpu.D, sizeof cpu.D},
	{"r0",	(void *) &cpu.R0, sizeof cpu.R0},
	{"r1",	(void *) &cpu.R1, sizeof cpu.R1},
	{"r2",	(void *) &cpu.R2, sizeof cpu.R2},
	{"r3",	(void *) &cpu.R3, sizeof cpu.R3},
	{"r4",	(void *) &cpu.R4, sizeof cpu.R4},
	{0}
	};

void	do_break();
void	do_config();
void	do_continue();
void	do_decimal();
void	do_delete();
void	do_disass();
void	do_finish();
void	do_go();
void	do_grob();
void	do_help();
void	do_io();
void	do_instr();
void	do_intr();
void	do_key();
void	do_lcd();
void	do_load();
void	do_modify();
void	do_pagelen();
void	do_print();
void	do_quiet();
void	do_mod();
void	do_new();
void	dump_ram();
void	dump_rom();
void	do_reg();
void	do_rpl();
void	do_save();
void	do_step();
void	do_trace();
void	do_watch();
void	do_where();

struct cmd {
	char	*name;
	void	(*func)();
	char	*help;
	} cmd_tbl[] = {
{"break",	do_break,	"break <addr>	Set breakpoint address"},
{"cont", 	do_continue,	"cont [n]	Continue until next [n] breakpoints",},
{"config", 	do_config,	"config		Display history of CONFIG command"},
{"clear", 	clear_lcd,	"clear		Clear X11 LCD window"},
{"delete",	do_delete,	"delete [n]	Delete specified break/watchpoint"},
{"decimal", 	do_decimal,	"decimal n	Set decimal trace flag to n"},
{"disass", 	do_disass,	"disass staddr endaddr\n		Disassemble section of memory"},
{"finish",	do_finish,	"finish		Continue execution until current routine returns"},
{"exit",	(void (*)()) exit, "exit	Terminate emulator"},
{"go", 		do_go,		"go addr	Set PC to addr"},
{"grob",	do_grob,	"grob addr	Display GROB in hex/stars at address"},
{"help",	do_help,	"help		Print out this information"},
{"instr",	do_instr,	"instr nn	Stop when instruction number nn is executed"},
{"io", 		do_io,		"io nn		Set IO watch moed to nn"},
{"intr", 	do_intr,	"intr		Generate an interrupt"},
{"key", 	do_key,		"key nn		Pretend key nn pressed. If nn > ff clear keybd map"},
{"lcd", 	do_lcd,		"lcd		Update X11 LCD display"},
{"load", 	do_load,	"load [file]	Load calculator image from  file"},
{"modify",	do_modify,	"modify addr vvvv.. Load nibbles vvvvv... into address"},
{"new",		do_new,		"new n		Set new instruction mode to n."},
{"print",	do_print,	"print nnn	Try and print nnn symbolically"},
{"pagelen",	do_pagelen,	"pagelen nn	Set pagelen for --More-- prompt"},
{"quiet",	do_quiet,	"quiet		Toggle trace mode verbosity"},
{"ram", 	dump_ram,	"ram staddr [endaddr] Print RAM in specified range"},
{"range", 	do_watch,	"range loaddr hiaddr Put a watchpoint on a memory area"},
{"rom", 	dump_rom,	"rom staddr [endaddr] Print ROM in specified range"},
{"reg", 	do_reg,		"reg [A|B|C|D|st nn]	Dump registers or modify specified register."},
{"rpl", 	do_rpl,		"rpl <addr>	Disassemble memory as RPL code"},
{"step", 	do_step,	"step [nn]	Single step (nn) instructions."},
{"save", 	do_save,	"save [file]	Save calculator image in file"},
{"trace", 	do_trace,	"trace [nn]	Trace execution of nn instructions"},
{"watch", 	do_watch,	"watch loaddr hiaddr Put a watchpoint on a memory area"},
{"where", 	do_where,	"where		Print stack trace"},
{0, 0}
	};
static char hex[] = "0123456789abcdef";

long	print_str PROTO((long, long));
long	dis_real PROTO((long, int, int));
long	dis_code PROTO((long, long));
long	dis_program PROTO((long, long));
void	intr_handler();
long	decode_arg PROTO((char *));
void	print_64 PROTO((WORD64));
int	read_nibble PROTO((long));
void	print_mem PROTO((long, long, int));
int	print_grob PROTO((long, int));

int	new_flag = FALSE;
int	io_flag = FALSE;
int	step_quiet = FALSE;
int	quit_flag;
int cont_count;
int	line_no;
int	pagelen = 22;
int	indent;

extern FILE *cmd_fp;
extern FILE	*input_fp;
extern struct stats stats;
extern CPU cpu;
extern char *yellow, *green, *blue;

/**********************************************************************/
/*   Read  a  command  and  when  we  return  we  expect to continue  */
/*   execution  until  some  new  interesting  event happens or user  */
/*   hits ^C.							      */
/**********************************************************************/
void
get_command()
{	
	char	*cp;
	char	buf[BUFSIZ];
	struct cmd *cmdp;
	int	argc;
	long	args[8];
	char	*strv[8];
	static int first_time = TRUE;
	static int hasgotcr=1;

	/***********************************************/
	/*   Turn off the finish command.	       */
	/***********************************************/
	if (bkpt_tbl[MAX_BREAKPOINTS].addr) {
		bkpt_tbl[MAX_BREAKPOINTS].addr = 0;
		num_bkpts--;
		}
	if (cont_count && cont_count-- > 1) {
		if (exec_flags & EXEC_STEP)
			exec_flags |= EXEC_STOPPED;
		return;
		}

	if (first_time) {
		signal(SIGINT, intr_handler);
		first_time = FALSE;
		}
	exec_flags &= ~(EXEC_TRACE | EXEC_STEP | EXEC_RUN);
	cpu.debug_mode = TRUE;

	x11_set_async(FALSE);
	while ((exec_flags & EXEC_RUN) == 0) {
		x11_refresh_window();
		line_no = 0;
		if (input_fp == stdin) {
			sprintf(acOut,"\n%sInstr#%04lx (bad#%ld): %s", 
				yellow, cpu.instr_count, stats.illegal_instr, green);
			if (hasgotcr) { 
				nx_print(acOut);
				hasgotcr=0;
			}
			}
/*		if (fgets(buf, sizeof buf, input_fp) == NULL) {
			if (input_fp == stdin)
				exit(0);
			sprintf(acOut,"<EOF>\n"); nx_print(acOut);
			input_fp = stdin;
			continue;
			}
*/
		
		getACommand(buf);

		quit_flag = FALSE;
		if (input_fp != stdin) {
			sprintf(acOut,"Command read: %s", buf); 
			nx_print(acOut);
		}
		if (buf[strlen(buf)-1] == '\n') {
			buf[strlen(buf)-1] = '\0';
			hasgotcr++;
		} else hasgotcr=0;

		if (*buf == '#')
			continue;
		if (*buf == '\0') {
			exec_flags |= EXEC_TRACE;
			break;
			}
		cp = strtok(buf, " \t");
		for (cmdp = cmd_tbl; cmdp->name; cmdp++)
			if (strncmp(cp, cmdp->name, strlen(cp)) == 0)
				break;

		argc = 0;
		while ((cp = strtok((char *) NULL, " \t")) != (char *) NULL) {
			strv[argc] = cp;
			args[argc++] = decode_arg(cp);
			}
		if (cmdp->func)
			(*cmdp->func)(argc, args, strv);
	}
	cpu.debug_mode = FALSE;
	exec_flags &= ~(EXEC_RUN | EXEC_STOPPED);
	if (exec_flags & EXEC_STEP)
		exec_flags |= EXEC_STOPPED;
	if (num_bkpts)
		exec_flags |= EXEC_BKPT;
	else
		exec_flags &= ~EXEC_BKPT;
	x11_set_async(TRUE);
}
/**********************************************************************/
/*   Print out help information.				      */
/**********************************************************************/
void
do_help()
{	int	i;

	for (i = 0; cmd_tbl[i].name; i++) {
		if (print(cmd_tbl[i].help) < 0)
			break;
		nx_print("\n");
	}
}
void
intr_handler()
{
	exec_flags |= EXEC_STOPPED;
	quit_flag = TRUE;
	cont_count = 0;
}
/**********************************************************************/
/*   Print out history of config instructions.			      */
/**********************************************************************/
void
do_config(argc, argv)
int	argc;
long	*argv;
{	int	i;
	for (i = 0; i < MAX_CONFIGS; i++)
		if (cpu.configs[i]) {
			sprintf(acOut,"CONFIG %05lx\n", cpu.configs[i]);
			nx_print(acOut);
		}
	sprintf(acOut,"RAM 32K: %05lx\n", cpu.ram_32k); nx_print(acOut);
	if (argc)
		cpu.ram_32k = argv[0];
}
/**********************************************************************/
/*   Setup  key  state  to  pretend  a  key  is down and if possible  */
/*   cause  an  interrupt.  Key  is  specified  as  two  hex  digits  */
/*   corresponding to row and column				      */
/**********************************************************************/
void
do_key(argc, argv)
int	argc;
long	*argv;
{	int	k = argv[0];
	int	i;

	if (argc == 0 || k == 0)
		memset(&cpu.keybuf, 0, sizeof cpu.keybuf);
	else {
		if (k < 0xff) {
			do_intr(0, (long *) NULL);
			cpu.keybuf.rows[k >> 4] |= 1 << (k & 0xf);
			}
		for (i = 8; i >= 0; i--) 
			sprintf(acOut,"Row %d #%03x: %04x\n",
				i, 1 << i, cpu.keybuf.rows[i]); nx_print(acOut);
		}
}
/**********************************************************************/
/*   Fake an interrupt going off - just push the current PC.	      */
/**********************************************************************/
void
do_intr(argc, argv)
int	argc;
long	*argv;
{	extern int gen_intr;

	/***********************************************/
	/*   If   argument  specified  then  generate  */
	/*   interrupt at regular intervals	       */
	/***********************************************/
	if (argc) {
		gen_intr = argv[0];
		return;
		}

	if (cpu.ien) {
		push(cpu.pc);
		cpu.pc = 0xf;
		cpu.ien = FALSE;
		}
	else
		cpu.intr_pending = TRUE;

}
/**********************************************************************/
/*   Specify an instruction number to stop at.			      */
/**********************************************************************/
void
do_instr(argc, argv)
int	argc;
long	*argv;
{
	if (argc) {
		exec_flags |= EXEC_INSTR;
		stop_at_instr = argv[0];
		}
	else {
		exec_flags &= ~EXEC_INSTR;
		stop_at_instr = -1;
		}
}
/**********************************************************************/
/*   Set flag for showing BCD calculations.			      */
/**********************************************************************/
void
do_decimal(argc, argv)
int	argc;
long	*argv;
{
	show_dec = argv[0];
}
/**********************************************************************/
/*   io_flag   used  to  toggle  whether  we  are  stopping  on  I/O  */
/*   instructions.						      */
/**********************************************************************/
void
do_io(argc, argv)
int	argc;
long	*argv;
{
	io_flag = argv[0];
}
/**********************************************************************/
/*   Command  to  print out a grob using '*' and spaces. Just enough  */
/*   to  see  80  pixels  wide.  Used for testing that what is being  */
/*   created is correct.					      */
/**********************************************************************/
void
do_grob(argc, argv)
int	argc;
long	*argv;
{
	int	lim, len;
	int	nibbles;
	int	x, y;
	int	x1, y1, b;
	int	bytes;
	long	addr = argv[0];

	if (argc == 3) {
		x = argv[1];
		y = argv[2];
		}
	else {
		lim = read_nibbles(addr, 5);
		if (lim == 0x02B1E) {
			addr += 5;
			lim = read_nibbles(addr, 5);
			}
		y = read_nibbles(addr+5, 5);
		if (y > 250)
			y = 250;
		x = read_nibbles(addr+10, 5);
		if (x > 250)
			x = 250;
		addr += 15;
		}

	sprintf(acOut,"GROB %ld %ld:\n", x, y); nx_print(acOut);
	sprintf(acOut," 1 2 3 4 5 6 7 8 9 A B C D E F\n"); nx_print(acOut);
	len = lim - 15;
	bytes = (x + 7) >> 3;
	nibbles = bytes * 2;

	for (y1 = 0; y1 < y; y1++) {
		for (x1 = 0; x1 < nibbles; x1++) {
			b = read_nibble(addr + x1);
			sprintf(acOut,"%x", b); nx_print(acOut);
			}
		sprintf(acOut," ");
		for (x1 = 0; x1 < nibbles; x1++) {
			b = read_nibble(addr + x1);
			sprintf(acOut,"%s%s%s%s",
				b & 1 ? "X" : " ",
				b & 2 ? "X" : " ",
				b & 4 ? "X" : " ",
				b & 8 ? "X" : " "); nx_print(acOut);
			x11_draw_nibble(x1, y1, b);
			}
		addr += nibbles;
		if (x11_loaded) {
			sprintf(acOut,"\n"); nx_print(acOut);
		}
		else if (print("\n") < 0)
			break;
		}
	x11_flush();
}
/**********************************************************************/
/*   Update the LCD.						      */
/**********************************************************************/
void
do_lcd()
{
	draw_lcd();
	x11_flush();
}
/**********************************************************************/
/*   Print a grob in a particular section of the LCD.		      */
/**********************************************************************/
int
print_grob(addr, base_line)
long	addr;
int	base_line;
{
	int	lim;
	int	nibbles;
	int	x, y;
	int	x1, y1, b;
	int	bytes;

	lim = read_nibbles(addr, 5);
	if (lim == 0x02B1E) {
		addr += 5;
		lim = read_nibbles(addr, 5);
		}
	y = read_nibbles(addr+5, 5);
	if (y > 500)
		y = 500;
	x = read_nibbles(addr+10, 5);
	if (x > 500)
		x = 500;
	addr += 15;

	bytes = (x + 7) >> 3;
	nibbles = bytes * 2;

	for (y1 = 0; y1 < y; y1++) {
		for (x1 = 0; x1 < nibbles; x1++) {
			b = read_nibble(addr + x1);
			x11_draw_nibble(x1, base_line + y1, b);
			}
		addr += nibbles;
		}
	return y;
}
/**********************************************************************/
/*   Modify nibbles in memory. Format is: modify <addr> <nibbles>.    */
/**********************************************************************/
void
do_modify(argc, argv, strv)
int	argc;
long	*argv;
char	**strv;
{	char	*cp;
	int	n;
	long	addr = argv[0];

	if (argc != 2) {
		sprintf(acOut,"Usage: modify <addr> <nibbles>\n"); nx_print(acOut);
		return;
		}
	for (cp = strv[1]; *cp; cp++) {
		n = strchr(hex, *cp) - hex;
		if (n < 0 || n > 15)
			break;
		write_nibble(addr++, n);
		}
}
/**********************************************************************/
/*   Set a breakpoint.						      */
/**********************************************************************/
void
do_break(argc, argv)
int	argc;
long	*argv;
{	int	i;

	if (argc == 0) {
		for (i = 0; i < MAX_BREAKPOINTS; i++) {
			if (bkpt_tbl[i].flags == 0)
				continue;
			sprintf(acOut,"#%d ", i+1); nx_print(acOut);
			if (bkpt_tbl[i].flags & BP_RANGE) {
				sprintf(acOut,"range watchpoint at %05lx..%05lx\n",
					bkpt_tbl[i].addr, bkpt_tbl[i].end_addr); 
				nx_print(acOut);
			}
			else if (bkpt_tbl[i].flags & BP_EXEC) {
				sprintf(acOut,"breakpoint at %lx\n", bkpt_tbl[i].addr);
				nx_print(acOut);
			}
			else {
				sprintf(acOut,"watchpoint at %lx\n", bkpt_tbl[i].addr);
				nx_print(acOut);
			}
			}
		return;
		}
	for (i = 0; i < MAX_BREAKPOINTS; i++) {
		if (bkpt_tbl[i].addr == 0) {
			sprintf(acOut,"#%d break at %lx\n", i+1, argv[0]); nx_print(acOut);
			bkpt_tbl[i].flags = BP_EXEC;
			bkpt_tbl[i].addr = argv[0];
			num_bkpts++;
			return;
			}
		}
	sprintf(acOut,"Breakpoint table full\n"); nx_print(acOut);
}
/**********************************************************************/
/*   Set a watch break point -- trap rd/wr access.		      */
/**********************************************************************/
void
do_watch(argc, argv)
int	argc;
long	*argv;
{	int	i;

	if (argc == 0) {
		do_break(argc, argv);
		return;
		}
	for (i = 0; i < MAX_BREAKPOINTS; i++) {
		if (bkpt_tbl[i].addr == 0) {
			sprintf(acOut,"#%d watchpoint at %lx\n", i+1, argv[0]); nx_print(acOut);
			bkpt_tbl[i].flags = BP_READ | BP_WRITE;
			bkpt_tbl[i].addr = argv[0];
			if (argc == 2) {
				bkpt_tbl[i].flags = BP_RANGE;
				bkpt_tbl[i].end_addr = argv[1];
				}
			else
				bkpt_tbl[i].end_addr = 0;
			num_bkpts++;
			num_ranges++;
			return;
			}
		}
	sprintf(acOut,"Breakpoint table full\n"); nx_print(acOut);
}
/**********************************************************************/
/*   Delete a breakpoint.					      */
/**********************************************************************/
void
do_delete(argc, argv)
int	argc;
long	*argv;
{	int	i;

	if (argc != 1) {
		sprintf(acOut,"All breakpoints deleted\n"); nx_print(acOut);
		for (i = 0; i < MAX_BREAKPOINTS; i++) {
			bkpt_tbl[i].addr = 0;
			bkpt_tbl[i].flags = 0;
			}
		num_bkpts = 0;
		num_ranges = 0;
		return;
		}
	if (argv[0] < 1 || argv[0] > MAX_BREAKPOINTS) {
		sprintf(acOut,"Breakpoint out of range.\n"); nx_print(acOut);
		return; 
		}
	i = argv[0] - 1;

	if (bkpt_tbl[i].flags & BP_RANGE)
		num_ranges--;
	if (bkpt_tbl[i].flags)
		num_bkpts--;
	bkpt_tbl[i].addr = 0;
	bkpt_tbl[i].flags = 0;
}
void
do_new(argc, argv)
int	argc;
long	*argv;
{
	if (argv[0])
		exec_flags |= EXEC_NEW_MODE;
	else
		exec_flags &= ~EXEC_NEW_MODE;
}
/**********************************************************************/
/*   Finish execution of current subroutine. Stop when we return.     */
/**********************************************************************/
void
do_finish()
{
	bkpt_tbl[MAX_BREAKPOINTS].addr = cpu.rstk[cpu.rstk_ptr];
	bkpt_tbl[MAX_BREAKPOINTS].flags = BP_EXEC;
	exec_flags |= EXEC_RUN;
	num_bkpts++;
	cont_count = 1;
}
void
do_disass(argc, argv)
int	argc;
long	*argv;
{
	long	start = argv[0];
	long	end = argv[1];

	indent = 0;
	if (argc < 2)
		end = start + 0x80000L;
	dis_code(start, end);
}
/**********************************************************************/
/*   Disassemble RPL at address.				      */
/**********************************************************************/
void
do_rpl(argc, argv)
int	argc;
long	*argv;
{
	long	start = argv[0];
	long	end = argv[1];

	if (argc < 1)
		start = cpu.D0;
	indent = 0;
	if (argc < 2)
		end = start + 0x80000L;
	dis_program(start, end);
}
long
dis_code(start, end)
long	start;
long	end;
{
	long	prefix;
	while (start <= end && quit_flag == FALSE) {
		prefix = read_nibbles(start, 5);
		if (prefix == start + 5) {
			print("\n$%05X:PREFIX==> %05X %s\n", start, prefix, symbol(start));
			start += 5;
			}
		else if (prefix == DOCOL) {
			print("\n$%05X:PROGRAM OBJECT==> %05X %s\n", start, prefix, symbol(start));
			start = dis_program(start+5, end);
			}
		else if (prefix == DOEXT) {
			print("\n$%05X:UNIT OBJECT==> %05X %s\n", start, prefix, symbol(start));
			start += 5;
			}
		else if (prefix == DOCODE) {
			print("\n$%05X:CODE OBJECT==> %05X %s\n", start, prefix, symbol(start));
			start += 5;
			}
		else
			start = Disassemble(start, start+1);
		}
	return start;
}
/**********************************************************************/
/*   Disassemble  a  program  object  which is a sequence of objects  */
/*   and object pointers.					      */
/**********************************************************************/
long
dis_program(start, end)
long	start;
long	end;
{	long	prefix;
	int	ch;
	long	len;

	while (start <= end && quit_flag == FALSE) {
		prefix = read_nibbles(start, 5);
		print("$%05lX: %05lX %*s%s ", 
			start, prefix, indent * 2, " ", symbol(prefix));
		start += 5;
		switch (prefix) {
		  case DOBINT:
			prefix = read_nibbles(start, 5);
			start += 5;
		  	print("<%05lX>\n", prefix);
		  	break;
		  case DOCSTR:
			len = read_nibbles(start, 5) - 5;
			start += 5;
			print("\"");
			if (len > 32)
				len = 32;
			start = print_str(start, len);
			print("\"\n");
		  	break;
		  case DOEREAL:
		  	start = dis_real(start, 15, 5);
			print("\n");
		  	break;
		  case DOREAL:
		  	start = dis_real(start, 12, 3);
			print("\n");
		  	break;
		  case DOCOL:
		  	print("\n");
			indent++;
			start = dis_program(start, end);
			indent--;
			break;
		  case DOCODE:
			len = read_nibbles(start, 5) - 5;
			start += 5;
		  	print("len=%05lX\n", len);
			indent++;
			start = dis_code(start, start+len-1);
			indent--;
		  	break;
		  case DOIDNT:
		  case DOLAM:
		  	len = read_byte(start);
			print("'");
			start = print_str(start+2, len * 2);
			print("'\n");
			break;
		  case DOCHAR:
		  	ch = read_byte(start);
			start += 2;
			print(" %02x '%c'\n", ch, ch >= ' ' && ch <= 0x7e ? ch : '.');
			break;
		  case DOHSTR:
			len = read_nibbles(start, 5) - 5;
			start += 5;
			while (len-- > 0) {
				print("%c", hex[read_nibble(start++)]);
				}
		  	break;
		  default:
			print("\n");
			break;
		  }
		if (prefix == SEMI)
			break;
		}
	return start;
}
/**********************************************************************/
/*   Function to print out a byte string.			      */
/**********************************************************************/
long
print_str(start, len)
long	start;
long	len;
{	int	ch;

	while (len > 0) {
		len -= 2;
		ch = read_nibbles(start, 2);
		start += 2;
		if (ch < ' ' || ch >= 0x7f)
			print("\\x%02x", ch);
		else
			print("%c", ch);
		}
	return start;
}
/**********************************************************************/
/*   Disassemble a REAL constant.				      */
/**********************************************************************/
long
dis_real(start, mant_len, exp_len)
long	start;
int	mant_len;
int	exp_len;
{	long	exponent;
	unsigned char mantissa[64];
	int	i;
	int	sign;

	exponent = read_nibbles(start, exp_len);
	start += exp_len;
	for (i = 0; i < mant_len; i++)
		mantissa[i] = read_nibble(start++);
	sign = read_nibble(start++);

	if (sign == 0)
		print(" ");
	else
		print("-");
	print("%c.", hex[mantissa[0]]);
	for (i = 1; i < mant_len; i++)
		print("%c", hex[mantissa[i]]);
	if (exponent)
		print("E%lx", exponent);
	return start;
}
void
do_where()
{	int	i;
	long	addr;
	int	p = cpu.rstk_ptr;

	sprintf(acOut,"#0: %s\n", label(cpu.pc)); nx_print(acOut);
	sprintf(acOut,"    "); nx_print(acOut);
	Disassemble(cpu.pc, cpu.pc+1);
	for (i = 0; i < 8; i++) {
		addr = cpu.rstk[p];
		p = (p-1) & 7;
		sprintf(acOut,"#%d: %05lx  %s\n", i+1, addr, label(addr)); nx_print(acOut);
		}
}
void
do_go(argc, argv)
int	argc;
long	*argv;
{
	cpu.pc = argv[0];
	fprintf(cmd_fp, "instr %lx\nc\ngo %lx\n", cpu.instr_count, cpu.pc);
}
void
do_step(argc, argv)
int	argc;
long	*argv;
{
	exec_flags |= EXEC_STEP | EXEC_RUN;
	if (argc)
		cont_count = argv[0];
}
void
do_continue(argc, argv)
int	argc;
long	*argv;
{
	nx_print("Continuing...\nEnter a command to abort: ");
	exec_flags |= EXEC_RUN;
	if (argc)
		cont_count = argv[0];
}
void
do_trace(argc, argv)
int	argc;
long	*argv;
{
	exec_flags |= EXEC_TRACE | EXEC_RUN;
	if (argc)
		cont_count = argv[0];
}
void
do_reg(argc, argv, strv)
int	argc;
long	*argv;
char	**strv;
{
	if (argc != 2) {
		print_reg();
		Disassemble(cpu.pc, cpu.pc + 1, stdout);
		return;
		}
	if (strcmp(strv[0], "a") == 0)
		cpu.A.lo = argv[1];
	else if (strcmp(strv[0], "b") == 0)
		cpu.B.lo = argv[1];
	else if (strcmp(strv[0], "c") == 0)
		cpu.C.lo = argv[1];
	else if (strcmp(strv[0], "d") == 0)
		cpu.D.lo = argv[1];
	else if (strcmp(strv[0], "st") == 0)
		cpu.st = argv[1];
	
}
/**********************************************************************/
/*   Set pagelength for More prompt.				      */
/**********************************************************************/
void
do_pagelen(argc, argv)
int	argc;
long	*argv;
{
	pagelen = argv[0];
}
/**********************************************************************/
/*   Print out argument symbolically.				      */
/**********************************************************************/
void
do_print(argc, argv)
int	argc;
long	*argv;
{
	sprintf(acOut,"%08lx: %s\n", argv[0], label(argv[0])); nx_print(acOut);
}
void
do_quiet()
{
	step_quiet = !step_quiet;
	sprintf(acOut,"Single step %s\n", step_quiet ? "quietly" : "noisily"); nx_print(acOut);
}
void
dump_rom(argc, argv)
int	argc;
long	*argv;
{
	if (argc < 2)
		argv[1] = argv[0] + 320;
	print_mem(argv[0], argv[1], TRUE);
}
void
dump_ram(argc, argv)
int	argc;
long	*argv;
{
	if (argc < 2)
		argv[1] = argv[0] + 320;
	print_mem(argv[0], argv[1], FALSE);
}
/**********************************************************************/
/*   Routine to save a memory image so it can be reloaded.	      */
/**********************************************************************/
void
do_save(argc, argv, strv)
int	argc;
long	*argv;
char	**strv;
{	FILE	*fp;
	char	*filename;
	extern char *HPSavePath();

	if (argc == 0)
		filename = HPSavePath();
	else
		filename = strv[0];
	fp = fopen(filename, "w");
	if (fp == (FILE *) NULL) {
		perror(filename);
		return;
		}
	fwrite(&cpu, sizeof cpu, 1, fp);
	/***********************************************/
	/*   Save I/O address space.		       */
	/***********************************************/
#ifdef MY48
 	fwrite(cpu.ram, 0x100*2*sizeof(Nibble), 1, fp);
#else
  	fwrite(cpu.ram, 0x100, 1, fp);
#endif

	/***********************************************/
	/*   Save  internal  32K memory plus two 128K  */
	/*   card slots.			       */
	/***********************************************/
#ifdef MY48
 	fwrite(&cpu.ram[0x70000], (32 + 128 + 128) * 1024*sizeof(Nibble), 1, fp);
#else
  	fwrite(&cpu.ram[0x70000 >> 1], (32 + 128 + 128) * 1024, 1, fp);
#endif

	fclose(fp);
}
/**********************************************************************/
/*   Routine to reload a memory image.				      */
/**********************************************************************/
void
do_load(argc, argv, strv)
int	argc;
long	*argv;
char	**strv;
{	FILE	*fp;
	char	*filename;
	extern char *HPLoadPath();

	if (argc == 0)
		filename = HPLoadPath();
	else
		filename = strv[0];
	fp = fopen(filename, "r");
	if (fp == (FILE *) NULL) {
		perror(filename);
		return;
		}
	if (fread(&cpu, sizeof cpu, 1, fp) != 1) {
		fprintf(stderr, "Error reading CPU state info\n");
		exit(1);
		}
#ifdef MY48
 	if (fread(cpu.ram, 0x100*2*sizeof(Nibble), 1, fp) != 1) {
#else
  	if (fread(cpu.ram, 0x100, 1, fp) != 1) {
#endif
		fprintf(stderr, "Error reading RAM area\n");
		exit(1);
		}
#ifdef MY48
 	if (fread(&cpu.ram[0x70000], (32 + 128 + 128) * 1024*sizeof(Nibble), 1, fp) != 1) {
#else
  	if (fread(&cpu.ram[0x70000 >> 1], (32 + 128 + 128) * 1024, 1, fp) != 1) {
#endif
		fprintf(stderr, "Error reading RAM area\n");
		exit(1);
		}
	fclose(fp);
	sprintf(acOut,"Image loaded\n"); nx_print(acOut);
	do_lcd();
}
void
print_mem(addr, end, rom_flag)
long	addr;
long	end;
int	rom_flag;
{	int	i, j;

	while (addr <= end && quit_flag == FALSE) {
		sprintf(acOut,"%05lx: ", addr); nx_print(acOut);
		for (j = 0; j < 32; j++) {
			if ((j & 0xf) == 0)
				sprintf(acOut," "); nx_print(acOut);
			if (j == 0x8 || j == 0x18)
				sprintf(acOut,":"); nx_print(acOut);
			sprintf(acOut,"%x", 
				rom_flag ? (int) get_nibble((NAddr) addr+j)
					: read_nibble(addr+j)); nx_print(acOut);
			} 
		sprintf(acOut," "); nx_print(acOut);
		for (j = 0; j < 32; j += 2) {
			i = rom_flag ? (int) get_nibble((NAddr) addr+j)
					: read_nibble(addr+j);
			i |= (rom_flag ? (int) get_nibble((NAddr) addr+j+1)
					: read_nibble(addr+j+1)) << 4;
			if (i >= ' ' && i < 0x7e) 
				sprintf(acOut,"%c", i);
			else
				sprintf(acOut,"."); 
			nx_print(acOut);
			}
		sprintf(acOut," "); nx_print(acOut);
		for (j = 1; j < 33; j += 2) {
			i = rom_flag ? (int) get_nibble((NAddr) addr+j)
					: read_nibble(addr+j);
			i |= (rom_flag ? (int) get_nibble((NAddr) addr+j+1)
					: read_nibble(addr+j+1)) << 4;
			if (i >= ' ' && i < 0x7e) 
				sprintf(acOut,"%c", i); 
			else
				sprintf(acOut,"."); 
			nx_print(acOut);
			}
		addr += 32;
		print("\n");
		}
}
void
print_reg()
{	register int	i;
	char	*cp;
	int mode = cpu.debug_mode;

	cpu.debug_mode = TRUE;

	for (i = 0; i < 4; i++) {
		if ((cpu.reg[i].lo & ADDRESS_20) < 0x100)
			continue;
		cp = label(cpu.reg[i].lo & ADDRESS_20);
		if (*cp)
			print("%c=>%s\n", 'A' + i, cp);
		}
	sprintf(acOut," A:"); nx_print(acOut);
	print_64(cpu.A);
	sprintf(acOut,"  B:"); nx_print(acOut);
	print_64(cpu.B);
	sprintf(acOut,"  C:"); nx_print(acOut);
	print_64(cpu.C);
	sprintf(acOut,"  D:"); nx_print(acOut);
	print_64(cpu.D);
	print("\n");

	if (cpu.R0.hi || cpu.R0.lo ||
	    cpu.R1.hi || cpu.R1.lo ||
	    cpu.R2.hi || cpu.R2.lo ||
	    cpu.R3.hi || cpu.R3.lo) {
		sprintf(acOut,"R0:"); nx_print(acOut); 
		print_64(cpu.R0);
		sprintf(acOut," R1:"); nx_print(acOut);
		print_64(cpu.R1);
		sprintf(acOut," R2:"); nx_print(acOut);
		print_64(cpu.R2);
		sprintf(acOut," R3:"); nx_print(acOut);
		print_64(cpu.R3);
		print("\n");
		}
	sprintf(acOut,"  P:%x %c%c  ST:%04x %s%s%s%s OUT:%03x",
		cpu.p,
		cpu.cy ? 'C' : ' ',
		cpu.hexmode ? 'H' : 'D',
		cpu.st & 0xffff, 
		cpu.hst & MP ? "MP " : "",
		cpu.hst & SB ? "SB " : "",
		cpu.hst & SR ? "SR " : "",
		cpu.hst & XM ? "XM " : "",
		cpu.out_reg); nx_print(acOut);
	sprintf(acOut," INTR %s", cpu.ien ? "ENABLED" : "DISABLED"); nx_print(acOut);
	if (cpu.R4.hi || cpu.R4.lo) {
		sprintf(acOut," R4:"); nx_print(acOut);
		print_64(cpu.R4);
		}
	print("\n");

	sprintf(acOut,"D0:%05lx=>", cpu.D0); nx_print(acOut);
	for (i = 0; i < 5; i++) {
		sprintf(acOut,"%x%x%x%x%x ", 
			read_nibble(cpu.D0 + i*5),
			read_nibble(cpu.D0 + i*5 + 1),
			read_nibble(cpu.D0 + i*5 + 2),
			read_nibble(cpu.D0 + i*5 + 3),
			read_nibble(cpu.D0 + i*5 + 4));
		} nx_print(acOut);
	print("%s\n", label(cpu.D0));
	sprintf(acOut,"D1:%05lx=>", cpu.D1); nx_print(acOut);
	for (i = 0; i < 5; i++) {
		sprintf(acOut,"%x%x%x%x%x ", 
			read_nibble(cpu.D1 + i*5),
			read_nibble(cpu.D1 + i*5 + 1),
			read_nibble(cpu.D1 + i*5 + 2),
			read_nibble(cpu.D1 + i*5 + 3),
			read_nibble(cpu.D1 + i*5 + 4));
		} nx_print(acOut);
	print("%s\n", label(cpu.D1)); 

	/***********************************************/
	/*   Make sure registers dont contain	       */
	/*   extraneous bits			       */
	/***********************************************/
	if ((cpu.D0 & ~ADDRESS_20)) {
		sprintf(acOut,"*** D0 CORRUPTED!!!\n"); nx_print(acOut);
		exit(1);
		}
	if ((cpu.D1 & ~ADDRESS_20)) {
		sprintf(acOut,"*** D1 CORRUPTED!!!\n"); nx_print(acOut);
		exit(1);
		}
	sprintf(acOut,"RSTK:");
	for (i = 0; i < 8; i++) {
		sprintf(acOut,"%05lx ", cpu.rstk[(cpu.rstk_ptr - i) & 7]); nx_print(acOut);
		}
	print("\n");
	cpu.debug_mode = mode;
}
void
print_64(v)
WORD64	v;
{
	sprintf(acOut,"%08lx%08lx", v.hi, v.lo); nx_print(acOut);
}
/**********************************************************************/
/*   Function  which  returns a value from the string and skips over  */
/*   it.  nn  is  used  to  access  the  named register otherwise we  */
/*   accept a hex number.					      */
/**********************************************************************/
WORD64 *
get_value(strp)
char	**strp;
{	char	*str = *strp;
	struct map *mp;
	static WORD64	val;

	val.hi = val.lo = 0;
	if (*str == '$') {
		str++;
		for (mp = reg_tbl; mp->name; mp++)
			if (strncmp(mp->name, str, strlen(mp->name)) == 0) {
				if (mp->size == sizeof(WORD64))
					val = *(WORD64 *) mp->ptr;
				else if (mp->size == sizeof(WORD20)) {
					val.lo = *(WORD20 *) mp->ptr;
					}
				else {
					sprintf(acOut,"Help in get_value()\n");
					nx_print(acOut);
				}
				*strp = str + strlen(mp->name);
				return &val;
				}
		sprintf(acOut,"Unknown register name\n"); nx_print(acOut);
		return &val;
		}
	while (*str) {
		if (*str >= '0' && *str <= '9') {
			shift_left_nibbles(&val, 1);
			val.lo |= *str++ - '0';
			}
		else if (*str >= 'a' && *str <= 'f') {
			shift_left_nibbles(&val, 1);
			val.lo |= *str++ - 'a' + 10;
			}
		else
			break;
		}
	*strp = str;
	return &val;
}
long
decode_arg(str)
char	*str;
{static	long	n;
	long	n1;
	int	add;

	if (*str == '-' || *str == '+')
		goto rel_value;

	n = get_value(&str)->lo;
	if (*str == '\0')
		return n;
rel_value:
	switch (*str++) {
	  case '+':	add = 1; break;
	  case '-':	add = 2; break;
	  default:
	  	return n;
	  }
	n1 = get_value(&str)->lo;
	if (add == 1)
		n += n1;
	else
		n -= n1;
	return n;
}
/**********************************************************************/
/*   Check to see if a breakpoint was hit.			      */
/**********************************************************************/
int
check_breakpoint(type, addr)
int	type;
long	addr;
{
	struct breakpoint *bp;
	int	n;

	bp = bkpt_tbl;
	n = num_bkpts;
	for ( ; n > 0 ; bp++) {
		if (bp->flags == 0)
			continue;
		n--;
		if (bp->flags & BP_RANGE &&
		    addr >= bp->addr && addr <= bp->end_addr) {
		    	goto hit_it;
		    	}
		if (bp->flags & type && addr == bp->addr) {
hit_it:
			sprintf(acOut,"%s%05lx: *** BREAKPOINT (%s) HIT @%05lx%s\n", 
				blue, 
				cpu.pc,
				type == BP_READ ? "READ" :
				type == BP_WRITE ? "WRITE" :
					"EXEC",
				addr, green); nx_print(acOut);
			exec_flags |= EXEC_STOPPED;
			cont_count = 0;
			return TRUE;
			}
		}
	return FALSE;
}
/**********************************************************************/
/*   Function to print string and keep track of line number.	      */
/**********************************************************************/
# define	ARGS	a, b, c, d, e, f, g, h, i
int
print(str, ARGS)
char	*str;
long	ARGS;
{	char	buf[BUFSIZ];
	char	*cp;


		sprintf(acOut,str, ARGS); nx_print(acOut);
		return 0;
		
	sprintf(buf, str, ARGS);
	for (cp = buf; *cp; cp++) {
		putchar(*cp);
		if (*cp != '\n' || ++line_no < pagelen)
			continue;
		sprintf(acOut,"-- More --"); nx_print(acOut);
		gets(buf);
		if (*buf != '\n' && *buf != '\0') {
			quit_flag = TRUE;
			return -1;
			}
		line_no = 0;
		}
	return 0;

}
