/**********************************************************************/
/*   Routine to disassemble an instruction.			      */
/**********************************************************************/
# include	"hp48.h"

static char hex[] = "0123456789ABCDEF";

static char *opcode_0tbl[] = {
	"RETSETXM",
	"RET",
	"RETSETC",
	"RETCLRC",
	"SETHEX",
	"SETDEC",
	"PUSH.A C",
	"POP.A C",
	"CLR.X ST",
	"MOVE.X C,ST",
	"MOVE.X ST,C"
	"SWAP.X C,ST",
	"INC.1 P"
	"DEC.1 P",
	"logic",
	"RETI"
	};
static char	*opcode_13tbl[] = {
	"MOVE.A A,D0", 
	"MOVE.A A,D1", 
	"SWAP.A A,D0",
	"SWAP.A A,D1", 
	"MOVE.A C,D0", 
	"MOVE.A C,D1", 
	"SWAP.A C,D0", 
	"SWAP.A C,D1",
	"MOVE.4 A,D0", 
	"MOVE.4 A,D1", 
	"SWAP.4 A,D0", 
	"SWAP.4 A,D1", 
	"MOVE.4 C,D0",
	"MOVE.4 C,D1", 
	"SWAP.4 C,D0", 
	"SWAP.4 C,D1"
	};
static char *fs[] = {
	"P", "WP", "XS", "X", "S", "M", "B", "W",
	"?",  "?", "?", "?", "?", "?", "?", "A"
	};
static char	*opcode_14tbl[] = {
	"A,@D0", "A,@D1", "@D0,A", "@D1,A", "C,@D0", "C,@D1",
	"@D0,C", "@D1,C"};

static char reg1[] = {"BCACABCD"};

/**********************************************************************/
/*   Buffer for compiling the instruction.			      */
/**********************************************************************/
static char	ins_buf[64];
static	char	*insp;
static	char	*binp;
static	int	n1, n2, n3, n4;
static long	addr;

/**********************************************************************/
/*   Prototypes.						      */
/**********************************************************************/
int	get_nib PROTO((long));
void	handle_0 PROTO((void));
void	handle_1 PROTO((void));

char *
disassemble(addr_value, new_addr)
long	addr;
long	*new_addr;
{	int	len;

	sprintf(buf, "%05x:         ", addr);
	binp = buf + 7;
	insp = binp + 8;
	addr = addr_value;
	/***********************************************/
	/*   Look   at   first   nibble  to  try  and  */
	/*   classify instruction.		       */
	/***********************************************/
	n1 = get_nib(addr++);
  	n2 = get_nib(addr++);
	switch (n1) {
	  case 0: len = handle_0(); break;
	  case 1: len = handle_1(); break;
	  default:
	  	sprintf(insp, "Unimpl %x%x", n1, n2);
		len = 2;
		break;
	  }
	*new_addr = addr;
	return ins_buf;
}
void
handle_0()
{
	if (n2 != 0xE) {
		strcpy(insp, opcode_0tbl[n2]);
		return;
		}
	n3 = get_nib();
	n4 = get_nib();
	sprintf(insp, "%s=%s%c%s",
		fs[n3],
		fs[n3],
		n4 < 8 ? '&' : '|',
		reg1[n4 & 7]);
}
void
handle_1()
{	int	n;

	n3 = get_nib();
	switch ((n1 << 4) | n2) {
	  case 0x10:
	  	sprintf(insp, "MOVE.W %c.,R%d", 
			n3 < 8 ? 'A' : 'C',
			n3 & 7);
		break;
	  case 0x11:
	  	sprintf(insp, "MOVE.W R%d,%c", 
			n3 & 7,
			n3 < 8 ? 'A' : 'C');
	  	break;
	  case 0x12:
	  	sprintf(insp, "SWAP.W %c.,R%d", 
			n3 < 8 ? 'A' : 'C',
			n3 & 7);
		break;
	  case 0x13:
	  	strcpy(insp, opcode_13tbl[n3]);
		break;
	  case 0x14:
	  	sprintf(insp, "MOVE.%c ",
			n3 < 8 ? 'A' : 'B',
			opcode_14tbl[n3 & 7]);
		break;
	  case 0x15:
		n4 = get_nib();
	  	if (n4 < 8)
		  	sprintf(insp, "MOVE.%s %s", fs[n4],
		  		opcode_14tbl[n3 & 7]);
		else
		  	sprintf(insp, "MOVE.%d %s", n4+1,
		  		opcode_14tbl[n3 & 7]);
		break;
	  case 0x16:
	  	sprintf(insp, "ADD.A %d,D0", n3+1);
		break;
	  case 0x17:
	  	sprintf(insp, "ADD.A %d,D1", n3+1);
		break;
	  case 0x18:
	  	sprintf(insp, "SUB.A %d,D0", n3+1);
		break;
	  case 0x19:
		break;
	  }
	else {
		sprintf(insp, "Unimpl %03x", n);
		}
}
/**********************************************************************/
/*   Routine  to  get  the  next  nibble for disassembly and dump an  */
/*   ASCII version of the nibble in the compiled instruction.	      */
/**********************************************************************/
int
get_nib()
{	int	n;

	n = get_nibble(addr++);
	*binp++ = hex[n];
	return n;
}
