/**********************************************************************/
/*                                                                    */
/*  File:          draw.c                                             */
/*  Author:        P. D. Fox                                          */
/*  Created:       25 Nov 1991                     		      */
/*                                                                    */
/*  Copyright (c) 1990, 1991, 1992 Paul Fox                           */
/*                All Rights Reserved.                                */
/*                                                                    */
/*                                                                    */
/*--------------------------------------------------------------------*/
/*  Description:  Drawing Area widget.                                */
/*                                                                    */
/**********************************************************************/

/*static char sccs_id[] = "%Z% %M% %R%.%L%";*/

# include 	<X11/copyright.h>
# include 	<X11/IntrinsicP.h>
# include 	<X11/StringDefs.h>
# include 	<X11/Shell.h>
# include	<X11/Xatom.h>
# include	<signal.h>
# include 	"drawP.h"

/**********************************************************************/
/*   Following  is  used  if  we  want  to  use my version of malloc  */
/*   checking.							      */
/**********************************************************************/
# if 0
# 	include	<chkalloc.h>
# else
#	define	chk_alloc	malloc
#	define	chk_free	free
# endif

# include	<stdio.h>
# include	<ctype.h>
# include	<memory.h>
# include	<stdlib.h>
# include	<string.h>

# if !defined(staticf)
#	define	staticf	static
# endif

static XtResource resources[] = {
#define offset(field) XtOffset(DrawWidget, draw.field)
	/* {name, class, type, size, offset, default_type, default_addr}, */
	{ XtNresizeCallback, XtCResizeCallback, XtRCallback, sizeof(caddr_t),
	  offset(resize_callback), XtRCallback, (caddr_t) NULL},
	{ XtNexposeCallback, XtCExposeCallback, XtRCallback, sizeof(caddr_t),
	  offset(expose_callback), XtRCallback, (caddr_t) NULL},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof(unsigned long),
	  offset(foreground), XtRString, "white"},
#undef offset
};

staticf void DrawInput();

static XtActionsRec actions[] = {
  /* {name, procedure}, */
    {"input",		DrawInput},
    };
static char translations[] =
"\
<KeyPress>:		input()\n\
";

staticf void	realize();
staticf void	Resize();
staticf void	redisplay();

DrawClassRec drawClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Draw",
    /* widget_size		*/	sizeof(DrawRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	NULL,
    /* initialize_hook		*/	NULL,
    /* realize			*/	realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	Resize,
    /* expose			*/	redisplay,
    /* set_values		*/	NULL,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	XtInheritAcceptFocus,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* draw fields */
    /* empty			*/	0
  }
};

WidgetClass drawWidgetClass = (WidgetClass)&drawClassRec;

# define	superclass	(&widgetClassRec)
staticf void
realize(w, valueMask, attributes)
Widget	w;
XtValueMask	*valueMask;
XSetWindowAttributes	*attributes;
{
	Arg	args[10];
	int	n;
	XSetWindowAttributes	win_attr;

	(*superclass->core_class.realize)(w, valueMask, attributes);

	win_attr.bit_gravity = NorthWestGravity;
	XChangeWindowAttributes(XtDisplay(w), XtWindow(w),
		CWBitGravity, &win_attr);
	
	n = 0;
	XtSetArg(args[n], XtNinput, 1); n++;
	XtSetValues(w, args, n);

}
/* ARGSUSED */
staticf void
Resize(w)
DrawWidget	w;
{

	if (!XtIsRealized(w))
		return;
	XtCallCallbacks(w, XtNresizeCallback, (caddr_t) NULL);
}
/**********************************************************************/
/*   Action routine to handle redrawing of window.		      */
/**********************************************************************/
staticf void 
DrawExpose(w, event, x, y)
Widget	w;
XEvent	*event;
String	*x;
Cardinal	*y;
{

	XtCallCallbacks(w, XtNexposeCallback, (caddr_t) event);
}
/**********************************************************************/
/*   Come here on a keypress event.				      */
/**********************************************************************/
staticf void 
DrawInput(w, event, x, y)
Widget	w;
XEvent	*event;
String	*x;
Cardinal	*y;
{
}
staticf void
redisplay(w, event)
Widget	w;
XExposeEvent *event;
{
/*	exposed_region(w, event->x, event->y, event->width, event->height);*/
	DrawExpose(w, event, (String *) NULL, (Cardinal *) NULL);
}

