/**********************************************************************/
/*   Program  to  emulate the Saturn CPU and provide an emulation of  */
/*   the HP48 on other machines.				      */
/**********************************************************************/
#ifdef MY48
#define FILE_EM48
#endif

#define MAXKEYROW 10
int keyeventlist[MAXKEYROW];

# include "hp48.h"
# include "dis48.h"
# include "proto.h"
# include <string.h>
# include <fcntl.h>
#import <mach/cthreads.h>
#import <mach/mach.h>
#import <mach/message.h>

msg_header_t    header;

mutex_t shutdownMutex;
mutex_t shutdownWaitMutex;
mutex_t runningMutex;
mutex_t timerMutex;
condition_t timerCondition;
condition_t shutdownCondition;
condition_t shutdownWaitCondition;
any_t timerThread(any_t arg);
int keydirty=0;
int SHUTDOWNED = 0;

# if 1
# define CHECK_HEX() if (cpu.hexmode == FALSE) printf("%05lx: In decimal\n", cpu.pc);
# endif
/**********************************************************************/
/*   Color sequences for fcterm.				      */
/**********************************************************************/
# if !defined(MSDOS) && !defined(__MSDOS__) 
char	*yellow = "";
char	*green = "";
char	*red = "";
char	*blue = "";
# else
char	*yellow = "";
char	*green = "";
char	*red = "";
char	*blue = "";
# endif

struct stats stats;
int	exec_flags = EXEC_TRACE | EXEC_TRACE;
extern int tracing;
int	step_flag = STEP;
extern int num_bkpts;
extern long nibble_masks[];
extern int	io_flag;
extern int step_quiet;
extern int cont_count;
extern int line_no;
extern long stop_at_instr;
extern enum run_state last_mode;

int	gen_intr;
int	intr_instr;
/**********************************************************************/
/*   Some useful 64 bit constants -- initialised at start up.	      */
/**********************************************************************/
WORD64	constant_0;
WORD64	constant_1;

/**********************************************************************/
/*   If  an  instruction  isnt  decoded  remmeber  the  PC so we can  */
/*   print it out.						      */
/**********************************************************************/
long	old_pc;
int	n1, n2, n3, n4, n5, n6, n7;	/* Nibbles constituting current instr*/
/**********************************************************************/
/*   Table  used  to  detect opcodes we've not so far single stepped  */
/*   for debugging.						      */
/**********************************************************************/
char	new_instr[0x80000];

/**********************************************************************/
/*   Structure  used  to  construct  tables  of src and dst register  */
/*   values for instruction.					      */
/**********************************************************************/
struct pairs {
	WORD64	*src;
	WORD64	*dst;
	};
/**********************************************************************/
/*   Destination  registers  for  certain opcodes when combined with  */
/*   registers A, B, C, D (in that order).			      */
/**********************************************************************/
int	dest_reg[] = {regB, regC, regA, regC};

CPU	cpu;
FILE	*input_fp;
FILE	*cmd_fp;
/**********************************************************************/
/*   Prototypes							      */
/**********************************************************************/
void	usage PROTO((void));
int	do_switches PROTO((int, char **));
int	handle_0 PROTO((void));
int	handle_1 PROTO((void));
int	handle_0e PROTO((void));
int	handle_8 PROTO((void));
int	handle_80 PROTO((void));
int	handle_808 PROTO((void));
int	handle_81 PROTO((void));
int	handle_9 PROTO((void));
int	handle_a PROTO((void));
int	handle_b PROTO((void));
int	handle_c PROTO((void));
int	handle_d PROTO((void));
int	handle_e PROTO((void));
int	handle_f PROTO((void));
int	handle_13 PROTO((void));
int	handle_14 PROTO((void));
int	handle_15 PROTO((void));
void	do_jump45 PROTO((int));
void	do_jump67 PROTO((int));
void	execute_loop PROTO((void));
int	step_instruction PROTO((void));
extern long clock();

int
oldmain(argc, argv)
int	argc;
char	**argv;
{	int	arg_index;

	
	x11_init(&argc, argv);
	/***********************************************/
	/*   Warm start address.		       */
	/***********************************************/
/*	cpu.pc = 0x01fbd;*/
	cpu.pc = 0x00000;
	shutdownMutex=mutex_alloc();
	shutdownWaitMutex=mutex_alloc();
	timerMutex=mutex_alloc();
	runningMutex=mutex_alloc();
	shutdownCondition=condition_alloc();
	timerCondition=condition_alloc();
	shutdownWaitCondition=condition_alloc();
	constant_1.hi = 0;
	constant_1.lo = 1;

	arg_index = do_switches(argc, argv);
	cmd_fp = fopen("EM48.CMD", "w");

	input_fp = stdin;
	if (arg_index < argc) {
		/***********************************************/
		/*   Read  commands  from  em48.ini  file  to  */
		/*   avoid having to keep typing them in.      */
		/***********************************************/
		input_fp = fopen(argv[arg_index], "r");
		if (input_fp == NULL) {
			perror(argv[arg_index]);
			input_fp = stdin;
			}
		arg_index++;
		}

	read_rom();
	do_load(0,NULL,NULL);					/* hack - tjb */

	read_symtab("addr.list");

	if (arg_index < argc)
		sscanf(argv[arg_index++], "%x", &cpu.pc);
    	header.msg_size = sizeof header;
    	header.msg_local_port = (0);
	clock();

	/*cthread_detach(cthread_fork((cthread_fn_t)timerThread, (any_t)0));*/
	
	execute_loop();

	return 0;
}
void em_wait()
{
	msg_receive(&header, RCV_TIMEOUT, 50000);
}

/**********************************************************************/
/*   Main  execution  loop.  Execute  an  instruction  at a time. We  */
/*   avoid  doing  lots  of  emulator tests in the main loop by only  */
/*   checking if the exec_flags are non-zero.			      */
/**********************************************************************/
void
execute_loop()
{
	long	l;

	exec_flags = EXEC_STOPPED;
	Disassemble(cpu.pc, cpu.pc + 1);

	while (1) {
		/***********************************************/
		/*   Check for special debugger events.	       */
		/***********************************************/
		if (exec_flags) {
			/***********************************************/
			/*   See   if   user   wants   to  watch  the  */
			/*   instructions as they pass by.	       */
			/***********************************************/
			if ((exec_flags & (EXEC_TRACE | EXEC_STOPPED)) == EXEC_TRACE) {
				if (!step_quiet)
					print_reg();
				Disassemble(cpu.pc, cpu.pc + 1);
				exec_flags |= EXEC_STOPPED;
				}
			/***********************************************/
			/*   Check   reasons   for   going   back  to  */
			/*   debugger.				       */
			/***********************************************/
			if (exec_flags & EXEC_STOPPED) {
				exec_flags &= ~EXEC_STOPPED;
				get_command();
				}
			else if (exec_flags & EXEC_BKPT && check_breakpoint(BP_EXEC, cpu.pc)) {
				Disassemble(cpu.pc, cpu.pc + 1);
				get_command();
				}
			/***********************************************/
			/*   See   if   we've   reached  the  desired  */
			/*   instruction number.		       */
			/***********************************************/
			else if (exec_flags & EXEC_INSTR && stop_at_instr && 
			    cpu.instr_count >= stop_at_instr) {
			    	exec_flags &= ~EXEC_INSTR;
				stop_at_instr = 0;
				printf("*** INSTRUCTION COUNT REACHED\n");
				get_command();
				}
			else if (exec_flags & EXEC_INTR && intr_instr++ >= gen_intr) {
				/***********************************************/
				/*   Generate  a  pretend interrupt. The user  */
				/*   may  have  planted  a  breakpoint at the  */
				/*   interrupt address so need to check that.  */
				/***********************************************/
				intr_instr = 0;
				do_intr(0, (long *) NULL);
				if (exec_flags & EXEC_BKPT && check_breakpoint(BP_EXEC, cpu.pc)) {
					Disassemble(cpu.pc, cpu.pc + 1);
					get_command();
					}
				}
			else if (exec_flags & EXEC_SIGIO) {
				exec_flags &= ~EXEC_SIGIO;
				x11_refresh_window();
				}
			/***********************************************/
			/*   If  we've  not  executed  an instruction  */
			/*   at  this  address  before see if we need  */
			/*   to tell the user.			       */
			/***********************************************/
			if (new_instr[cpu.pc] == 0) {
				if (exec_flags & EXEC_NEW_MODE) {
					print_reg();
					Disassemble(cpu.pc, cpu.pc+ 1);
					exec_flags |= EXEC_STOPPED;
					}
				}
			new_instr[cpu.pc] = 1;
			}
		if ((++cpu.instr_count & 0x7f) == 0) {
			cpu.debug_mode = TRUE;
			n1 = read_nibble(TIMER1_CTRL);
			n2 = read_nibble(TIMER2_CTRL);
			cpu.debug_mode = FALSE;
			/***********************************************/
			/*   If  TIMER2  is running, then handle T1 &  */
			/*   T2 timer ticks.			       */
			/***********************************************/
			if (n2 & 0x01) {
				long l = cpu.timer2;
				int	gen_intr = TRUE;
				if (--cpu.timer2 < 0 && n2 & 0x02) {
					if (cpu.ien)
						do_intr(0, (long *) NULL);
					gen_intr = FALSE;
					}
				/***********************************************/
				/*   TIMER1 is ticked by bit 8 of TIMER2.      */
				/***********************************************/
				if ((l ^ cpu.timer2) & 0x100) {
					cpu.timer1 = (cpu.timer1 - 1) & 0x0f;
					if (cpu.timer1 == 0 && n1 & 0x02 && gen_intr)
						do_intr(0, (long *) NULL);
					}
				}
		
			if ((cpu.instr_count & 0x7fff) == 0) {
				/*x11_checkkeyinput();*/
				x11_refresh_window();
				do_lcd();
				l = clock() / 1000000;
				if (l)
					printf("Instr#%lx:%04lx  Instr/sec:%ld\n", 
					cpu.instr_count >> 16,
					cpu.instr_count & 0xffff,
					cpu.instr_count / l);
				}
			}
	
		/***********************************************/
		/*   Bash   out   one   instruction.  We  may  */
		/*   execute  an  illegal  instruction or hit  */
		/*   a  read/write  breakpoint  so we need to  */
		/*   go back to the debugger if we do.	       */
		/***********************************************/


		if (step_instruction()) {
			Disassemble(old_pc, old_pc+1);
			exec_flags |= EXEC_STOPPED;
			}
		}

}
#ifndef MY48
int
fast_nibble(addr)
long	addr;
{
	register int val;
	if (addr & 1)
		val = (cpu.ram[addr >> 1] >> 4) & 0x0f;
	else
		val = cpu.ram[addr >> 1] & 0x0f;

	return val;
}
#endif
void dotimerintr ()
{
	mutex_lock(timerMutex);
	do_intr(0, (long *) NULL);
	mutex_unlock(timerMutex);
}

		/***********************************************/
		/*   Generate  a  TIMER2  interrupt  if  tick  */
		/*   goes   to   zero.  Also  update  display  */
		/*   every  now  and  again.  Use the current  */
		/*   instruction number to scale the clock.    */
		/***********************************************/
any_t timerThread(any_t arg)
{

	while (1) {

			
			//fprintf(stderr,"T: TIMER_RUN=%d   now in wait mode\n", TIMER_RUN);
			//condition_wait(timerCondition, timerMutex);
			//fprintf(stderr,"T: TIMER_RUN=%d   now active\n", TIMER_RUN);	
			cpu.debug_mode = TRUE;
			n1 = read_nibble(TIMER1_CTRL);
			n2 = read_nibble(TIMER2_CTRL);
			cpu.debug_mode = FALSE;
			
			/***********************************************/
			/*   If  TIMER2  is running, then handle T1 &  */
			/*   T2 timer ticks.			       */
			/***********************************************/
			if (n2 & 0x01) {
				long l ;
				int	gen_intr = TRUE;
				
				l= cpu.timer2;
				if (--cpu.timer2 < 0 && n2 & 0x01) {
					if (cpu.ien) {
fprintf(stderr,"T: TIMER_RUN  INTR 2\n");
						mutex_lock(timerMutex);
						do_intr(0, (long *) NULL);
						mutex_unlock(timerMutex);
						condition_signal(shutdownCondition);
fprintf(stderr,"T: TIMER_RUN  INTR DONE\n");
					}
					gen_intr = FALSE;
					}
				/***********************************************/
				/*   TIMER1 is ticked by bit 8 of TIMER2.      */
				/***********************************************/
				if ((l ^ cpu.timer2) & 0x100) {
					cpu.timer1 = (cpu.timer1 - 1) & 0x0f;
					if (cpu.timer1 == 0 && n1 & 0x01 && gen_intr) {
fprintf(stderr,"T: TIMER_RUN  INTR 1\n");
						mutex_lock(timerMutex);
						do_intr(0, (long *) NULL);
						mutex_unlock(timerMutex);
						condition_signal(shutdownCondition);
fprintf(stderr,"T: TIMER_RUN  DONE\n");
						}
					}
			
				}
			 if (SHUTDOWNED&&(n2==5)&&!n1) {
					mutex_lock(timerMutex);
					condition_wait(shutdownCondition, timerMutex);
					mutex_unlock(timerMutex);
				}
			msg_receive(&header, RCV_TIMEOUT, 1000);
			//fprintf(stderr,"T: TIMER_RUN=%d   one run complete\n", TIMER_RUN);
	}
}
/**********************************************************************/
/*   Single step one instruction.				      */
/**********************************************************************/
int
step_instruction()
{	long	n;
	short	s;
	WORD64	*wp;

#ifdef ALWAYSLOCK 
	mutex_lock(timerMutex);
#endif
	/***********************************************/
	/*   Get  first  two  nibbles  --  we  always  */
	/*   need these.			       */
	/***********************************************/
#ifdef MY48
 	n1 = get_nibble(cpu.pc);
 	n2 = get_nibble(cpu.pc+1);
#else
	if (cpu.pc & 1) {
		n1 = cpu.rom[cpu.pc >> 1] >> 4;
		n2 = cpu.rom[(cpu.pc + 1) >> 1] & 0x0f;
		}
	else {
		n1 = cpu.rom[cpu.pc >> 1] & 0x0f;
		n2 = cpu.rom[(cpu.pc + 1) >> 1] >> 4;
		}
#endif
/*	n1 = get_nibble(cpu.pc);
	n2 = get_nibble(cpu.pc+1);*/

	switch (n1) {
	  case 0:
	  	if (handle_0() < 0)
			goto DEFAULT;
		break;
	  case 1:
	  	if (handle_1() < 0)
			goto DEFAULT;
		break;
	  case 2:
	  	/***********************************************/
	  	/*   MOVE.1 #$n,P			       */
	  	/***********************************************/
	  	cpu.p = n2;
		cpu.pc += 2;
		break;
	  case 3:
	  	/***********************************************/
	  	/*   MOVE.Pn #$nnnn, C			       */
	  	/***********************************************/
# define	WRAP_CHECK(n) \
		if (n + cpu.p >= 16) {\
			printf("%05lx: WRAP-AROUND!!!!!!!!!!!!\n", cpu.pc); \
			Disassemble(cpu.pc, cpu.pc+1); \
			}
	  	wp = get_imm_nibble(cpu.pc+2, n2 + 1);
		rotate_right_nibbles(&cpu.C, cpu.p);
		clear_field(&cpu.C, n2+1);
		cpu.C.lo |= wp->lo;
		cpu.C.hi |= wp->hi;
		rotate_left_nibbles(&cpu.C, cpu.p);
		cpu.pc += n2 + 3;
	  	break;
	  case 4:
		n3 = get_nibble(cpu.pc+2);
		if (n2 == 2 && n3 == 0) {
			cpu.pc += 3;
			break;
			}
	  	if (cpu.cy == 0) {
			cpu.pc += 3;
			break;
			}
		if (n3 == 0 && n2 == 0) {
			/***********************************************/
			/*   RETNC				       */
			/***********************************************/
			cpu.pc = pop();
			break;
			}
		/***********************************************/
		/*   BRNC label				       */
		/***********************************************/
		s = (n3 << 4) | n2;
		if (n3 & 0x8)
			s |= 0xff00;
	  	cpu.pc += s + 1;
	  	break;
	  case 5:
		n3 = get_nibble(cpu.pc+2);
	  	if (cpu.cy) {
			cpu.pc += 3;
			break;
			}
		if (n3 == 0 && n2 == 0) {
			/***********************************************/
			/*   RETCC				       */
			/***********************************************/
			cpu.pc = pop();
			break;
			}
		/***********************************************/
		/*   BRCC label				       */
		/***********************************************/
		s = (n3 << 4) | n2;
		if (n3 & 0x8)
			s |= 0xff00;
	  	cpu.pc += s + 1;
	  	break;
	  case 6:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
	  	/***********************************************/
	  	/*   JUMP.3 cpu.pc+$nnn			       */
	  	/***********************************************/
		s = (n4 << 8) | (n3 << 4) | n2;
		if (s == 0x300) {
			cpu.pc += 4;
			break;
			}
		if (s == 0x400) {
			cpu.pc += 5;
			break;
			}
		if (n4 & 8)
			s |= 0xf000;
	  	cpu.pc += s + 1;
		break;
	  case 7:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
	  	/***********************************************/
	  	/*   CALL.3 cpu.pc+$nnn			       */
	  	/***********************************************/
		push(cpu.pc + 4);
		n = ((n4 << 8) | (n3 << 4) | n2);
		if (n4 & 0x8)
			n |= 0xfffff000;
	  	cpu.pc += n + 4;
		break;
	  case 8:
	  	if (handle_8() < 0)
			goto DEFAULT;
		break;
	  case 9:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	if (handle_9() < 0)
			goto DEFAULT;
		break;
	  case 0xa:
		n3 = get_nibble(cpu.pc+2);
	  	if (handle_a() < 0)
			goto DEFAULT;
	  	break;
	  case 0xb:
		n3 = get_nibble(cpu.pc+2);
	  	if (handle_b() < 0)
			goto DEFAULT;
		break;
	  case 0xc:
	  	if (handle_c() < 0)
			goto DEFAULT;
	  	break;
	  case 0xd:
	  	if (handle_d() < 0)
			goto DEFAULT;
		break;
	  case 0xe:
	  	if (handle_e() < 0)
			goto DEFAULT;
		break;
	  case 0xf:
	  	if (handle_f() < 0)
			goto DEFAULT;
		break;
	  default:
	  DEFAULT:
	  	printf("%s??? Instruction not decoded ???%s\n", red, green);
		stats.illegal_instr++;
#ifdef ALWAYSLOCK 
		mutex_unlock(timerMutex);
#endif
		return -1;
	  }
#ifdef ALWAYSLOCK 
	mutex_unlock(timerMutex);
#endif
	return 0;
}
int
handle_0()
{	long	t;

	switch (n2) {
	  case 0x0:
	  	/***********************************************/
	  	/*   RETSETXM				       */
	  	/***********************************************/
		cpu.pc = pop();
		cpu.hst |= XM;
		break;
	  case 0x1:
	  	/***********************************************/
	  	/*   RET				       */
	  	/***********************************************/
		cpu.pc = pop();
		break;
	  case 0x2:
	  	/***********************************************/
	  	/*   RETSETC				       */
	  	/***********************************************/
		cpu.cy = 1;
		cpu.pc = pop();
		break;
	  case 0x3:
	  	/***********************************************/
	  	/*   RETCLRC				       */
	  	/***********************************************/
		cpu.cy = 0;
		cpu.pc = pop();
		break;
	  case 0x4: 
	  	/***********************************************/
	  	/*   SETHEX				       */
	  	/***********************************************/
		if (cpu.hexmode == FALSE && io_flag & IO_MODE) {
		  	printf("$%05lx: SETHEX [instr=%lx]\n", cpu.pc, cpu.instr_count);
			}
	  	cpu.hexmode = TRUE; 
		cpu.pc += 2;
		break;
	  case 0x5: 
	  	/***********************************************/
	  	/*   SETDEC				       */
	  	/***********************************************/
	  	cpu.hexmode = FALSE;
		if (io_flag & IO_MODE) {
		  	printf("$%05lx: SETDEC [instr=%lx]\n", cpu.pc, cpu.instr_count);
			old_pc = cpu.pc;
			cpu.pc += 2;
			return -1;
			}
		cpu.pc += 2;
		break;
	  case 0x6:
		push(cpu.C.lo & ADDRESS_20);
		cpu.pc += 2;
		break;
	  case 0x7:
		/***********************************************/
		/*   POP.A C				       */
		/***********************************************/
		cpu.C.lo &= ~ADDRESS_20;
		cpu.C.lo |= pop();
		cpu.pc += 2;
		break;
	  case 0x8:
		cpu.st &= 0xf000;
		cpu.pc += 2;
		break;
	  case 0x9:
	  	cpu.C.lo = (cpu.C.lo & ~EXP_MASK) | (cpu.st & EXP_MASK);
		cpu.pc += 2;
	  	break;
	  case 0xa:
	  	cpu.st = (cpu.st & ~EXP_MASK) | (cpu.C.lo & EXP_MASK);
		cpu.pc += 2;
		break;
	  case 0xb:
	  	/***********************************************/
	  	/*   SWAP.X C, ST			       */
	  	/***********************************************/
		t = cpu.C.lo & EXP_MASK;
	  	cpu.C.lo = (cpu.C.lo & ~EXP_MASK) | (cpu.st & EXP_MASK);
		cpu.st = (cpu.st & ~EXP_MASK) | t;
		cpu.pc += 2;
		break;
	  case 0xc:
	  	/***********************************************/
	  	/*   INC.1 P				       */
	  	/***********************************************/
		if (cpu.p == 0xf)
			cpu.cy = 1;
		else
			cpu.cy = 0;
		cpu.p = (cpu.p + 1) & 0xf;
		cpu.pc += 2;
		break;
	  case 0xd:
	  	/***********************************************/
	  	/*   DEC.1 P				       */
	  	/***********************************************/
		if (cpu.p == 0x0)
			cpu.cy = 1;
		else
			cpu.cy = 0;
		cpu.p = (cpu.p - 1) & 0xf;
		cpu.pc += 2;
		break;
	  case 0xe:
		return handle_0e();
	  case 0xf:
	  	/***********************************************/
	  	/*   RETI				       */
	  	/***********************************************/
		if (cpu.intr_pending) {
			cpu.intr_pending = FALSE;
			cpu.pc = 0xf;
			cpu.ien = FALSE;
			}
		else {
			cpu.pc = pop();
			cpu.ien = TRUE;
			}
		break;
	  default: 
		old_pc = cpu.pc;
	  	return -1;
		}
	return 0;
}
int
handle_0e()
{	WORD64	src, dst;
	int	r1 = 0, r2 = 0;

	n3 = get_nibble(cpu.pc+2);
	n4 = get_nibble(cpu.pc+3);
	switch (n4 & 7) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  	r1 = n4 & 3;
		r2 = dest_reg[n4 & 3];
	  	break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	r1 = dest_reg[n4 & 3];
		r2 = n4 & 3;
	  	break;
	  }

	src = *extract_field(&cpu.reg[r1], n3);
	dst = *extract_field(&cpu.reg[r2], n3);

	if (n4 & 8) {
		src.lo |= dst.lo;
		src.hi |= dst.hi;
		}
	else {
		src.lo &= dst.lo;
		src.hi &= dst.hi;
		}
	/***********************************************/
	/*   Put the result back.		       */
	/***********************************************/
	replace_field(&cpu.reg[r1], n3, &src);
	cpu.pc += 4;
	return 0;
}
int
handle_1()
{	WORD64	tmp;

	n3 = get_nibble(cpu.pc+2);
	switch (n2) {
	  case 0x0:
		switch (n3) {
		  case 0:
		  case 1:
		  case 2:
		  case 3:
		  case 4:
			/***********************************************/
			/*   MOVE.W A,Rn			       */
			/***********************************************/
			cpu.Rreg[n3] = cpu.A;
			break;
		  case 8:
		  case 9:
		  case 0xa:
		  case 0xb:
		  case 0xc:
			/***********************************************/
			/*   MOVE.W C,Rn			       */
			/***********************************************/
			cpu.Rreg[n3 & 7] = cpu.C;
			break;
		  default:
			old_pc = cpu.pc;
		  	return -1;
		  }
		cpu.pc += 3;
		break;
	  case 0x1:
		switch (n3) {
		  case 0:
		  case 1:
		  case 2:
		  case 3:
		  case 4:
			/***********************************************/
			/*   MOVE.W Rn,A			       */
			/***********************************************/
			cpu.A = cpu.Rreg[n3];
			break;
		  case 8:
		  case 9:
		  case 0xa:
		  case 0xb:
		  case 0xc:
			/***********************************************/
			/*   MOVE.W Rn,C			       */
			/***********************************************/
			cpu.C = cpu.Rreg[n3 & 7];
			break;
		  default:
			old_pc = cpu.pc;
		  	return -1;
		  }
		cpu.pc += 3;
		break;
	  case 0x2:
		switch (n3) {
		  case 0:
		  case 1:
		  case 2:
		  case 3:
		  case 4:
		  	/***********************************************/
		  	/*   SWAP.W A, Rn			       */
		  	/***********************************************/
		  	tmp = cpu.A;
			cpu.A = cpu.Rreg[n3];
			cpu.Rreg[n3] = tmp;
			break;
		  case 8:
		  case 9:
		  case 0xa:
		  case 0xb:
		  case 0xc:
		  	/***********************************************/
		  	/*   SWAP.W C, Rn			       */
		  	/***********************************************/
		  	tmp = cpu.C;
			cpu.C = cpu.Rreg[n3 & 7];
			cpu.Rreg[n3 & 7] = tmp;
			break;
		  default: 
			old_pc = cpu.pc;
		  	return -1;
		  }
		cpu.pc += 3;
	  	break;
	  case 0x3:
		return handle_13();
	  case 0x4:
		return handle_14();
	  case 0x5:
		return handle_15();
	  case 0x6:
		cpu.D0 += n3 + 1;
handle_D0:
		cpu.pc += 3;
		if (cpu.D0 & ~ADDRESS_20) {
			cpu.cy = 1;
			cpu.D0 &= ADDRESS_20;
			}
		else
			cpu.cy = 0;
		break;
	  case 0x7:
		cpu.D1 += n3 + 1;
handle_D1:
		cpu.pc += 3;
		if (cpu.D1 & ~ADDRESS_20) {
			cpu.cy = 1;
			cpu.D1 &= ADDRESS_20;
			}
		else
			cpu.cy = 0;
		break;
	  case 0x8:
		cpu.D0 -= n3 + 1;
		goto handle_D0;
	  case 0x9:
		cpu.D0 &= ~0xffL;
		cpu.D0 |= (WORD20) get_imm_nibble(cpu.pc + 2, 2)->lo;
		cpu.pc += 4;
		break;
	  case 0xa:
		cpu.D0 &= ~0xffffL;
		cpu.D0 |= (WORD20) get_imm_nibble(cpu.pc + 2, 4)->lo;
		cpu.pc += 6;
		break;
	  case 0xb:
		cpu.D0 = (WORD20) (get_imm_nibble(cpu.pc + 2, 5)->lo & ADDRESS_20);
		cpu.pc += 7;
		break;
	  case 0xc:
		cpu.D1 -= n3 + 1;
		goto handle_D1;
	  case 0xd:
		cpu.D1 &= ~0xffL;
		cpu.D1 |= (WORD20) get_imm_nibble(cpu.pc + 2, 2)->lo;
		cpu.pc += 4;
		break;
	  case 0xe:
		cpu.D1 &= ~0xffffL;
		cpu.D1 |= (WORD20) get_imm_nibble(cpu.pc + 2, 4)->lo;
		cpu.pc += 6;
		break;
	  case 0xf:
		cpu.D1 = (WORD20) (get_imm_nibble(cpu.pc + 2, 5)->lo & ADDRESS_20);
		cpu.pc += 7;
		break;
	  default: 
		old_pc = cpu.pc;
	  	return -1;
	  }
	return 0;
}
/**********************************************************************/
/*   Handle 13x opcodes.					      */
/**********************************************************************/
int
handle_13()
{	long	t;

	switch (n3) {
	  case 0:	cpu.D0 = cpu.A.lo & ADDRESS_20;	break;
	  case 1:	cpu.D1 = cpu.A.lo & ADDRESS_20;	break;
	  case 2:	
	  		t = cpu.D0;
	  		cpu.D0 = cpu.A.lo & ADDRESS_20;
	  		cpu.A.lo = (cpu.A.lo & ~ADDRESS_20) | t;
			break;
	  case 3:	t = cpu.D1;
	  		cpu.D1 = cpu.A.lo & ADDRESS_20;
	  		cpu.A.lo = (cpu.A.lo & ~ADDRESS_20) | t;
			break;

	  case 4:	cpu.D0 = cpu.C.lo & ADDRESS_20;	break;
	  case 5:	cpu.D1 = cpu.C.lo & ADDRESS_20;	break;
	  case 6:	t = cpu.D0;
	  		cpu.D0 = cpu.C.lo & ADDRESS_20;
	  		cpu.C.lo = (cpu.C.lo & ~ADDRESS_20) | t;
			break;
	  case 7:	t = cpu.D1;
	  		cpu.D1 = cpu.C.lo & ADDRESS_20;
	  		cpu.C.lo = (cpu.C.lo & ~ADDRESS_20) | t;
			break;
	  case 8:
	  		/***********************************************/
	  		/*   MOVE.4 A, D0			       */
	  		/***********************************************/
			cpu.D0 = (cpu.D0 & 0xf0000L) | (cpu.A.lo & 0xffff);
			break;
	  case 9:
	  		/***********************************************/
	  		/*   MOVE.4 A, D1			       */
	  		/***********************************************/
			cpu.D1 = (cpu.D1 & 0xf0000L) | (cpu.A.lo & 0xffff);
			break;
	  case 0xc:
	  		/***********************************************/
	  		/*   MOVE.4 C, D0			       */
	  		/***********************************************/
			cpu.D0 = (cpu.D0 & 0xf0000L) | (cpu.C.lo & 0xffff);
			break;
	  case 0xD:
	  		/***********************************************/
	  		/*   MOVE.4 C, D1			       */
	  		/***********************************************/
			cpu.D1 = (cpu.D1 & 0xf0000L) | (cpu.C.lo & 0xffff);
			break;
	  default:	
			old_pc = cpu.pc;
	  		return -1;
	  }
	cpu.pc += 3;
	return 0;
}
/**********************************************************************/
/*   Handle the 14x opcodes.					      */
/**********************************************************************/
int
handle_14()
{	int	size;
	long	mask;

	if (n3 & 8) {
		size = 2;
		mask = 0xff;
		}
	else {
		size = 5;
		mask = ADDRESS_20;
		}
	switch (n3 & 7) {
	  case 0:
	  	write_nibbles(cpu.D0, cpu.A.lo, size);
		break;
	  case 1:
	  	write_nibbles(cpu.D1, cpu.A.lo, size);
		break;
	  case 2:
	  	cpu.A.lo &= ~mask;
		cpu.A.lo |= read_nibbles(cpu.D0, size);
	  	break;
	  case 3:
	  	cpu.A.lo &= ~mask;
		cpu.A.lo |= read_nibbles(cpu.D1, size);
	  	break;

	  case 4:
	  	write_nibbles(cpu.D0, cpu.C.lo, size);
		break;
	  case 5:
	  	write_nibbles(cpu.D1, cpu.C.lo, size);
		break;
	  case 6:
	  	cpu.C.lo &= ~mask;
		cpu.C.lo |= read_nibbles(cpu.D0, size);
	  	break;
	  case 7:
	  	cpu.C.lo &= ~mask;
		cpu.C.lo |= read_nibbles(cpu.D1, size);
	  	break;
	  default:	
		old_pc = cpu.pc;
	  	return -1;
	  }
	cpu.pc += 3;
	return 0;
}
/**********************************************************************/
/*   Handle 15xx ocpu.pcode.					      */
/**********************************************************************/
int
handle_15()
{	WORD64	*wp;
	WORD64	tmp;
	WORD20	dreg;

	n4 = get_nibble(cpu.pc+3);
  	dreg = n3 & 1 ? cpu.D1 : cpu.D0;
	switch (n3) {
	  case 0:
	  case 1:
	  	wp = extract_field(&cpu.A, n4);
		shift_right_nibbles(wp, cpu.field_offset);
		write_nibbles(dreg, wp->lo, cpu.field_size);
		if (cpu.field_size > 8)
			write_nibbles(dreg + 8, wp->hi, cpu.field_size - 8);
		break;
	  case 2:
	  case 3:
	  	/***********************************************/
	  	/*   MOVE.? @D0, A			       */
	  	/***********************************************/
		wp = extract_indirect_field(dreg, n4);
		shift_left_nibbles(wp, cpu.field_offset);
		replace_field(&cpu.A, n4, wp);
	  	break;

	  case 4:
	  case 5:
	  	wp = extract_field(&cpu.C, n4);
		shift_right_nibbles(wp, cpu.field_offset);
		write_nibbles(dreg, wp->lo, cpu.field_size);
		if (cpu.field_size > 8)
			write_nibbles(dreg + 8, wp->hi, cpu.field_size - 8);
		break;
	  case 6:
	  case 7:
	  	/***********************************************/
	  	/*   MOVE.? @D0, C			       */
	  	/***********************************************/
		wp = extract_indirect_field(dreg, n4);
		shift_left_nibbles(wp, cpu.field_offset);
		replace_field(&cpu.C, n4, wp);
	  	break;
	  case 0x8:
	  case 0x9:
	  case 0xC:
	  case 0xD:
	  	/***********************************************/
	  	/*   MOVE.n A/C,@dn			       */
	  	/***********************************************/
		if (n3 & 4)
		  	tmp = cpu.C;
		else
		  	tmp = cpu.A;
		if (n4 >= 8) {
			write_nibbles(dreg, tmp.lo, 8);
			write_nibbles(dreg+8, tmp.hi, n4 + 1 - 8);
			}
		else
			write_nibbles(dreg, tmp.lo, n4 + 1);
		break;
	  case 0xA:
	  case 0xB:
	  case 0xE:
	  case 0xF:
	  	/***********************************************/
	  	/*   MOVE.n @Dn, A/C			       */
	  	/***********************************************/
		if (n3 & 4)
			wp = &cpu.C;
		else
			wp = &cpu.A;
		if (n4 < 8) {
		  	tmp.lo = read_nibbles(dreg, n4 + 1);
			tmp.hi = 0;
			}
		else {
		  	tmp.lo = read_nibbles(dreg, 8);
		  	tmp.hi = read_nibbles(dreg + 8, n4 + 1 - 8);
			}
		clear_field(wp, n4 + 1);
		wp->hi |= tmp.hi;
		wp->lo |= tmp.lo;
	  	break;
	  default:
		old_pc = cpu.pc;
	  	return -1;
	  }
	cpu.pc += 4;
	return 0;
}
/**********************************************************************/
/*   Handle 8xxx opcodes.					      */
/**********************************************************************/
int
handle_8()
{
	long	t, t1, t2;


	switch (n2) {
	  case 0x0:
	  	return handle_80();

	  case 0x1:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	return handle_81();

	  case 0x2:
		n3 = get_nibble(cpu.pc+2);
		cpu.hst &= ~n3;
		cpu.pc += 3;
		break;
	
	  case 0x3:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	do_jump45((cpu.hst & n3) == 0);
	  	break;
	  case 0x4:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		cpu.pc += 3;
		cpu.st &= ~(1 << n3);
		break;
	  case 0x5:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		cpu.pc += 3;
		cpu.st |= (1 << n3);
		break;

	  case 0x6:
	  	/***********************************************/
	  	/*   BRBC #nn, st, addr			       */
	  	/***********************************************/
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		do_jump45((cpu.st & (1 << n3)) == 0);
		break;
	  case 0x7:
	  	/***********************************************/
	  	/*   BRBS #nn, st, addr			       */
	  	/***********************************************/
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		do_jump45(cpu.st & (1 << n3));
	  	break;

	  case 0x8:
	  	/***********************************************/
	  	/*   BRNE   p,nn,addr			       */
	  	/***********************************************/
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		do_jump45(cpu.p != n3);
	  	break;
	  case 0x9:
	  	/***********************************************/
	  	/*   BREQ   p,nn,addr			       */
	  	/***********************************************/
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
		do_jump45(cpu.p == n3);
	  	break;

	  case 0xa:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	t = cpu.reg[n3 & 3].lo & ADDRESS_20;
	  	t1 = cpu.reg[dest_reg[n3 & 3]].lo & ADDRESS_20;
/*		CHECK_HEX();*/
		t2 = t - t1;
	  	if (n3 <= 3) {
	  		/***********************************************/
	  		/*   BREQ.f s,d,PC+3+yy			       */
	  		/***********************************************/
			do_jump45(t2 == 0);
			break;
			}
		if (n3 <= 7) {
	  		/***********************************************/
	  		/*   BRNE.f s,d,PC+3+yy			       */
	  		/***********************************************/
			do_jump45(t2 != 0);
			break;
			}
		if (n3 <= 0xb) {
			do_jump45(t == 0);
			break;
			}
		do_jump45(t != 0);
		break;
	  case 0xb:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	t = cpu.reg[n3 & 3].lo & ADDRESS_20;
		t1 = cpu.reg[dest_reg[n3 & 3]].lo & ADDRESS_20;
/*		CHECK_HEX();*/
		t2 = t - t1;
	  	if (n3 <= 3) {
	  		/***********************************************/
	  		/*   BRGT.f s,d,PC+3+yy			       */
	  		/***********************************************/
			do_jump45(t2 > 0);
			break;
			}
		if (n3 <= 7) {
	  		/***********************************************/
	  		/*   BRLT.f s,d,PC+3+yy			       */
	  		/***********************************************/
			do_jump45(t2 < 0);
			break;
			}
		if (n3 <= 0xb) {
	  		/***********************************************/
	  		/*   BRGE.f s,d,PC+3+yy			       */
	  		/***********************************************/
			do_jump45(t2 >= 0);
			break;
			}
	  	/***********************************************/
	  	/*   BRLE.f s,d,PC+3+yy			       */
	  	/***********************************************/
		do_jump45(t2 <= 0);
		break;
	  case 0xc:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	/***********************************************/
	  	/*   JUMP.4 label			       */
	  	/***********************************************/
	  	n6 = get_nibble(cpu.pc+5);
		t = (unsigned short) 
			((n6 << 12) | (n5 << 8) | (n4 << 4) | n3);
		if (n6 & 0x8)
			t |= 0xffff0000L;
		cpu.pc = cpu.pc + t + 2;
		break;

	  case 0xd:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	/***********************************************/
	  	/*   JUMP.5 label			       */
	  	/***********************************************/
	  	n6 = get_nibble(cpu.pc+5);
	  	n7 = get_nibble(cpu.pc+6);
		cpu.pc = ((long) n7 << 16) | 
			(unsigned short) ((n6 << 12) | (n5 << 8) | (n4 << 4) | n3);
		break;

	  case 0xe:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	/***********************************************/
	  	/*   CALL.4 addr			       */
	  	/***********************************************/
		push(cpu.pc + 6);
	  	n6 = get_nibble(cpu.pc+5);
		t = (unsigned short) 
			((n6 << 12) | (n5 << 8) | (n4 << 4) | n3);
		if (n6 & 0x8)
			t |= 0xffff0000L;
		cpu.pc = cpu.pc + t + 6;
/*cpu.cy = 1;*/
		break;
	  case 0xf:
		n3 = get_nibble(cpu.pc+2);
		n4 = get_nibble(cpu.pc+3);
		n5 = get_nibble(cpu.pc+4);
	  	/***********************************************/
	  	/*   CALL.A addr			       */
	  	/***********************************************/
		push(cpu.pc + 7);
	  	n6 = get_nibble(cpu.pc+5);
	  	n7 = get_nibble(cpu.pc+6);
		cpu.pc = ((long) n7 << 16) | 
			(unsigned short) ((n6 << 12) | (n5 << 8) | (n4 << 4) | n3);
/*cpu.cy = 1;*/
		break;
	  default:
		old_pc = cpu.pc;
	  	return -1;
	  }
	return 0;
}
int
handle_80()
{	long	mask;
	int	shift;
	WORD64	tmp;
	long	t;
	int	i, j;


	n3 = get_nibble(cpu.pc+2);
	switch (n3) {
	  case 1:
		/***********************************************/
		/*   OUT.x C				       */
		/***********************************************/
		cpu.out_reg = cpu.C.lo & 0xfff;
		cpu.pc += 3;
		if (io_flag & IO_OUT) {
			old_pc = cpu.pc-3;
			return -1;
			}
		break;
	  case 2:
		/***********************************************/
		/*   IN.4 A				       */
		/***********************************************/
		cpu.pc += 3;
		cpu.A.lo &= ~0xffffL;
		j = 0;
		for (i = 0; i < 9; i++)
			if (cpu.out_reg & (1 << i)) {
				j |= cpu.keybuf.rows[i];
				keyeventlist[i]=0;
			}
		cpu.A.lo |= j;
		if (io_flag & IO_IN) {
			old_pc = cpu.pc-3;
			return -1;
			}
		break;
	  case 3:
		/***********************************************/
		/*   IN.4 C				       */
		/***********************************************/
		cpu.pc += 3;
		cpu.C.lo &= ~0xffffL;
		j = 0;
		for (i = 0; i < 9; i++)
			if (cpu.out_reg & (1 << i)) {
				j |= cpu.keybuf.rows[i];
				keyeventlist[i]=0;
			}
		cpu.C.lo |= j;
		if (io_flag & IO_IN) {
			old_pc = cpu.pc-3;
			return -1;
			}
		break;
	  case 4:
		/***********************************************/
		/*   UNCNFG				       */
		/***********************************************/
		cpu.pc += 3;
		if (io_flag & IO_CONFIG) {
			printf("%05lx: UNCNFG %05lx [instr=%lx]\n",
				cpu.pc, cpu.C.lo & ADDRESS_20,
				cpu.instr_count);
			}
		break;
	  case 5:
		/***********************************************/
		/*   CONFIG -- only handle 00100, 70000 and    */
		/*   F0000 in C register.		       */
		/***********************************************/
		cpu.pc += 3;
		for (i = 0; i < MAX_CONFIGS-1; i++)
			cpu.configs[i] = cpu.configs[i+1];
		switch (cpu.C.lo & ADDRESS_20) {
		  case 0x80000:
		  case 0xC0000:
		  case 0xD0000:
			break;
		  case 0x00100:
			break;
		  case 0x70000:
			cpu.ram_32k = 0x70000L;
			break;
		  case 0xF0000:
			cpu.ram_32k = 0xF0000L;
			break;
		  case 0xFE000:
			cpu.ram_32k = 0xFE000L;
			break;
		  case 0xFC000:
			cpu.ram_32k = 0xFC000L;
			break;
		  default:
			break;
			}
		cpu.configs[MAX_CONFIGS-1] = cpu.C.lo & ADDRESS_20;
		if (cpu.configs[MAX_CONFIGS-1] == 0x70000L &&
			cpu.configs[MAX_CONFIGS-2] == 0xFE000L)
			cpu.ram_32k = 0xFE000L;
		else if (cpu.configs[MAX_CONFIGS-1] == 0x70000L &&
			cpu.configs[MAX_CONFIGS-2] == 0xFC000L)
			cpu.ram_32k = 0xFC000L;
		if (io_flag & IO_CONFIG) {
			printf("%05lx: CONFIG %05lx [instr=%lx]\n",
				cpu.pc, cpu.C.lo & ADDRESS_20,
				cpu.instr_count);
			}
		break;
	  case 6:
		/***********************************************/
		/*   MOVE.A ID, C			       */
		/***********************************************/
		{static int daisy_state[] = {5, 0xf6, 7, 0xf8, 1, 0xf2, 0};
		cpu.pc += 3;
		cpu.C.lo = (cpu.C.lo & ~EXP_MASK) | daisy_state[cpu.daisy_state];
		if (io_flag & IO_CONFIG) {
			printf("%05lx: MOVE.A ID,C [instr=%lx]\n",
				cpu.pc, 
				cpu.instr_count);
			}
		if (cpu.daisy_state > 5) {
			if (io_flag & IO_OTHER) {
				old_pc = cpu.pc-3;
				return -1;
				}
			break;
			}
		cpu.daisy_state++;
		break;
		}
	  case 7:{
	  	static long last_instr;
		/***********************************************/
		/*   SHUTDN				       */
		/***********************************************/

		if (io_flag & IO_OTHER) {
			printf("%05lx: SHUTDN[instr=%lx]\n",
				cpu.pc, 
				cpu.instr_count);
			}
			
		cpu.pc += 3;
		/***********************************************/
		/*   CPU  is  going to sleep so now is a good  */
		/*   time   to  check  for  any  key  events.  */
		/*   Avoid  eating  up  all  the  CPU drawing  */
		/*   the display if we are in a SHUTDN loop.   */
		/***********************************************/
		SHUTDOWNED=1;
 		/*checkTimer();*/

		if ((cpu.pc!=0x019fe)&&(!keydirty)) {
			do_lcd();
			x11_refresh_window();
//	printf("%05lx: SHUTDN[instr=%lx]\n",cpu.pc, cpu.instr_count);

			condition_wait(shutdownCondition, timerMutex);
		}
		if (keydirty>0) keydirty = keydirty--;
		SHUTDOWNED=0;
		break;
		}
	  case 8:
		return handle_808();
	  case 9:
		cpu.pc += 3;
		tmp.lo = cpu.p + 1 + (cpu.C.lo & ADDRESS_20);
		if (tmp.lo & ~ADDRESS_20)
			cpu.cy = 1;
		else
			cpu.cy = 0;
		cpu.C.lo &= ~ADDRESS_20;
		cpu.C.lo |= (tmp.lo & ADDRESS_20);
		break;
	  case 0xa:
		/***********************************************/
		/*   RESET				       */
		/***********************************************/
		cpu.pc += 3;
		memset((char *)cpu.configs, 0, sizeof cpu.configs);
		cpu.daisy_state = 0;
		if (io_flag & IO_CONFIG) {
			printf("%05lx: RESET [instr=%lx]\n", cpu.pc, cpu.instr_count);
			}
		break;
	  case 0xc:
		/***********************************************/
		/*   MOVE.1 p,c,#$n			       */
		/***********************************************/
		n4 = get_nibble(cpu.pc+3);
		cpu.pc += 4;
		if (n4 >= 8) {
			shift = 4 * (n4 - 8);
			mask = 0xfL << shift;
			cpu.C.hi &= ~mask;
			cpu.C.hi |= (long) cpu.p << shift;
			}
		else {
			shift = 4 * n4;
			mask = 0xfL << shift;
			cpu.C.lo &= ~mask;
			cpu.C.lo |= (long) cpu.p << shift;
			}
		break;
	  case 0xd:
		/***********************************************/
		/*   MOVE.1 c,#$n,p			       */
		/***********************************************/
		n4 = get_nibble(cpu.pc+3);
		cpu.pc += 4;
		tmp = cpu.C;
		shift_right_nibbles(&tmp, n4);
		cpu.p = (WORD4) (tmp.lo & 0xf);
		break;
	  case 0xf:
		/***********************************************/
		/*   SWAP.1 P,C,x			       */
		/***********************************************/
		n4 = get_nibble(cpu.pc+3);
		cpu.pc += 4;
		tmp = cpu.C;
		t = cpu.p;
		shift_right_nibbles(&tmp, n4);
		cpu.p = (WORD4) (tmp.lo & 0xf);
		tmp.lo = t;
		shift_left_nibbles(&tmp, n4);
		if (n4 < 8) {
			cpu.C.lo &= ~nibble_masks[n4];
			cpu.C.lo |= tmp.lo;
			}
		else {
			cpu.C.hi &= ~nibble_masks[n4 & 7];
			cpu.C.hi |= tmp.hi;
			}
		break;
	  default: 
		old_pc = cpu.pc;
	  	return -1;
	  }
	return 0;
}
/**********************************************************************/
/*   Handle 808 instructions.					      */
/**********************************************************************/
int
handle_808()
{	WORD64	*tmp;

	n4 = get_nibble(cpu.pc+3);
	switch (n4) {
	  case 0:
	  	/***********************************************/
	  	/*   INTON				       */
	  	/***********************************************/
		cpu.ien = TRUE;
		cpu.pc += 4;
		if (cpu.intr_pending) {
			cpu.intr_pending = FALSE;
			push(cpu.pc);
			cpu.pc = 0xf;
			}
		break;
	  case 1:
	  	/***********************************************/
	  	/*   RSI				       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
	  	if (n5 != 0) {
			old_pc = cpu.pc;
			return -1;
			}
		/***********************************************/
		/*   We  shouldnt  really  zero  the keyboard  */
		/*   map  --  we need a mechanism to indicate  */
		/*   that a key is release.		       */
		/***********************************************/
		memset((char *)&cpu.keybuf, 0, sizeof cpu.keybuf);
		cpu.intr_pending = 0;
/*		cpu.ien = TRUE;*/
		if (io_flag & IO_OTHER) {
			printf("%05lx: RSI [instr=%lx]\n",
				cpu.pc, 
				cpu.instr_count);
			}
		cpu.pc += 5;
		break;
	  case 2:
	  	/***********************************************/
	  	/*   MOVE.Pn #...., A			       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		WRAP_CHECK(n5);
	  	tmp = get_imm_nibble(cpu.pc+5, n5 + 1);
		rotate_right_nibbles(&cpu.A, cpu.p);
		clear_field(&cpu.A, n5 + 1);
		cpu.A.lo |= tmp->lo;
		cpu.A.hi |= tmp->hi;
		rotate_left_nibbles(&cpu.A, cpu.p);
		cpu.pc += n5 + 6;
	  	break;
	  case 4:
	  	/***********************************************/
	  	/*   CLRB #..,A				       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		cpu.A.lo &= ~(1L << n5);
		cpu.pc += 5;
	  	break;
	  case 5:
	  	/***********************************************/
	  	/*   SETB #..,A				       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		cpu.A.lo |= (1L << n5);
		cpu.pc += 5;
	  	break;
	  case 6:
	  	/***********************************************/
	  	/*   BRBC x,A,label			       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		do_jump67((cpu.A.lo & (1L << n5)) == 0);
		break;
	  case 7:
	  	/***********************************************/
	  	/*   BRBS x,A,label			       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		do_jump67((cpu.A.lo & (1L << n5)) != 0);
		break;
	  case 8:
	  	/***********************************************/
	  	/*   CLRB #..,C				       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		cpu.C.lo &= ~(1L << n5);
		cpu.pc += 5;
	  	break;
	  case 9:
	  	/***********************************************/
	  	/*   SETB #..,C				       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		cpu.C.lo |= (1L << n5);
		cpu.pc += 5;
	  	break;
	  case 0xa:
	  	/***********************************************/
	  	/*   BRBC #nn,c,label			       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		do_jump67((cpu.C.lo & (1L << n5)) == 0);
	  	break;
	  case 0xb:
	  	/***********************************************/
	  	/*   BRBS #nn,c,label			       */
	  	/***********************************************/
		n5 = get_nibble(cpu.pc+4);
		do_jump67((cpu.C.lo & (1L << n5)) != 0);
	  	break;
	  case 0xc:
	  	/***********************************************/
	  	/*   JUMP.A @A				       */
	  	/***********************************************/
		cpu.pc = read_nibbles(cpu.A.lo & ADDRESS_20, 5);
	  	break;
	  case 0xe:
	  	/***********************************************/
	  	/*   JUMP.A @C				       */
	  	/***********************************************/
		cpu.pc = read_nibbles(cpu.C.lo & ADDRESS_20, 5);
	  	break;
	  case 0xf:
	  	/***********************************************/
	  	/*   INTOFF				       */
	  	/***********************************************/
		cpu.ien = FALSE;
		cpu.pc += 4;
		break;
	  default:
		old_pc = cpu.pc;
	  	return -1;
	  }
	return 0;
}
int
handle_81()
{
	WORD64	tmp, tmp1, src;
	long	t;
	int	mode;

  	switch (n3) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  	t = cpu.reg[n3].hi >> 28;
		cpu.reg[n3].hi = (cpu.reg[n3].hi << 4) | 
				(cpu.reg[n3].lo >> 28);
		cpu.reg[n3].lo = (cpu.reg[n3].lo << 4) | t;
	  	break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	/***********************************************/
	  	/*   SRN				       */
	  	/***********************************************/
	  	n3 &= 3;
	  	t = cpu.reg[n3].lo << 28;
		cpu.reg[n3].lo = (cpu.reg[n3].lo >> 4) | 
				(cpu.reg[n3].hi << 28);
		cpu.reg[n3].hi = (cpu.reg[n3].hi >> 4) | t;
		if (t)
			cpu.hst |= SB;
	  	break;
	  case 8:
	  	/***********************************************/
	  	/*   ADD/SUB.? #$n,reg			       */
	  	/***********************************************/
		tmp.hi = 0;
		tmp.lo = get_nibble(cpu.pc+5) + 1;
		src = *extract_field(&cpu.reg[n5 & 3], n4);
		mode = cpu.hexmode;
		cpu.hexmode = TRUE;
		switch (n5) {
		  case 0: case 1: case 2: case 3:
			add_fs(&src, &tmp, n4);
			break;
		  case 8: case 9: case 0xa: case 0xb:
			sub_fs(&src, &tmp, n4);
			break;
		  default:
			old_pc = cpu.pc;
			cpu.hexmode = mode;
		  	return -1;
		  }
		cpu.hexmode = mode;
		replace_field(&cpu.reg[n5 & 3], n4, &src);
		cpu.pc += 6;
		return 0;
	  case 9:
	  	/***********************************************/
	  	/*   SRB.f d				       */
	  	/***********************************************/
		tmp = *extract_field(&cpu.reg[n5 & 3], n4);
		shift_right_nibbles(&tmp, cpu.field_offset);
		if (tmp.lo & 1)
			cpu.hst |= SB;
		tmp.lo = ((tmp.hi & 1) << 31) | (tmp.lo >> 1);
		tmp.hi >>= 1;
		shift_right_nibbles(&tmp, cpu.field_offset);
		replace_field(&cpu.reg[n5 & 3], n4, &tmp);
		cpu.pc += 5;
		return 0;
	  case 0xa:
	  	n6 = get_nibble(cpu.pc+5);
		switch (n5) {
		  case 0:
		  	/***********************************************/
		  	/*   MOVE.f A/C, Rn			       */
		  	/***********************************************/
			if (n6 & 8)
				tmp = *extract_field(&cpu.C, n4);
			else
				tmp = *extract_field(&cpu.A, n4);
			replace_field(&cpu.Rreg[n6 & 7], n4, &tmp);
			break;
		  case 1:
		  	/***********************************************/
		  	/*   MOVE.f Rn, A/C			       */
		  	/***********************************************/
			tmp = *extract_field(&cpu.Rreg[n6 & 7], n4);
			if (n6 & 8)
				replace_field(&cpu.C, n4, &tmp);
			else
				replace_field(&cpu.A, n4, &tmp);
			break;
		  case 2:
		  	/***********************************************/
		  	/*   SWAP? A/C, Rn			       */
		  	/***********************************************/
			tmp = *extract_field(&cpu.Rreg[n6 & 7], n4);
			switch (n6) {
			  case 0: case 1: case 2: case 3: case 4:
				tmp1 = *extract_field(&cpu.A, n4);
				replace_field(&cpu.A, n4, &tmp);
				replace_field(&cpu.Rreg[n6 & 7], n4, &tmp1);
				break;
			  case 8: case 9: case 0xa: case 0xb: case 0xc:
				tmp1 = *extract_field(&cpu.C, n4);
				replace_field(&cpu.C, n4, &tmp);
				replace_field(&cpu.Rreg[n6 & 7], n4, &tmp1);
				break;
			  default:
				old_pc = cpu.pc;
			  	return -1;
			  }
			break;
		  default:
			old_pc = cpu.pc;
		  	return -1;
		  }
		cpu.pc += 6;
		return 0;
	  case 0xb:
		switch (n4) {
		  case 2:
	  		/***********************************************/
	  		/*   JUMP.A A				       */
	  		/***********************************************/
	  		cpu.pc = cpu.A.lo & ADDRESS_20;
			break;
		  case 3:
	  		/***********************************************/
	  		/*   JUMP.A C				       */
	  		/***********************************************/
	  		cpu.pc = cpu.C.lo & ADDRESS_20;
			break;
		  default: 
			old_pc = cpu.pc;
		  	return -1;
		  }
		return 0;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n3 &= 3;
	  	t = cpu.reg[n3].lo & 1;
		cpu.reg[n3].lo = (cpu.reg[n3].lo >> 1) | 
				(cpu.reg[n3].hi << 31);
		cpu.reg[n3].hi >>= 1;
		if (t)
			cpu.hst |= SB;
	  	break;
	  default: 
		old_pc = cpu.pc;
	  	return -1;
	  }
	cpu.pc += 3;
	return 0;
}
int
handle_9()
{	WORD64	src, dst;
	int	is_equal;
static	struct pairs pair1_tbl[] = {
		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.A, &cpu.C},
		{&cpu.C, &cpu.D},

		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.A, &cpu.C},
		{&cpu.C, &cpu.D},

		{&cpu.A, &constant_0},
		{&cpu.B, &constant_0},
		{&cpu.C, &constant_0},
		{&cpu.D, &constant_0},

		{&cpu.A, &constant_0},
		{&cpu.B, &constant_0},
		{&cpu.C, &constant_0},
		{&cpu.D, &constant_0},
		};
static struct pairs pair2_tbl[] = {
		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.C, &cpu.A},
		{&cpu.D, &cpu.C},

		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.C, &cpu.A},
		{&cpu.D, &cpu.C},

		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.C, &cpu.A},
		{&cpu.D, &cpu.C},

		{&cpu.A, &cpu.B},
		{&cpu.B, &cpu.C},
		{&cpu.C, &cpu.A},
		{&cpu.D, &cpu.C},
		};

	if ((n2 & 0x8) == 0) {
		src = *extract_field(pair1_tbl[n3].src, n2 & 7);
		dst = *extract_field(pair1_tbl[n3].dst, n2 & 7);
		}
	else {
		src = *extract_field(pair2_tbl[n3].src, n2 & 7);
		dst = *extract_field(pair2_tbl[n3].dst, n2 & 7);
		}

	is_equal = src.lo == dst.lo && src.hi == dst.hi;
	if ((n2 & 0x08) == 0) {
		if (n3 & 4)
			is_equal = !is_equal;
		/***********************************************/
		/*   Equality operator.			       */
		/***********************************************/
		do_jump45(is_equal);
		}
	else {
/*		CHECK_HEX();*/
		/***********************************************/
		/*   Relational operator.		       */
		/***********************************************/
		sub_fs(&src, &dst, n2 & 7);
		switch (n3 & 0xc) {
		  case 0x0:
		  	do_jump45(!is_equal && cpu.cy == 0);
			break;
		  case 0x4:
		  	do_jump45(!is_equal && cpu.cy != 0);
			break;
		  case 0x8:
		  	do_jump45(cpu.cy == 0 || is_equal);
			break;
		  case 0xc:
		  	do_jump45(cpu.cy || is_equal);
			break;
		  }
		}
	return 0;
}
int
handle_a()
{	WORD64	*wp;
	WORD64	src, dst;
static struct pairs ex_tbl[] = {
	{&cpu.A, &cpu.B},
	{&cpu.B, &cpu.C},
	{&cpu.A, &cpu.C},
	{&cpu.C, &cpu.D}
	};

	if (n2 < 8) {
		src = *extract_field(&cpu.reg[n3 & 3], n2);
		switch (n3) {
		  case 0: dst = *extract_field(&cpu.B, n2); break;
		  case 1: dst = *extract_field(&cpu.C, n2); break;
		  case 2: dst = *extract_field(&cpu.A, n2); break;
		  case 3: dst = *extract_field(&cpu.C, n2); break;
		  case 4: dst = *extract_field(&cpu.A, n2); break;
		  case 5: dst = *extract_field(&cpu.B, n2); break;
		  case 6: dst = *extract_field(&cpu.C, n2); break;
		  case 7: dst = *extract_field(&cpu.D, n2); break;
		  case 8:
		  case 9:
		  case 0xa:
		  case 0xb:
		  	dst = *extract_field(&cpu.reg[dest_reg[n3 & 3]], n2); 
			add_fs(&src, &dst, n2);
			replace_field(&cpu.reg[dest_reg[n3 & 3]], n2, &src);
			cpu.pc += 3;
			return 0;
		  case 0xc:
		  case 0xd:
		  case 0xe:
		  case 0xf:
		  	dst = constant_1;
			shift_left_nibbles(&dst, cpu.field_offset);
			sub_fs(&src, &dst, n2);
			replace_field(&cpu.reg[n3 & 3], n2, &src);
			cpu.pc += 3;
			return 0;
		  }
		add_fs(&src, &dst, n2);
		replace_field(&cpu.reg[n3 & 3], n2, &src);
		cpu.pc += 3;
		return 0;
		}
	cpu.pc += 3;
	switch (n3) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
		/***********************************************/
		/*   CLR.? reg				       */
		/***********************************************/
		wp = extract_field(&cpu.reg[n3], n2 & 7);
		cpu.reg[n3].lo &= ~wp->lo;
		cpu.reg[n3].hi &= ~wp->hi;
		break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	/***********************************************/
	  	/*   MOVE.? s,d				       */
	  	/***********************************************/
		n3 &= 3;
		wp = extract_field(&cpu.reg[dest_reg[n3]], n2 & 7);
		replace_field(&cpu.reg[n3], n2 & 7, wp);
		break;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
	  	/***********************************************/
	  	/*   MOVE.? s,d (other way round).	       */
	  	/***********************************************/
		n3 &= 3;
		wp = extract_field(&cpu.reg[n3], n2 & 7);
		replace_field(&cpu.reg[dest_reg[n3]], n2 & 7, wp);
	  	break;

	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n3 &= 3;
		n2 &= 7;
	  	src = *extract_field(ex_tbl[n3].src, n2);
	  	dst = *extract_field(ex_tbl[n3].dst, n2);
		replace_field(ex_tbl[n3].src, n2, &dst);
		replace_field(ex_tbl[n3].dst, n2, &src);
	  	break;
	  }
	return 0;
}
/**********************************************************************/
/*   Handle various arithmetic instructions.			      */
/**********************************************************************/
int
handle_b()
{	WORD64	src, dst;
	unsigned long	tmp;
static struct pairs tbl[] = {
	{&cpu.A, &cpu.B},
	{&cpu.B, &cpu.C},
	{&cpu.C, &cpu.A},
	{&cpu.D, &cpu.C},

	{&cpu.A, &constant_1},
	{&cpu.B, &constant_1},
	{&cpu.C, &constant_1},
	{&cpu.D, &constant_1},

	{&cpu.B, &cpu.A},
	{&cpu.C, &cpu.B},
	{&cpu.A, &cpu.C},
	{&cpu.C, &cpu.D},

	{&cpu.A, &cpu.B},
	{&cpu.B, &cpu.C},
	{&cpu.C, &cpu.A},
	{&cpu.D, &cpu.C}
	};

	if ((n2 & 0x8) == 0) {
		src = *extract_field(tbl[n3].src, n2);
		dst = *extract_field(tbl[n3].dst, n2);
		switch (n3) {
		  case 0:
		  case 1:
		  case 2:
		  case 3:
		  	sub_fs(&src, &dst, n2);
			break;
		  case 4:
		  case 5:
		  case 6:
		  case 7:
		  	/***********************************************/
		  	/*   A=A+1				       */
		  	/***********************************************/
		  	dst.lo = 1;
			dst.hi = 0;
		  	shift_left_nibbles(&dst, cpu.field_offset);
		  	add_fs(&src, &dst, n2);
			break;
		  case 8:
		  case 9:
		  case 0xa:
		  case 0xb:
		  	sub_fs(&src, &dst, n2);
			break;
		  case 0xc:
		  case 0xd:
		  case 0xe:
		  case 0xf:
		  	sub_fs(&dst, &src, n2);
			src = dst;
			break;
		  }
		replace_field(tbl[n3].src, n2, &src);
		cpu.pc += 3;
		return 0;
		}

	n2 &= 7;
	switch (n3) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  	/***********************************************/
	  	/*   ?SL				       */
	  	/***********************************************/
		src = *extract_field(&cpu.reg[n3], n2);
		tmp = src.lo >> 28;
		src.lo <<= 4;
		src.hi = (src.hi << 4) | tmp;
		replace_field(&cpu.reg[n3], n2, &src);
		break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	/***********************************************/
	  	/*   ?SR				       */
	  	/***********************************************/
		n3 &= 3;
		src = *extract_field(&cpu.reg[n3], n2);
		if (cpu.field_offset >= 8) {
			if (src.hi & (0xf << (4 * (cpu.field_offset-8))))
				cpu.hst |= SB;
			}
		else {
			if (src.lo & (0xf << (4 * cpu.field_offset)))
				cpu.hst |= SB;
			}
		tmp = src.hi << 28;
		src.lo = (src.lo >> 4) | tmp;
		src.hi >>= 4;
		replace_field(&cpu.reg[n3], n2, &src);
		break;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
		n3 &= 3;
		src = *extract_field(&cpu.reg[n3], n2);
		dst.hi = 0;
		dst.lo = 0;
		sub_fs(&dst, &src, n2);
		if (dst.hi || dst.lo)
			cpu.cy = 1;
		else
			cpu.cy = 0;
		replace_field(&cpu.reg[n3], n2, &dst);
	  	break;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	/***********************************************/
	  	/*   One's     complement     --     actually  */
	  	/*   reg.f = -reg.f - 1			       */
	  	/***********************************************/
		n3 &= 3;
		src = *extract_field(&cpu.reg[n3], n2);
		if (cpu.hexmode) {
			src.hi = ~src.hi;
			src.lo = ~src.lo;
			}
		else {
			dst.hi = 0;
			dst.lo = 0;
			sub_fs(&dst, &src, n2);
			src = dst;
			dst.hi = 0;
			dst.lo = 1;
			shift_left_nibbles(&dst, cpu.field_offset);
			sub_fs(&src, &dst, n2);
			}
		replace_field(&cpu.reg[n3], n2, &src);
		cpu.cy = 0;
	  	break;
	  }
	cpu.pc += 3;
	return 0;
}
int
handle_c()
{
	unsigned long	a1,a2;

	switch (n2) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  	a1 = cpu.reg[n2].lo & ADDRESS_20;
		a2 = cpu.reg[dest_reg[n2]].lo & ADDRESS_20;
do_add:
		if (cpu.hexmode) {
			a1 += a2;
			cpu.cy = a1 & ~ADDRESS_20 ? 1 : 0;
			}
		else
			a1 = add_dec(a1, a2, 5);
		cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | (a1 & ADDRESS_20);
	  	break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	n2 &= 3;
		a1 = a2 = cpu.reg[n2].lo & ADDRESS_20;
		goto do_add;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
	  	n2 &= 3;
	  	a1 = cpu.reg[n2].lo & ADDRESS_20;
		a2 = cpu.reg[dest_reg[n2]].lo & ADDRESS_20;
		n2 = dest_reg[n2];
		goto do_add;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n2 &= 3;
		a1 = cpu.reg[n2].lo & ADDRESS_20;
		if (cpu.hexmode)
			cpu.cy = a1-- == 0 ? 1 : 0;
		else {
			a1 = sub_dec(a1, 1L, 5);
			}
	  	cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | (a1 & ADDRESS_20);
	  	break;
	  }
	cpu.pc += 2;
	return 0;
}
int
handle_d()
{	long	t;
	WORD64	*wp1, *wp2;
static	struct pairs tbl[] = {
	{&cpu.A, &cpu.B},
	{&cpu.B, &cpu.C},
	{&cpu.A, &cpu.C},
	{&cpu.C, &cpu.D}
	};

	switch (n2) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
		/***********************************************/
		/*   CLR.A reg				       */
		/***********************************************/
		cpu.reg[n2].lo &= ~ADDRESS_20;
	  	break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
		cpu.reg[n2 & 3].lo = (cpu.reg[n2 & 3].lo & ~ADDRESS_20) |
			(cpu.reg[dest_reg[n2 & 3]].lo & ADDRESS_20);
		break;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
		cpu.reg[dest_reg[n2 & 3]].lo = 
			(cpu.reg[dest_reg[n2 & 3]].lo & ~ADDRESS_20) |
			(cpu.reg[n2 & 3].lo & ADDRESS_20);
		break;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n2 &= 3;
		wp1 = tbl[n2].src;
		wp2 = tbl[n2].dst;
	  	t = wp1->lo & ADDRESS_20;
		wp1->lo = (wp1->lo & ~ADDRESS_20) | (wp2->lo & ADDRESS_20);
		wp2->lo = (wp2->lo & ~ADDRESS_20) | t;
		break;
	  default: 
		old_pc = cpu.pc;
	  	return -1;
	  }
	cpu.pc += 2;
	return 0;
}
int
handle_e()
{	
	WORD20	t;
	WORD20	t1;

	switch (n2) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  	t = cpu.reg[n2].lo & ADDRESS_20;
		t1 = cpu.reg[dest_reg[n2]].lo & ADDRESS_20;
		if (cpu.hexmode) {
			t -= t1;
			cpu.cy = t & ~ADDRESS_20 ? 1 : 0;
			}
		else
			t = sub_dec(t, t1, 5);
		cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | (t & ADDRESS_20);
		break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	t = cpu.reg[n2 & 3].lo & ADDRESS_20;
		if (cpu.hexmode)
			cpu.cy = ++t & ~ADDRESS_20 ? 1 : 0;
		else
			t = add_dec(t, 1L, 5);
		cpu.reg[n2 & 3].lo = (cpu.reg[n2 & 3].lo & ~ADDRESS_20) | (t & ADDRESS_20);
	  	break;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
	  	t = cpu.reg[dest_reg[n2 & 3]].lo & ADDRESS_20;
		t1 = cpu.reg[n2 & 3].lo & ADDRESS_20;
		if (cpu.hexmode) {
			t -= t1;
			cpu.cy = t & ~ADDRESS_20 ? 1 : 0;
			}
		else
			t = sub_dec(t, t1, 5);
		cpu.reg[dest_reg[n2 & 3]].lo = (cpu.reg[dest_reg[n2 & 3]].lo & ~ADDRESS_20) | (t & ADDRESS_20);
	  	break;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n2 &= 3;
	  	t = cpu.reg[n2].lo & ADDRESS_20;
		t1 = cpu.reg[dest_reg[n2]].lo & ADDRESS_20;
		if (cpu.hexmode) {
			t = t1 - t;
			cpu.cy = t & ~ADDRESS_20 ? 1 : 0;
			}
		else
			t = sub_dec(t1, t, 5);
		cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | (t & ADDRESS_20);
		break;
	  }
	cpu.pc += 2;
	return 0;
}
int
handle_f()
{	WORD20	t;

	switch (n2) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
/*		CHECK_HEX();*/
	  	t = cpu.reg[n2].lo & ADDRESS_20;
	  	cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | 
			((t << 4) & ADDRESS_20);
	  	break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  	n2 &= 3;
	  	t = cpu.reg[n2].lo & ADDRESS_20;
	  	cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | 
			((t >> 4) & ADDRESS_20);
		if (t & 0xf)
			cpu.hst |= SB;
	  	break;
	  case 8:
	  case 9:
	  case 0xa:
	  case 0xb:
	  	n2 &= 3;
	  	t = cpu.reg[n2].lo & ADDRESS_20;
		if (cpu.hexmode)
			t = (-t) & ADDRESS_20;
		else
			t = tens_complement(t, 5);
		cpu.cy = t != 0;
		cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | t;
	  	break;
	  case 0xc:
	  case 0xd:
	  case 0xe:
	  case 0xf:
	  	n2 &= 3;
	  	t = cpu.reg[n2].lo & ADDRESS_20;
		if (cpu.hexmode) {
			t = (~t) & ADDRESS_20;
			}
		else {
			t = sub_dec(0, t, 5);
			t = sub_dec(t, 1, 5) & ADDRESS_20;
			}
		cpu.reg[n2].lo = (cpu.reg[n2].lo & ~ADDRESS_20) | t;
		cpu.cy = 0;
	  	break;
	  }
	cpu.pc += 2;
	return 0;
}
int
do_switches(argc, argv)
int	argc;
char	**argv;
{	int	i;
	char	*cp;

	for (i = 1; i < argc; i++) {
		cp = argv[i];
		if (*cp++ != '-')
			return i;
		while (*cp) {
			switch (*cp++) {
			  default: usage();
			  }
			}
		}
	return i;
}
void
usage()
{
	fprintf(stderr, "em48 -- HP48 emulator\n");
	fprintf(stderr, "Usage: em48 [start-pc]\n");
	exit(1);
}
void
do_jump45(expr)
int	expr;
{	short	s;

	if (expr == 0) {
		cpu.pc += 5;
		cpu.cy = 0;
		return;
		}
	cpu.cy = 1;
	s = (n5 << 4) | n4;
	if (s == 0) {
		cpu.pc = pop();
		}
	else {
		if (n5 & 8)
			s |= 0xff00;
		cpu.pc += s + 3;
		}
}
void
do_jump67(expr)
int	expr;
{	short	s;

	if (expr == 0) {
		cpu.pc += 7;
		cpu.cy = 0;
		return;
		}
	cpu.cy = 1;
	n6 = get_nibble(cpu.pc+5);
	n7 = get_nibble(cpu.pc+6);
	s = (n7 << 4) | n6;
	if (s == 0)
		cpu.pc = pop();
	else {
		if (n7 & 8)
			s |= 0xff00;
		cpu.pc += s + 5;
		}
}
