/**********************************************************************/
/*   Routines to manipulate fields.				      */
/**********************************************************************/
# include	"hp48.h"
# include	"dis48.h"
# include "proto.h"

void	set_carry PROTO((WORD64 *, int));
/**********************************************************************/
/*   Masks to extract nibbles from 64 bit quantities.		      */
/**********************************************************************/
long nibble_masks[] = {
	0x0000000f,
	0x000000f0,
	0x00000f00,
	0x0000f000,
	0x000f0000,
	0x00f00000,
	0x0f000000,
	0xf0000000,
	0x0000000f,
	0x000000f0,
	0x00000f00,
	0x0000f000,
	0x000f0000,
	0x00f00000,
	0x0f000000,
	0xf0000000
	};

extern CPU cpu;
extern int	show_dec;
/**********************************************************************/
/*   Routine  to  rotate a  64-bit  number  by  specified  number of  */
/*   nibbles.							      */
/**********************************************************************/
void
rotate_left_nibbles(wp, num)
WORD64	*wp;
int	num;
{	unsigned long	t;
	while (num-- > 0) {
		t = wp->hi >> 28;
		wp->hi = (wp->hi << 4) | (wp->lo >> 28);
		wp->lo = (wp->lo << 4) | t;
		}
}
/**********************************************************************/
/*   Routine  to  shift  a  64-bit  number  by  specified  number of  */
/*   nibbles.							      */
/**********************************************************************/
void
shift_left_nibbles(wp, num)
WORD64	*wp;
int	num;
{
	while (num-- > 0) {
		wp->hi = (wp->hi << 4) | (wp->lo >> 28);
		wp->lo <<= 4;
		}
}
/**********************************************************************/
/*   Routine  to  shift right a 64-bit number by specified number of  */
/*   nibbles.							      */
/**********************************************************************/
void
shift_right_nibbles(wp, num)
WORD64	*wp;
int	num;
{
	while (num-- > 0) {
		wp->lo = (wp->lo >> 4) | (wp->hi << 28);
		wp->hi >>= 4;
		}
}
/**********************************************************************/
/*   Routine  to  rotate  right  a 64-bit number by specified number  */
/*   of nibbles.						      */
/**********************************************************************/
void
rotate_right_nibbles(wp, num)
WORD64	*wp;
int	num;
{	unsigned long t;

	while (num-- > 0) {
		t = wp->lo << 28;
		wp->lo = (wp->lo >> 4) | (wp->hi << 28);
		wp->hi = (wp->hi >> 4) | t;
		}
}
/**********************************************************************/
/*   Function  to  return  a  'left-justified'  field  based  on the  */
/*   field mode specified.					      */
/**********************************************************************/
WORD64 *
extract_field(val, mode)
WORD64	*val;
int	mode;
{	static WORD64 res;
	int	i;

	switch (mode) {
	  case 0:
	  	/***********************************************/
	  	/*   P -- 1 nibble at position P	       */
	  	/***********************************************/
		res.lo = res.hi = 0;
		if (cpu.p >= 8)
			res.hi = val->hi & nibble_masks[cpu.p];
		else
			res.lo = val->lo & nibble_masks[cpu.p];
		cpu.field_size = 1;
		cpu.field_offset = cpu.p;
		break;
	  case 1:
	  	/***********************************************/
	  	/*   WP -- n nibbles at 0..P		       */
	  	/***********************************************/
		res.lo = res.hi = 0;
		for (i = 0; i <= cpu.p; i++)
			if (i >= 8)
				res.hi |= val->hi & nibble_masks[i];
			else
				res.lo |= val->lo & nibble_masks[i];
		cpu.field_size = cpu.p+1;
		cpu.field_offset = 0;
		break;
	  case 2:
	  	/***********************************************/
	  	/*   XS -- Get sign of exponent.	       */
	  	/***********************************************/
	  	res.lo = val->lo & SEXP_MASK;
		res.hi = 0;
		cpu.field_size = 1;
		cpu.field_offset = 2;
	  	break;
	  case 3:
	  	/***********************************************/
	  	/*   X -- get 3 nibble of exponent	       */
	  	/***********************************************/
		res.lo = val->lo & EXP_MASK;
		res.hi = 0;
		cpu.field_size = 3;
		cpu.field_offset = 0;
		break;
	  case 4:
	  	/***********************************************/
	  	/*   S -- get mantissa sign		       */
	  	/***********************************************/
		res.lo = 0;
		res.hi = val->hi & SIGN_MASK;
		cpu.field_size = 1;
		cpu.field_offset = 15;
	  	break;
	  case 5:
	  	/***********************************************/
	  	/*   M -- get mantissa			       */
	  	/***********************************************/
		res.lo = val->lo & MANT_LO_MASK;
		res.hi = val->hi & MANT_HI_MASK;
		cpu.field_offset = 3;
		cpu.field_size = 12;
		break;
	  case 6:
	  	/***********************************************/
	  	/*   B -- Get byte			       */
	  	/***********************************************/
		res.lo = val->lo & 0xff;
		res.hi = 0;
		cpu.field_size = 2;
		cpu.field_offset = 0;
		break;
	  case 7:
	  	/***********************************************/
	  	/*   W					       */
	  	/***********************************************/
		res = *val;
		cpu.field_offset = 0;
		cpu.field_size = 16;
		break;
	  case 0xf:
	  	/***********************************************/
	  	/*   A -- 5 nibbles for address.	       */
	  	/***********************************************/
		res.lo = val->lo & ADDRESS_20;
		res.hi = 0;
		cpu.field_offset = 0;
		cpu.field_size = 5;
		break;
	  default:
	  	printf("%05lx: ??? EXTRACT FIELD -- UNSUPPORTED MODE 0x%x\n", cpu.pc, mode);
	  }
	return &res;
}
/**********************************************************************/
/*   Function  to  insert  a  field  into the destination. We assume  */
/*   that newval has no extraneous bits outside of the nibble field.  */
/**********************************************************************/
void
replace_field(val, mode, newval)
WORD64	*val;
int	mode;
WORD64	*newval;
{
	int	i;

	switch (mode) {
	  case 0:
	  	/***********************************************/
	  	/*   P -- 1 nibble at position P	       */
	  	/***********************************************/
		if (cpu.field_offset >= 8) {
			val->hi &= ~nibble_masks[cpu.p];
			val->hi |= newval->hi & nibble_masks[cpu.p];
			}
		else {
			val->lo &= ~nibble_masks[cpu.p];
			val->lo |= newval->lo & nibble_masks[cpu.p];
			}
		break;
	  case 1:
	  	/***********************************************/
	  	/*   WP -- n nibbles at 0..P		       */
	  	/***********************************************/
		for (i = 0; i < cpu.field_size; i++)
			if (i >= 8)
				val->hi &= ~nibble_masks[i];
			else
				val->lo &= ~nibble_masks[i];
		for (; i < 16; i++)
			if (i >= 8)
				newval->hi &= ~nibble_masks[i];
			else
				newval->lo &= ~nibble_masks[i];
		val->hi |= newval->hi;
		val->lo |= newval->lo;
		break;
	  case 2:
	  	/***********************************************/
	  	/*   XS -- Get sign of exponent.	       */
	  	/***********************************************/
		val->lo &= ~SEXP_MASK;
		val->lo |= newval->lo & SEXP_MASK;
	  	break;
	  case 3:
	  	val->lo &= ~EXP_MASK;
		val->lo |= newval->lo & EXP_MASK;
		break;
	  case 4:
	  	/***********************************************/
	  	/*   S -- get mantissa sign		       */
	  	/***********************************************/
		val->hi &= ~SIGN_MASK;
		val->hi |= newval->hi & SIGN_MASK;
	  	break;
	  case 5:
	  	/***********************************************/
	  	/*   M -- get mantissa			       */
	  	/***********************************************/
		val->lo &= ~MANT_LO_MASK;
		val->hi &= ~MANT_HI_MASK;
		val->lo |= newval->lo & MANT_LO_MASK;
		val->hi |= newval->hi & MANT_HI_MASK;
		break;
	  case 6:
	  	/***********************************************/
	  	/*   B -- Get byte			       */
	  	/***********************************************/
		val->lo &= ~BYTE_MASK;
		val->lo |= newval->lo & BYTE_MASK;
		break;
	  case 7:
	  	/***********************************************/
	  	/*   W					       */
	  	/***********************************************/
		*val = *newval;
		break;
	  case FIELD_XP:
	  	/***********************************************/
	  	/*   XP -- n nibbles at P..P+n-1	       */
	  	/***********************************************/
		for (i = 0; i < cpu.field_offset; i++)
			if (i >= 8)
				newval->hi &= ~nibble_masks[i];
			else
				newval->lo &= ~nibble_masks[i];
		for ( ; i < cpu.field_offset + cpu.field_size; i++)
			if (i >= 8)
				val->hi &= ~nibble_masks[i];
			else
				val->lo &= ~nibble_masks[i];
		for (; i < 16; i++)
			if (i >= 8)
				newval->hi &= ~nibble_masks[i];
			else
				newval->lo &= ~nibble_masks[i];
		val->hi |= newval->hi;
		val->lo |= newval->lo;
		break;
	  case 0xf:
	  	/***********************************************/
	  	/*   A -- 5 nibbles for address field.	       */
	  	/***********************************************/
		val->lo = (val->lo & ~ADDRESS_20) | (newval->lo & ADDRESS_20);
	  	break;
	  default:
	  	printf("%05lx: ??? REPLACE FIELD -- UNSUPPORTED MODE 0x%x\n", cpu.pc, mode);
	  }
}
/**********************************************************************/
/*   Function to clear a field of nibbles.			      */
/**********************************************************************/
void
clear_field(wp, num)
WORD64	*wp;
int	num;
{
static unsigned long masks[] = {
	0xffffffff,
	0xfffffff0,
	0xffffff00,
	0xfffff000,
	0xffff0000,
	0xfff00000,
	0xff000000,
	0xf0000000,
	0x00000000
	};
	if (num <= 7)
		wp->lo &= masks[num];
	else {
		wp->lo = 0;
		wp->hi &= masks[num - 8];
		}
}

/**********************************************************************/
/*   Routine  similar  to  extract_field  but  we  have a pointer to  */
/*   memory.  We  are  going to load the field into a register so we  */
/*   only need the number of nibbles expressed by 'mode'.	      */
/**********************************************************************/
WORD64 *
extract_indirect_field(addr, mode)
long	addr;
int	mode;
{	static WORD64	val;

	switch (mode) {
	  case 0:
	  	/***********************************************/
	  	/*   P -- 1 nibble at position P	       */
	  	/***********************************************/
		val.hi = 0;
		val.lo = read_nibbles(addr, 1);
		cpu.field_size = 1;
		cpu.field_offset = cpu.p;
		break;
	  case 1:
	  	/***********************************************/
	  	/*   WP -- n nibbles at 0..P		       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, cpu.p+1);
		if (cpu.p > 7)
			val.hi = read_nibbles(addr+8, cpu.p+1-8);
		else
			val.hi = 0;
		cpu.field_size = cpu.p+1;
		cpu.field_offset = 0;
		break;
	  case 2:
	  	/***********************************************/
	  	/*   XS -- Get sign of exponent.	       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 1);
		val.hi = 0;
		cpu.field_size = 1;
		cpu.field_offset = 2;
	  	break;
	  case 3:
	  	/***********************************************/
	  	/*   X -- get 3 nibble of exponent	       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 3);
		val.hi = 0;
		cpu.field_size = 3;
		cpu.field_offset = 0;
		break;
	  case 4:
	  	/***********************************************/
	  	/*   S -- get mantissa sign		       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 1);
		val.hi = 0;
		cpu.field_size = 1;
		cpu.field_offset = 15;
	  	break;
	  case 5:
	  	/***********************************************/
	  	/*   M -- get mantissa			       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 8);
		val.hi = read_nibbles(addr + 8, 4);
		cpu.field_offset = 3;
		cpu.field_size = 12;
		break;
	  case 6:
	  	/***********************************************/
	  	/*   B -- Get byte			       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 2);
		val.hi = 0;
		cpu.field_size = 2;
		cpu.field_offset = 0;
		break;
	  case 7:
	  	/***********************************************/
	  	/*   W					       */
	  	/***********************************************/
		val.lo = read_nibbles(addr, 8);
		val.hi = read_nibbles(addr + 8, 8);
		cpu.field_offset = 0;
		cpu.field_size = 16;
		break;
	  default:
	  	printf("%05lx: ??? EXTRACT INDIRECT FIELD -- UNSUPPORTED MODE 0x%x\n", cpu.pc, mode);
	  }
	return &val;
}
/**********************************************************************/
/*   Routine  to  perform  64-bit  add  and set carry appropriate to  */
/*   the field type.						      */
/**********************************************************************/
void
add_fs(w1, w2, fs)
WORD64	*w1;
WORD64	*w2;
int	fs;
{
	if (cpu.hexmode == FALSE) {
		add_dec64(w1, w2);
		if (cpu.field_offset + cpu.field_size == 16)
			return;
		cpu.cy = 0;
		set_carry(w1, fs);
		if (show_dec)
			printf("        size=%x off=%x cy=%d %08lx:%08lx\n", 
				cpu.field_size, cpu.field_offset, cpu.cy, w1->hi, w1->lo);
		}
	else {
		add_64(w1, w2);
		set_carry(w1, fs);
		}
}
void
sub_fs(w1, w2, fs)
WORD64	*w1;
WORD64	*w2;
int	fs;
{

	if (cpu.hexmode == FALSE) {
		sub_dec64(w1, w2);
		if (cpu.field_offset + cpu.field_size == 16)
			return;
		cpu.cy = 0;
		set_carry(w1, fs);
		if (show_dec)
			printf("        size=%x off=%x cy=%d %08lx:%08lx\n", 
				cpu.field_size, cpu.field_offset, cpu.cy, w1->hi, w1->lo);
		}
	else {
		sub_64(w1, w2);
		set_carry(w1, fs);
		}
}
void
add_hexfs(w1, w2, fs)
WORD64	*w1;
WORD64	*w2;
int	fs;
{
	add_64(w1, w2);
	set_carry(w1, fs);
}
void
sub_hexfs(w1, w2, fs)
WORD64	*w1;
WORD64	*w2;
int	fs;
{
	sub_64(w1, w2);
	set_carry(w1, fs);
}
void
set_carry(w1, fs)
WORD64	*w1;
int	fs;
{	WORD64	*tmp;

	if (cpu.cy)
		return;
	tmp = extract_field(w1, fs);
	if (tmp == w1) {
		printf("***OOPS -- bad code detected\n");
		exit(1);
		}
	if (tmp->lo == w1->lo && tmp->hi == w1->hi)
		cpu.cy = 0;
	else
		cpu.cy = 1;
	*w1 = *tmp;
}
