/**********************************************************************/
/*   Include file used to maintain state of HP48 execution.	      */
/**********************************************************************/
#define ALWAYSLOCK
/**********************************************************************/
/*   Following  macro  used  to  allow  compilation  with ANSI C and  */
/*   non-ANSI C compilers automatically.			      */
/**********************************************************************/
#ifdef MY48
typedef unsigned char Nibble;
#else
 typedef unsigned char Nibble;
#endif

# if !defined(PROTO)
# 	if	defined(__STDC__) || defined(MSDOS)
#		define	PROTO(x)	x
#	else
#		define	PROTO(x)	()
#		define	volatile
#	endif
# endif

# if !defined(TRUE)
#	define	TRUE	1
#	define	FALSE	0
# endif

# if defined(__MSDOS__) && !defined(MSDOS)
#	define	MSDOS
# endif
/**********************************************************************/
/*   Some useful addresses.					      */
/**********************************************************************/
# define	TIMER1_CTRL	0x12E
# define	TIMER2_CTRL	0x12F
/**********************************************************************/
/*   Size of ROM in nibbles.					      */
/**********************************************************************/
# define	ROM_SIZE (512L * 1024L)
# define	ROM_SIZE_BYTES (ROM_SIZE / 2)
/**********************************************************************/
/*   Under  MSDOS  or any 16 bit machine we cannot read ROM into one  */
/*   segment  so  we  need an array of chunks. On 32 bit machines we  */
/*   dont need to worry about this.				      */
/*   NOTE::  Support  for  MSDOS has lapsed so these definitions are  */
/*   not actually used.						      */
/**********************************************************************/
# if defined(MSDOS)
#	define	MEM_CHUNK_SIZE	(16 * 1024)
#	define	MEM_SHIFT	14
# else
#	define	MEM_CHUNK_SIZE	ROM_SIZE_BYTES
#	define	MEM_SHIFT	20 /* Any number such that when shifted */
				   /* we get zero. */
# endif
# define	NUM_CHUNKS	(ROM_SIZE_BYTES / MEM_CHUNK_SIZE)
# define	RAM_SIZE	(1024 * 1024 / 2)
/**********************************************************************/
/*   Execution  mode  flags. We stuff all these in a single variable  */
/*   so  that  during  the  time critical inner loop we only need to  */
/*   make  one  test  if  no  special  debug  events  are  set (e.g.  */
/*   breakpoints, new mode, etc).				      */
/**********************************************************************/
# define	EXEC_STEP	0x01
# define	EXEC_TRACE	0x02
# define	EXEC_NEW_MODE	0x04
# define	EXEC_BKPT	0x08
# define	EXEC_INSTR	0x10
# define	EXEC_STOPPED	0x20
# define	EXEC_INTR	0x40
# define	EXEC_RUN	0x80
# define	EXEC_SIGIO	0x100
extern int exec_flags;
/**********************************************************************/
/*   Breakpoint types.						      */
/**********************************************************************/
# define	BP_RANGE	0x08
# define	BP_READ		0x04
# define	BP_WRITE	0x02
# define	BP_EXEC		0x01
/**********************************************************************/
/*   Instructions to watch which affect the hardware.		      */
/**********************************************************************/
# define	IO_CONFIG	0x01
# define	IO_IN		0x02
# define	IO_OUT		0x04
# define	IO_INTR		0x08
# define	IO_OTHER	0x10
# define	IO_MODE		0x20
/**********************************************************************/
/*   Various  masks  for  accessing 64 bit word. Note that we assume  */
/*   we  can  only  access  in  32-bit  chunks  so masks may only be  */
/*   applied to certain parts of the hi/lo structure.		      */
/**********************************************************************/
# define	EXP_MASK	0xfff
# define	SEXP_MASK	0xf00
# define	SIGN_MASK	0xf0000000L
# define	MANT_HI_MASK	0xfffffffL
# define	MANT_LO_MASK	0xfffff000L
# define	ADDRESS_20	0xfffffL
# define	BYTE_MASK	0xff
/**********************************************************************/
/*   Misc.							      */
/**********************************************************************/
enum run_state {
	CONTINUE,
	STEP,
	STOPPED
	};
/**********************************************************************/
/*   Following  structure  used to communicate with the symbol table  */
/*   library.							      */
/**********************************************************************/
# define	SF_PREFIX	0x01
typedef struct	symtab {
	long	addr;
	int	flags;
	char	*comment;
	} symtab_t;

/**********************************************************************/
/*   Keep statistics on execution.				      */
/**********************************************************************/
struct stats {
	long	illegal_instr;
	};
/**********************************************************************/
/*   Typedefs to match the sizes in the saturn CPU.		      */
/**********************************************************************/
typedef struct WORD64 {
	unsigned long	hi, lo;
	} WORD64;
typedef long	WORD20;
typedef short	WORD16;
typedef short	WORD12;
typedef short	WORD4;

/**********************************************************************/
/*   Structure  used  to  contain  information required to implement  */
/*   keyboard logic.						      */
/**********************************************************************/
typedef struct keystate_t {
	int	rows[9];
	} keystate_t;
/**********************************************************************/
/*   Following  structure  is  used  to  contain the entire state of  */
/*   the CPU.							      */
/**********************************************************************/
typedef struct CPU {
	WORD4	p;

	WORD64	reg[4];

	WORD20	rstk[8];	/* 8 level call/ret stack.	*/
	int	rstk_ptr;

	WORD20	D0;
	WORD20	D1;

	WORD20	pc;	/* Program counter.		*/

	WORD64	Rreg[5];

	WORD16	st;

# define	MP	0x08
# define	SR	0x04
# define	SB	0x02
# define	XM	0x01
	WORD4	hst;	/* Hardware status register. */

	int		cy;
	int		hexmode;
	WORD16		in_reg;
	WORD12		out_reg;
	keystate_t	keybuf;
	int		ien;
	int		intr_pending;

	/***********************************************/
	/*   Following  do  not  exist in the CPU but  */
	/*   used to keep track of state machine.      */
	/***********************************************/
	int	field_size;	/* Used during address calculation. */
	int	field_offset;	/* Number of nibbles.		    */

	Nibble	*rom;
	Nibble	*ram; 

	WORD20	ram_32k;	/* Based on the CONFIG instruction */
				/* figure out where RAM is.	   */
# define	MAX_CONFIGS	8
	WORD20	configs[MAX_CONFIGS];

	int	debug_mode;	/* Set to TRUE when in debugger to */
				/* avoid CRC being corrupted.	   */

	int	crc;		/* Current value of CRC.		*/

	int	daisy_state;	/* Used to emulate the DAISY-IN hardware */
				/* input used by MOVE.X ID,C; CONFIG and UNCNFG */
				/* instructions.			*/

	int	timer1;		/* 1-nibble value of TIMER1.		*/
	long	timer2;		/* Current value of 32-bit TIMER2	*/
	long	instr_count;
	} CPU;

/**********************************************************************/
/*   Indexes  for  the  registers.  Not  used  most  of the time but  */
/*   occasionally useful.					      */
/**********************************************************************/
# define	regA	0
# define	regB	1
# define	regC	2
# define	regD	3

# define	A	reg[regA]
# define	B	reg[regB]
# define	C	reg[regC]
# define	D	reg[regD]

# define	R0	Rreg[0]
# define	R1	Rreg[1]
# define	R2	Rreg[2]
# define	R3	Rreg[3]
# define	R4	Rreg[4]

/**********************************************************************/
/*   Field types.						      */
/**********************************************************************/
# define	FIELD_P		0
# define	FIELD_WP	1
# define	FIELD_XS	2
# define	FIELD_X		3
# define	FIELD_S		4
# define	FIELD_M		5
# define	FIELD_B		6
# define	FIELD_W		7
# define	FIELD_XP	0xe
# define	FIELD_A		0xf

symtab_t	*lookup_by_addr PROTO((long));
void	add_64 PROTO((WORD64 *, WORD64 *));
void	sub_64 PROTO((WORD64 *, WORD64 *));
unsigned long add_dec PROTO((unsigned long, unsigned long, int));
unsigned long tens_complement PROTO((unsigned long, int));

void	shift_left_nibbles PROTO((WORD64 *, int));
void	rotate_left_nibbles PROTO((WORD64 *, int));
void	shift_right_nibbles PROTO((WORD64 *, int));
void	rotate_right_nibbles PROTO((WORD64 *, int));
WORD64	*extract_field PROTO((WORD64 *, int));
WORD64	*extract_indirect_field PROTO((long, int));
void	replace_field PROTO((WORD64 *, int, WORD64 *));
void	clear_field PROTO((WORD64 *, int));

void	do_lcd PROTO((void));
void	do_intr PROTO((int, long *));
void	x11_init PROTO((int *, char **));
void	x11_refresh_window PROTO((void));
void	x11_clear PROTO((void));
void	x11_flush PROTO((void));

