#include "hp48.h"
#include "dis48.h"

char	hex[] = "0123456789ABCDEF";

char	*InStr0[16] = {
	"retsetxm", "ret", "retsetc", "retclrc", "sethex", "setdec",
	"push.a\t\tc", "pop.a\t\tc", "clr.x\t\tst", "move.x\t\tst, c",
	"move.x\t\tc, st", "swap.x\t\tc, st", "inc.1\t\tp", "dec.1\t\tp",
	"AND/OR", "reti"
};

char	*OpStr0[8] = {
	"b, a", "c, b", "a, c", "c, d", "a, b", "b, c", "c, a", "d, c"
};

char	*InStr9[8] = {"eq", "ne", "z", "nz", "gt", "lt", "ge", "le"};

char	*OpStr9[4] = {"a, b", "b, c", "c, a", "d, c"};

char	*OpStraf[16] = {
	"b, a", "c, b", "a, c", "c, d", "a, a", "b, b", "c, c", "d, d",
	"a, b", "b, c", "c, a", "d, c", "b, a", "c, b", "a, c", "c, d"
};

#ifdef ANSI
char	*Instr1(NAddr *addr, char *out);
char	*Instr8(NAddr *addr, char *out);
#else
char	*Instr1();
char	*Instr8();
#endif

NAddr
#ifdef ANSI
Instruction(NAddr addr, char *out)
#else
Instruction(addr, out)
NAddr	addr;
char	*out;
#endif
{
	Nybble	n;
	Nybble	fn;
	char	*p;
	char	c;
	NAddr	disp, pc;
	
	switch (n = get_nibble(addr++)) {
	case 0:
		if ((n = get_nibble(addr++)) != 0xe) {
			p = AppendStr(out, InStr0[n]);
			if ((n < 4) || (n == 0xf))
				itype = ujump;
				
			break;
		}
		
		fn = get_nibble(addr++);
		n = get_nibble(addr++);
		p = AppendStr(out, (n < 8) ? "and" : "or");
		p = AppendField(p, fn);
		APPEND_TAB(p);
		p = AppendStr(p, OpStr0[n & 7]);
		break;
		
	case 1:
		p = Instr1(&addr, out);
		break;
		
	case 2:
		p = AppendStr(out, "move.1\t\t");
		p = AppendImmNyb(p, &addr, 1);
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'p');
		break;
		
	case 3:
		p = AppendStr(out, "move.p");
		fn = get_nibble(addr++);
		p = Append16(p, fn);
		APPEND_TAB(p);
		if ((p - out) < 8)
			APPEND_TAB(p);
			
		p = AppendImmNyb(p, &addr, fn + 1);
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'c');
		break;
		
	case 4:
	case 5:
		pc = addr;
		disp = GetInt(&addr, 2);
		if (disp == 2) {
			p = AppendStr(out, "nop3");
			break;
		}
		
		p = AppendStr(out, (disp == 0) ? "retc" : "brc");
		APPEND_CHAR(p, (n == 4) ? 's' : 'c');
		if (disp != 0) {
			APPEND_TAB(p);
			APPEND_TAB(p);
			p = AppendRAddr(p, pc, disp, 2, 1);
		} else
			target = NOADDR;
		
		itype = branch;
		break;
	
	case 6:
		pc = addr;
		disp = GetInt(&addr, 3);
		if ((disp == 3) || (disp == 4)) {
			p = AppendStr(out, "nop");
			APPEND_BDIGIT(p, disp + 1);
			if (disp == 4)
				addr++;
				
			break;
		}
			
		p = AppendStr(out, "jump.3\t\t");
		p = AppendRAddr(p, pc, disp, 3, 1);
		itype = jump;
		break;
		
	case 7:
		pc = addr + 3;
		disp = GetInt(&addr, 3);
		p = AppendStr(out, "call.3\t\t");
		p = AppendRAddr(p, pc, disp, 3, 4);
		itype = call;
		break;
		
	case 8:
		fn = get_nibble(addr);	/* peek! */
		if ((fn != 0xa) && (fn != 0xb)) {
			p = Instr8(&addr, out);
			break;
		}
		
		/* fall through */
	case 9:
		if (n == 8) {
			c = (char) ((fn == 0xa) ? 0 : 1);
			fn = 0xf;
			addr++;
			
		} else {
			fn = get_nibble(addr++);
			c = (char) ((fn < 8) ? 0 : 1);
			fn &= 7;
		}

		n = get_nibble(addr++);
		pc = addr;
		disp = GetInt(&addr, 2);
		p = AppendStr(out, (disp == 0) ? "ret" : "br");
		p = AppendStr(p, InStr9[((n >> 2) & 3) + 4 * c]);
		p = AppendField(p, fn);
		if ((p - out) < 8)
			APPEND_TAB(p);
			
		if ((c == 0) && (n >= 8)) {
			APPEND_CHAR(p, (n & 3) + 'a');
			
		} else {
			p = AppendStr(p, OpStr9[n & 3]);
		}
		
		if (disp != 0) {
			APPEND_COMMA(p);
			p = AppendRAddr(p, pc, disp, 2, 3);
		} else
			target = NOADDR;
		
		itype = branch;
		break;
		
	default:
		switch (n) {
		case 0xa:
			fn = get_nibble(addr++);
			c = (char) ((fn < 8) ? 0 : 1);
			fn &= 7;
			disp = 0xa;
			break;
			
		case 0xb:
			fn = get_nibble(addr++);
			c = (char) ((fn < 8) ? 0 : 1);
			fn &= 7;
			disp = 0xb;
			break;
			
		case 0xc:
		case 0xd:
			fn = 0xf;
			c = (char) (n & 1);
			disp = 0xa;
			break;
			
		case 0xe:
		case 0xf:
			fn = 0xf;
			c = (char) (n & 1);
			disp = 0xb;
			break;
		}
		
		n = get_nibble(addr++);
		pc = 0;
		switch (disp) {
		case 0xa:
			if (c == 0) {
				if (n < 0xc) {
					p = "add";
				
				} else {
					p = "dec";
					pc = 1;
				}
				
			} else {
				if (n < 4) {
					p = "clr";
					pc = 1;
					
				} else if (n >= 0xc) {
					p = "swap";
				
				} else {
					p = "move";
					if (n < 8)
						n -= 4;
				}
			}
			
			break;
			
		case 0xb:
			if (c == 0) {
				if (n >= 0xc) {
					p = "subn";
				
				} else if ((n >= 4) && (n <= 7)) {
					p = "inc";
					pc = 1;
					n -= 4;
				
				} else {
					p = "sub";
				}
				
			} else {
				pc = 1;
				if (n < 4)
					p = "sln";
					
				else if (n < 8)
					p = "srn";
					
				else if (n < 0xc)
					p = "neg";
					
				else
					p = "not";
			}
			
			break;
		}
		
		p = AppendStr(out, p);
		p = AppendField(p, fn);
		if ((p - out) < 9)
			APPEND_TAB(p);
			
		if (pc == 1)
			APPEND_CHAR(p, (n & 3) + 'a');
		
		else
			p = AppendStr(p, OpStraf[n]);
			
		break;
	}
	
	TERMINATE(p);
	return(addr);
}
char	*OpStr1[8] = {
	"a, @d0", "a, @d1", "@d0, a", "@d1, a",
	"c, @d0", "c, @d1", "@d0, c", "@d1, c"
};

char *
#ifdef ANSI
Instr1(NAddr *addr, char *out)
#else
Instr1(addr, out)
NAddr	*addr;
char	*out;
#endif
{
	Nybble	n;
	Nybble	fn;
	char	*p;
	char	c;
	
	switch (n = get_nibble((*addr)++)) {
	case 0:
	case 1:
		p = AppendStr(out, "move.w\t\t");
		fn = get_nibble((*addr)++);
		c = (char) ((fn < 8) ? 'a' : 'c');
		fn = (Nybble) ((fn & 7) + '0');
		if (n == 0)
			APPEND_CHAR(p, c);
		else {
			APPEND_CHAR(p, 'r');
			APPEND_CHAR(p, fn);
		}
		
		APPEND_COMMA(p);
		if (n == 0) {
			APPEND_CHAR(p, 'r');
			APPEND_CHAR(p, fn);
			
		} else
			APPEND_CHAR(p, c);
		
		break;
		
	case 2:
		p = AppendStr(out, "swap.w\t\t");
		n = get_nibble((*addr)++);
		APPEND_CHAR(p, (n < 8) ? 'a' : 'c');
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'r');
		APPEND_BDIGIT(p, n & 7);
		break;
	
	case 3:
		n = get_nibble((*addr)++);
		p = AppendStr(out, (n & 2) ? "swap" : "move");
		APPEND_CHAR(p, '.');
		APPEND_CHAR(p, (n < 8) ? 'a' : '4');
		APPEND_TAB(p);
		APPEND_TAB(p);
		APPEND_CHAR(p, (n & 4) ? 'c' : 'a');
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'd');
		APPEND_BDIGIT(p, n & 1);
		break;
		
	case 4:
	case 5:
		p = AppendStr(out, "move");
		fn = get_nibble((*addr)++);
		if (n == 4) {
			APPEND_CHAR(p, '.');
			APPEND_CHAR(p, (fn < 8) ? 'a' : 'b');
			APPEND_TAB(p);
		
		} else {
			n = get_nibble((*addr)++);
			if (fn < 8)
				p = AppendField(p, n);
				
			else {
				APPEND_CHAR(p, '.');
				p = Append16(p, n);
				APPEND_TAB(p);
			}
		}
		
		if ((p - out) < 8)
			APPEND_TAB(p);
			
		p = AppendStr(p, OpStr1[fn & 7]);
		break;
	
	case 6:
	case 7:
		p = AppendStr(out, "add.a\t\t");
		APPEND_IMMMARK(p);
		p = Append16(p, get_nibble((*addr)++));
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'd');
		APPEND_BDIGIT(p, n & 1);
		break;
		
	default:
		c = (char) ((n < 0xc) ? '0' : '1');
		if ((n & 3) == 0) {
			p = AppendStr(out, "sub.a\t\t");
			APPEND_IMMMARK(p);
			p = Append16(p, get_nibble((*addr)++));
		
		} else {
			p = AppendStr(out, "move.");
			switch (n & 3) {
			case 1: n = 2; break;
			case 2: n = 4; break;
			case 3: n = 5; break;
			}
			
			APPEND_BDIGIT(p, n);
			APPEND_TAB(p);
			APPEND_TAB(p);
			p = AppendImmNyb(p, addr, n);
		}
		
		APPEND_COMMA(p);
		APPEND_CHAR(p, 'd');
		APPEND_CHAR(p, c);
		break;
	}
	
	TERMINATE(p);
	return(p);
}

