/**********************************************************************/
/*   Routines  to  grab  current  contents  of the LCD and cause the  */
/*   LCD to be redrawn.						      */
/**********************************************************************/
# include	"hp48.h"
# include	"proto.h"

# define	DABR		0x120	/* Display Address Base Register */
# define	MENU_OFFSET	0x128	/* Offset to where menu starts.	  */
# define	MENU_BITMAP	0x130	/* Place where menu keys are located */

/**********************************************************************/
/*   Each row is a multiple of 8 bits long.			      */
/**********************************************************************/
# define	LCD_ROWS	64
# define	LCD_COLUMNS	132
# define	BYTES_PER_ROW	17

# define	NIBBLES_PER_BYTE	2
typedef struct row_t {
	char	bits[BYTES_PER_ROW];
	} row_t;

row_t	lcd[LCD_ROWS];
row_t	old_lcd[LCD_ROWS];
extern CPU cpu;

long	draw_row PROTO((long, int));
/**********************************************************************/
/*   Draw the LCD on the display.				      */
/**********************************************************************/
void
draw_lcd()
{	long	addr;
	int	menu_offset;
	int	i;

	/***********************************************/
	/*   Save  old  copy  of  LCD  and  just draw  */
	/*   ints which are different.		       */
	/***********************************************/
	memcpy((char *) &old_lcd, (char *) &lcd, sizeof lcd);
	addr = read_nibbles(DABR, 5);
	menu_offset = read_nibbles(MENU_OFFSET, 2) & 0x3f;
	for (i = 0; i <= menu_offset; i++)
		addr = draw_row(addr, i);

	/***********************************************/
	/*   If  we  have  a  menu  area then display  */
	/*   that as well.			       */
	/***********************************************/
	if (i < LCD_ROWS) {
		addr = read_nibbles(MENU_BITMAP, 5);
		for ( ; i < LCD_ROWS; i++)
			addr = draw_row(addr, i);
		}
}
/**********************************************************************/
/*   Grab a row of pixels and update bits which have changed.	      */
/**********************************************************************/
long
draw_row(addr, row)
long	addr;
int	row;
{	int	i, v, col;
#ifdef MY48
 	register int adp;
 	register char *p=&lcd[row].bits[0];
 	for(i=BYTES_PER_ROW,adp=addr;--i>=0;adp+=2)
 	    *p++=cpu.ram[adp]|(cpu.ram[adp+1]<<4);
#else
	memcpy(&lcd[row].bits[0], &cpu.ram[addr >> 1], BYTES_PER_ROW);
#endif
	for (i = 0; i < BYTES_PER_ROW; i++) {
		v = lcd[row].bits[i];
		if (v != old_lcd[row].bits[i]) {
			col = i * NIBBLES_PER_BYTE;
			x11_draw_nibble(col, row, v & 0x0f);
			x11_draw_nibble(col+1, row, (v >> 4) & 0x0f);
			}
		}
	return addr + (BYTES_PER_ROW * 2);
}
/**********************************************************************/
/*   Clear LCD so we can redraw it from scratch.		      */
/**********************************************************************/
void
clear_lcd()
{
	memset((char *) lcd, 0, sizeof lcd);
	x11_clear();
}
