/**********************************************************************/
/*   Routines for reading and manipulating the symbol table.	      */
/**********************************************************************/
# include	"hp48.h"
# include	<stdio.h>
/*# include	<malloc.h>*/
# include	<string.h>
# include	<stdlib.h>

extern char acOut[512];

# define	SIZE_INCR	1024
symtab_t	*sym_tbl;
int		sym_count;
int		sym_used;

void
read_symtab(filename)
char	*filename;
{	FILE	*fp;
	char	*cp;
	int	i;
	symtab_t	*sp;
	long	addr;
	int	len;
	char	lbuf[BUFSIZ];

	sym_count = SIZE_INCR;
	sym_tbl = (symtab_t *) malloc(sizeof(symtab_t) * sym_count);
	if (sym_tbl == NULL) {
		fprintf(stderr, "Cannot allocate memory for symbol table\n");
		exit(1);
		}
	sym_used = 0;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		perror(filename);
		return;
		}
	while (!feof(fp)) {
		if (fgets(lbuf, sizeof lbuf - 1, fp) == NULL)
			break;
		/***********************************************/
		/*   Make sure have exactly five hex digits    */
		/*   at start of line.			       */
		/***********************************************/
		addr = 0;
		for (i = 0; i < 5; i++) {
			int	ch = lbuf[i];
			if (ch >= '0' && ch <= '9')
				addr = (addr << 4) + ch - '0';
			else if (ch >= 'A' && ch <= 'F')
				addr = (addr << 4) + ch - 'A' + 10;
			else if (ch >= 'a' && ch <= 'f')
				addr = (addr << 4) + ch - 'f' + 10;
			else
				break;
			}
		if (i != 5)
			continue;
# if defined(MSDOS)
		if (addr >= 0x20000 && addr < 0x70000 &&
		    (lbuf[8] != 'H' || lbuf[9] != 'P' || lbuf[10] != ':'))
			continue;
# endif
		/***********************************************/
		/*   If same address as previous then just     */
		/*   junk it.				       */
		/***********************************************/
		if (sym_used && addr == sym_tbl[sym_used-1].addr)
			continue;

		if (sym_used >= sym_count - 1) {
			sym_count += SIZE_INCR;
			sp = (symtab_t *) realloc(sym_tbl, sym_count * sizeof(symtab_t));
			if (sp == NULL) {
				fprintf(stderr, "Out of memory reading symbol table\n");
				return;
				}
			sym_tbl = sp;
			}
		sp = &sym_tbl[sym_used++];
		sp->addr = addr;
		sp->flags = 0;
		if (lbuf[6] == 'P')
			sp->flags = SF_PREFIX;
		for (cp = lbuf + 8; *cp == ' '; )
			cp++;
		len = strlen(cp);
		if (cp[len-1] == '\n')
			cp[len-1] = '\0';
		sp->comment = strdup(cp);
		if (sp->comment == (char *) NULL) {
			fprintf(stderr, "Out of memory allocating symbol entry\n");
			exit(1);
			}
		}
	fclose(fp);
	sprintf(acOut,"Symbols read: %d\n", sym_used); nx_print(acOut);
}
int
cmp(p1, p2)
long	p1;
symtab_t	*p2;
{	long	l = p1 - p2->addr;
	if (l < 0)
		return -1;
	if (l > 0)
		return 1;
	return 0;
}
symtab_t *
lookup_by_addr(addr)
long	addr;
{	symtab_t	*sp, *spend;

	spend = &sym_tbl[sym_used];
	sp = sym_tbl;
	sp = (symtab_t *) bsearch(addr, sym_tbl, sym_used, sizeof(symtab_t), cmp);
	return sp;
}
char *
symbol(addr)
long	addr;
{	symtab_t *sp = lookup_by_addr(addr);
	static char buf[10];

	if (sp)
		return sp->comment;
	sprintf(buf, "$%05X", addr);
	return buf;
}
char *
label(addr)
long	addr;
{	symtab_t	*sp;
	symtab_t	*spend = &sym_tbl[sym_used];
	static char buf[64];
	int	diff;

	if (addr == 0)
		return "";

	for (sp = sym_tbl+1; sp < spend; sp++) {
		if (sp->addr == addr)
			return sp->comment;
		if (sp->addr > addr) {
			diff = addr - sp[-1].addr;
			if (diff >= 0x100)
				break;
			sprintf(buf, "%.48s+%lx",
				sp[-1].comment, diff);
			return buf;
			}
		}
	return "";
}
